// cfplexer.h
// This file holds the classes for the configpp lexer, which manages the file reading and
//  extraction of tokens for parsing.

#ifndef INC_cfplexh
#define INC_cfplexh

#include "cfpinc.h"

//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace Configpp{





//--------------------------------------------------------------------------------
//---- CfpInputSource ------------------------------------------------------------
//--------------------------------------------------------------------------------

// CfpInputSource is base class for a file or a string from which tokens are popped off;
//  we use a stack of these sources so that we can do nested includes, etc.

class CfpInputSource
{
protected:
  Configpp *configppp;					// pointer to configpp class
  istream *inputstreamp;				// pointer to the string or file input stream
  string streamname;					// name of file or string (for error reporting)
  long position;					// current position in stream
  long curline;						// current line number
  long curlinepos;					// current offset into line
public:
  CfpInputSource();					// constructor
  ~CfpInputSource();					// destructor
public:
  virtual void destruct()=0;				// returns either isfile or isstring (=0 makes it pure)
  virtual void openstream()=0;				// allocate and open the stream
  virtual void close()=0;				// close stream
  bool reopenclosed();					// reopen the stream if it has been closed (ret true)
  void positionreset();					// reset the internal position vars
public:
  virtual CfpInputSourceType sourcetype()=0;		// returns either isfile or isstring (=0 makes it pure)
  virtual string gettypename()=0;			// returns either 'file' or 'string' (=0 makes it pure)
  istream *getstreamp() {return inputstreamp;};		// accessor
  string getname() {return streamname;};		// accessor
  long getpos() {return position;};			// accessor
  long getcurline() {return curline;};			// accessor
  long getcurlinepos() {return curlinepos;};		// accessor
public:
  virtual int readchar();				// get a char from the stream
  virtual int peekchar();				// peek a char from the stream
  virtual string getcurfilename() {return "";};		// accessor
};






//--------------------------------------------------------------------------------
//---- CfpSourceFile -------------------------------------------------------------
//--------------------------------------------------------------------------------

// File input stream source

class CfpSourceFile:public CfpInputSource
{
private:
  string extname;
public:
  CfpSourceFile(Configpp *inconfigppp,string filename, const string ext="");
  virtual ~CfpSourceFile() {cout << "DESTRUCT LOCAL";};
  virtual void openstream();					// allocate and open the stream
  virtual void close();						// close the string, but keep misc.
  virtual void destruct();					// replacement for delete
public:
  virtual CfpInputSourceType sourcetype() {return filetype;};	// rtti-like accessor
  virtual string gettypename() {return "File";};		// rtti-like accessor
public:
  virtual int peekchar();					// peek a char from the stream
  virtual string getcurfilename();				// accessor
};





//--------------------------------------------------------------------------------
//---- CfpSourceString -----------------------------------------------------------
//--------------------------------------------------------------------------------

// String input stream source

class CfpSourceString:public CfpInputSource
{
private:
  char *parsecopy;							// buffer of the incoming string
public:
  CfpSourceString(Configpp *inconfigppp,string parsestring, string inname="");	// constructor
  virtual void close();							// close the string, but keep misc.
  virtual void destruct();						// replacement for delete
  virtual void openstream();						// allocate and open the stream
public:
  virtual reopenclosed() {;};						// stringstreams always open
  virtual CfpInputSourceType sourcetype() {return stringtype;};		// rtti-like accessor
  virtual string gettypename() {return "String";};			// rtti-like accessor
};











//--------------------------------------------------------------------------------
//---- CfpToken ------------------------------------------------------------------
//--------------------------------------------------------------------------------

// CfpToken holds a single abstract lexer token, and the best guess about where it was from

class CfpToken
{
private:
  CfpTokenType tokentype;				// see cfpsymbols.h for enumerations
  string tokenval;					// string value of token
  CfpLoc streamloc;					// location where token was scanned
public:
  CfpToken(CfpTokenType intt=initVal,string intv="");	// constructor
  void set(CfpTokenType intype, string instrval);	// set the token values
public:
  void reset() {set(initVal,"");};			// reset the token to uninitialized state
  CfpLoc *getlocp() {return &streamloc;};		// accessor
  CfpTokenType gettype() {return tokentype;};		// accessor
  string getstr()	 {return tokenval;};		// accessor
  void seteoflag() {streamloc.seteoflag();};		// set the loc's eof flag
};

// non-member token inserter function for debugging tokens
ostream &operator<< (ostream &stream, CfpToken token);	// replacement inserter, for printing debug info



//--------------------------------------------------------------------------------
//---- CfpLexer ------------------------------------------------------------------
//--------------------------------------------------------------------------------

// The CfpLexer class is used to manage the input streams and the extraction of tokens
//  It pushes files and strings onto a stack and pops tokens off it.

class CfpLexer
{
private:
  Configpp *configppp;					// pointer to parent parser
  CfpToken curtoken;					// current token being parsed
  bool semihint;					// hint whether we think they might forgoteen semicolon
private:
  deque<CfpToken> peektokens;				// stack of peeked tokens
  deque<CfpInputSource *> inputhistory;			// history of all input sources (for error reports)
  deque<CfpInputSource *> inputs;			// stack of currently active input sources
public:
  CfpLexer(Configpp *inconfigppp);			// constructor
  ~CfpLexer();						// constructor
public:
  void pushfile(string filename);			// push a file on stack
  void pushstring(string parsestring, string name="");	// push a string on stack
  void popsource();					// pop the current stream off the stack
public:
  CfpToken readtoken(TokenReadMode rmode=tokenMode,bool skipping=false);	// read and return a lexical token
  CfpToken peektoken(TokenReadMode rmode=tokenMode,bool advance=false);		// peek and return a lexical token

  void pushtoken(CfpToken toke) {peektokens.push_front(toke);};			// push token on peek stack

  void consumepeeks();								// consume peeked tokens
  CfpToken readpuretoken();	// read and return a lexical token
public:
  CfpToken *getcurtokenp() {return &curtoken;};		// accessor
  long getcurline();					// accessor
  long getcurlinepos();					// accessor
  CfpInputSource *getcursourcep();			// accessor
  CfpLoc *getlocp() {return curtoken.getlocp();};	// accessor
  bool getsemihint() {return semihint;};		// accessor
  long getcurpos();					// accessor
  string getcurfilename();				// accessor
  CfpInputSource *getcursource();			// accessor
private:
  void internalreadtoken(TokenReadMode rmode);		// read in a lexical token. the big function.
  int lexreadchar();					// get a char from the front stream, can be eof
  int lexpeekchar();					// peek a char from the front stream, and be eof
private:
  CfpTokenType toketypechar(int c);			// figure out token type from first char
  bool ischarseperator(CfpTokenType &ttype,int c,int lc, int len);  // is the char a seperator for the token
public:
  void showpeeks();
};





//}
// Namespace close-brace
#endif


