//  Jesse Reichler, January 1996
//  Fred Delcomyn
//
//  chain.C
//  This file contains the class methods for the Chain class, but not
//   the Link class.
//   see header file (.h) for description.
//  Also the class Rot is also used and stored in rot.C (.h)
//
//

#include "misc.def"
#include "matrixinc.h"
#include "chain.h"
#include "link.h"
#include "rot.h"
#include "world.h"
#include "robot.h"

#include <stdio.h>
#include <string.h>
#include <math.h>



#define ERRORVELOCITY 999999




Chain:: Chain(const char *chainname,World *inworldp,Robot *inrobotp)
{
  // initialize the chain, normally this would imply loading a
  //  chain configuration file with parameters in.
  // Set the world it is in, and the robot it is attached to.
  Link *linkpbase;
  Shape *shapep;
  
  // from config
  robotp=inrobotp;
  
  // zero out any hip info (the legs connect at c.o.m)
  hip_distance.fill(0.0);
  hip_alpha_twist=1.570796327;
  hip_theta_angle=0;

  // initialize
  setthing(chainname,inworldp,NO);
  links=0;

  // All chains have their first link as a dummy virtual Hip link
  // WE must create this link and add it to the world.
  // we must also create and add a dummy shape because every physobject needs a shape
  linkpbase=new Link("baselink",inworldp);
  shapep=new Shape();
  linkpbase->setshape(shapep);
  linkpbase->done_construct();

  // add the hip link to the chain
  addlink(linkpbase);
}


void Chain::sethip(perVectorR3 inhipdistance,percision inat,percision inta)
{
  // setthe hip information.  this must be called *before* done_construct()
  hip_distance=inhipdistance;
  hip_alpha_twist=inat;
  hip_theta_angle=inta;
}

void Chain::sethip(float hipdx,float hipdy,float hipdz,percision inat,percision inta)
{
  // setthe hip information.  this must be called *before* done_construct()
  perVectorR3 inhipdistance;

  inhipdistance(0)=hipdx;
  inhipdistance(1)=hipdy;
  inhipdistance(2)=hipdz;
  hip_distance=inhipdistance;
  hip_alpha_twist=inat;
  hip_theta_angle=inta;
}



void Chain:: done_construct()
{
  // AFTER all links have been added, we allocate and init our
  //  DH transform matrix classes, we pass it our pointer so it
  //  can know who it is attached to.  See rot.C (.h) for details.
  rot.allocate(this, homeworld);
  
  // Lastly, update the global (with respect to frame 0, the world)
  //  position and orientation of the links, so that they may be
  //  imediately drawn.
  updatefordraw(0.0);
}










Chain:: ~Chain()
{
  // deconstruct a chain
  int count;
  
  // Now we release the ROT class, note links must be >0
  rot.deallocate();
  
  // delete each of the links
  for (count=0;count<links;++count)
    delete link[count];
  links=0;
}



void Chain:: addlink(Link *linkp)
{
  // add a link to the END of the chain
  // NOTE: links are added consecutively from the body outward
  
  if (links>=MAX_LINKS_PERCHAIN)
    {
      printf ("$$$ERROR: Too many links (%d) already attached to chain '%s'.\n",links,name);
      return;
    }
  
  link[links]=linkp;
  ++links;
}


void Chain::resetforces()
{
  int count;
  // Clear the previous forces/torques at the joints
  // note we don't reset the userTorq, which is static rather than pulse based.
  
  // clear contact forces
  for (count=0;count<links;++count)
    link[count]->extTorq=0.0;
}



void Chain::getforces(double deltat)
{
  // Get the forces exerted on each part of a chain.
  // This includes actuator created forces
  //  and the ground forces (contact forces) if any
}



void Chain::update(double deltat)
{
  // calc. new position&velocity of chain, given that new accellerations
  //  have already been calculated.
  int count;
  percision accel, veloc;
  int c;
  char tstring[80];
  
  //  cout << "... In chain update 1.\n";

  // integrate via easy, standard Euler method, to calculate the new postition and veloc. of each link.
  //  link[0] is the base so we don't update that one
  for (count=1;count<links;++count)
    {
      // if it is a fixed (ie nonturning link), then don't update
      accel=link[count]->ddtheta_angle;
      link[count]->dtheta_angle+=accel*deltat;

      //      cout << "theta_angle = '" << link[count]->theta_angle << "' \n";
      //      cout << "dtheta_angle = '" << link[count]->dtheta_angle << "' \n";
      //      cout << "ddtheta_angle = '" << link[count]->ddtheta_angle << "' \n";

      // fix up modulus angle
      while (link[count]->theta_angle>M_PI*2)
	link[count]->theta_angle-=(M_PI*2);
      while (link[count]->theta_angle<M_PI*(-2.0))
	link[count]->theta_angle+=(M_PI*2);

      veloc=link[count]->dtheta_angle;
      link[count]->setangle((link[count]->theta_angle)+(veloc + (0.5*accel*deltat))*deltat);

      //      cout << "veloc = '" << veloc << "' \n";

      if (fabs(veloc)>ERRORVELOCITY)
	{
	  printf("%c",7);
	  cout << "\n\n-----------------------------------------------------------------------------\n";
	  cout << "NUMERICAL ERROR: Dynamics equations have become unstable. Could be due to insufficiently small timestep, or excessive forces being generated within simulation.  Sorry about that.\n\n";
	  cout << "-----------------------------------------------------------------------------\n\n";
	  for (;;)
	    {
	      cout << "Would you like to reset velocities and accelerations and attempt to continue (results obtained after this point should not be considered valid)?  [y/n]: ";
	      fgets(tstring,70,stdin);
	      if (tstring[0]=='n' || tstring[0]=='N')
		{
		  cout << "Exiting...\n\n\n";
		  exit(1);
		}
	      if (tstring[0]=='y' || tstring[0]=='Y')
		break;
	    }
	  for (count=1;count<links;++count)
	    {
	      link[count]->dtheta_angle=0.0;
	      link[count]->ddtheta_angle=0.0;
	    }
	  if (robotp!=NULL)
	    robotp->resetbaseav();
	  cout << "Reseting velocities and accelerations and resuming...\n\n\n";
	  break;
	}
    }
  
  //  cout << "... In chain update 2.\n";

  // Lastly, update the global (with respect to frame 0, the world)
  //  position and orientation of the links, so that they may be drawn.
  updatefordraw(deltat);

  //  cout << "... In chain update 3.\n";
}



void Chain::updatefordraw(double deltat)
{
  // Update the global (with respect to frame 0, the world)
  // position and orientation of the links, so that they may be drawn.
  // The rot matrices, which hold info on joint angles,
  // must have been updated sometime previous to this function.
  int count;
  perVector tempv1(6);
  
  // Make the call to update the DH transform rotational matrices
  //  this must be done after the angle update, to keep
  //  the matrices always up to date
  rot.update();

  // the base of the chain is where we start, this should be in
  //  world frame.  Even when the mobile base is used, link 0 is
  //  the copy of the base orient and loc in world coord.
  
  // now iteratively calculate the positions and orientation
  //  w.r.t. the world frame
  //  for (count=0;count<links;++count)
  for (count=1;count<links;++count)
    {
      // The location of a link is the pos. vect. wrt. frame 0;
      // end location is just the pos. of the next link;
      tempv1=rot.DTp(RI_EARTH,count);

      // first set orientation
      link[count]->setorientation(rot.DTA33(RI_EARTH,count));

      // set endpoint
      //      (link[count]->shape)->setendlocation(rot.DTp(RI_EARTH,count+1));

      if (deltat!=NODELTAT)
	{
	  link[count]->setlocation(tempv1,deltat);
	}
      else
	{
	  link[count]->setlocation(tempv1); // don't touch linear velocity
	  // but we DO need to update the contact locations
	  link[count]->updatecontactlocs(deltat);
	}
    }
}



void Chain::draw(GrobVideoCamera *cam)
{
  // draw the chain
  int count;
  perVectorR3 p1,p2;
  Base *basep;
  
  // draw links
  for (count=1;count<links;++count)
    link[count]->draw(cam);
}











// The new dynamics routines











void Chain::iterate(double deltat)
{
  // NEW leg dynamics routines, designed for multichain mobile base
  
  //  cout << "chain iterate 1.\n";
  // 1. Initialize any calculations
  setbaselinks(deltat);
  
  //  cout << "chain iterate 2.\n";
  // 2. Forward leg kinematics - propagate kinematic info from base to feet
  forward_kinematics(deltat);
  
  //  cout << "chain iterate 3.\n";
  // 3. Calculate external forces at each link, specifically foot (ground contact) forces
  getforces(deltat);
  foot_forces(deltat);
  
  //  cout << "chain iterate 4.\n";
  // 4. Backward leg dynamics - propagate information from the foot to the hip
  backward_dynamics(deltat);

  //  cout << "chain iterate 5.\n";
  // 5. Hip equations - calculate how legs influence the acceleration of the base
  hip_equations(deltat);

  //  cout << "chain iterate 6.\n";
}









void Chain::movie_load(FILE *moviefp)
{
  // load all data for a chain
  int count;

  // load the links
  for (count=0;count<links;++count)
    link[count]->movie_load(moviefp);
  
  // NOW, in the new movie format, we must recalc positions and orients,
  //  given angles, to save space.
  double deltat=NODELTAT;
  
  // 2/4/99 trying without updatefordraw
  //  updatefordraw(deltat);
}


void Chain::movie_save(FILE *moviefp)
{
  // save all data for a chain
  int count;
  
  // save the links
  for (count=0;count<links;++count)
    link[count]->movie_save(moviefp);
}











// NOTE:
// Methods:
//  forward_kinematics(), hip_equations(), footforces(),fillNmat() 
//  in file dynamics.C



