//  Jesse Reichler, January 1996
//  Fred Delcomyn
//
//  physobject.C
//  This file contains the class definition for the Physobject class,
//   for description see physobject.h
//  
//

#include "misc.def"
#include "world.h"
#include "physobject.h"
#include "misc.h"
#include "contact.h"

#include <math.h>
#include <stdlib.h>
#include <string.h>

bool showforce=false;
double forcezoom=1.0;


// ------------------------------------------------------------------------------------------
// -- Physicalobject members
// ------------------------------------------------------------------------------------------


Physicalobject:: Physicalobject(const char *objname, World *worldp, int wowned)
{
  // initialize the object
  setthing(objname,worldp,wowned);

  // null out the shape pointer (warning - a shape MUST be added)
  shape=(Shape *)NULL;
  // to keep disasters from happening, let's allocate a default shape, and then free it when we reset it.
  shape=new Shape();

  // allocate vectors
  velocity=perVector(3);
  inertia=perMatrix(3,3);
  Com=perVector(3);

  
  // Set initial mass based data  -  user should fill these in
  inertia.fill(0.0);
  Com.fill(0.0);
  kilograms=0.0;


  // clear external forces
  velocity.fill(0.0);
  clearextforce();

  // Now attach the physical object to the world
  worldp->addthing(this);

  // test2
  inertia(0,0)=.0008;
  inertia(1,1)=.02;
  inertia(2,2)=.02;
  //  Com(0)=0;
  //  Com(1)=0;
  //  Com(2)=0;
  kilograms=.03;
}







Physicalobject:: ~Physicalobject()
{
  // deconstruct the object
  if (shape!=(Shape *)NULL)
    delete shape;
}




void Physicalobject:: setshape(class Shape *setshape)
{
  // set the shape of the object

  // first, we check if the current shape is a blank placeholder.  if so, release it.
  if (shape!=NULL)
    {
      if (shape->shapetype()==SHAPE_NOSHAPE)
	{
	  // note that what we might want to do before deleting it is copy any relevant info (like location)
	  perVectorR3 oldloc;
	  oldloc=shape->getlocation();
	  delete shape;
	  shape=setshape;
	  shape->setlocation(oldloc);
	  return;
	}
    }
  
  shape=setshape;
}



void Physicalobject:: draw(GrobVideoCamera *cam)
{
  // draw the object
  if (shape)
    shape->draw(cam);
  // might want to draw some force lines
  if (shapetype()!=SHAPE_TRIANGLE && shapetype()!=SHAPE_MESH)
    drawforce(cam);
}



void Physicalobject:: setlocation(float x,float y,float z)
{
  // set the loc of the object, and update linear veloc (w.r.t world fram).
  perVectorR3 loc;

  loc(0)=x;loc(1)=y;loc(2)=z;
  setlocation(loc);
}


void Physicalobject:: setlocation(perVector loc, percision deltat)
{
  // set the loc of the object, and update linear veloc (w.r.t world fram).
  percision dx,dy,dz;
  perVector oldloc;
  
  if (deltat!=0.0)
    {
      // calculate the veloc if this is not some initialization
      oldloc=locationval();
      dx=loc(0)-oldloc(0);
      dy=loc(1)-oldloc(1);
      dz=loc(2)-oldloc(2);
      velocity(0)=dx/deltat;
      velocity(1)=dy/deltat;
      velocity(2)=dz/deltat;
    }
  else
    {
      // initialize velocity
      velocity.fill(0.0);
    }
  
  // set the new location
  shape->setlocation(loc);
  
  // now tell the shape to adjust its contact tracking points
  if (shape!=(Shape *)NULL)
    shape->updatecontactlocs(this,deltat);
}



void Physicalobject::clearextforce()
{
   // Clear any external forces that are present on the object
   extF.fill(0.0);
   extN.fill(0.0);
}


void Physicalobject::addextforce(perVectorR3 F, perVectorR3 N)
{
  // Add the passed external forces to the current external forces
  extF=extF+F;
  extN=extN+N;
}







void Physicalobject::movie_load(FILE *moviefp)
{
  // load all data for a physobject
  //  perVector inlocation(3);
  //  perMatrix inorientation(3,3);

  (shape->grobp)->readlocation(moviefp);
  (shape->grobp)->readorientation(moviefp);
  Jvecload(shape->endlocation,moviefp,3);

  //  Jvecload(inlocation,moviefp,3);
  //  setlocation(inlocation);
  //  Jmatload(inorientation,moviefp,3,3);
  //  setorientation(inorientation);

  Jvecload(extF,moviefp,3);

  // these could be saved&loaded if we want this info
  //  Jvecload(userextF,moviefp,3);
  //  Jvecload(extN,moviefp,3);
  //  Jvecload(userextN,moviefp,3);
  //  Jvecload(velocity,moviefp,3);
}

void Physicalobject::movie_save(FILE *moviefp)
{
  // save all data for a base
  //  perVector inlocation(3);
  //  perMatrix inorientation(3,3);

  (shape->grobp)->writelocation(moviefp);
  (shape->grobp)->writeorientation(moviefp);
  Jvecsave(shape->endlocation,moviefp,3);

  //  inlocation=locationval();
  //  Jvecsave(inlocation,moviefp,3);
  //  inorientation=orientationval();
  //  Jmatsave(inorientation,moviefp,3,3);
  Jvecsave(extF,moviefp,3);

  // these could be saved if we want this info
  //  Jvecsave(extN,moviefp,3);
  //  Jvecsave(userextF,moviefp,3);
  //  Jvecsave(userextN,moviefp,3);
  //  Jvecsave(velocity,moviefp,3);
}






















// ------------------------------------------------------------------------------------------
// -- Base Shape members
// ------------------------------------------------------------------------------------------


Shape:: Shape()
{
  // base constructor
  contactpoints=0;
  smartsprings=NO;
  grobp=NULL;

  // allocate the Grob graphic object
  grobp=new Grob();
}


Shape:: ~Shape()
{
  // Destructor for shape
  int count;
  
  // free the Grob
  if (grobp)
    delete grobp;
  
  // free up the contact points
  for (count=0;count<contactpoints;++count)
    delete point[count];
  contactpoints=0;
}




void Shape::addcontacts(Physicalobject *thisobj,Physicalobject *obj2)
{
  // Create one contact for each contact point on this shape,
  //  and register it with the world
  int count;
  int ss;
  PointContact *pc;
  
  if (smartsprings==YES || (obj2->shape)->smartsprings==YES)
    ss=YES;
  else
    ss=NO;

  for (count=0;count<contactpoints;++count)
    {
      pc=new PointContact(thisobj,obj2,point[count],ss);
      (thisobj->homeworld)->addcontact(pc);
    }
}


void Shape::addpoint(perVectorR3 comdist)
{
  // add a point track
  PointTrack *pt;

  pt=new PointTrack(comdist);
  point[contactpoints++]=pt;
}







// ------------------------------------------------------------------------------------------
// -- Derived Shape members
// ------------------------------------------------------------------------------------------


// ------------------------------------------------------------------------------------------
// -- Sphere
// ------------------------------------------------------------------------------------------

Sphere::Sphere(percision inradius)
  : Shape()
{
  radius=inradius;
  makegrob();
  makecontactpoints();
}



void Sphere::makecontactpoints()
{
  // allocate and register the potential contact points on the shape
  // make one point at the center
  perVectorR3 comdist;

  comdist(0)=0.0; comdist(1)=0.0; comdist(2)=0.0;
  addpoint(comdist);
}


void Sphere::updatecontactlocs(Physicalobject *thisobj,percision deltat)
{
  // update the contact location points
  point[0]->updatepoint(thisobj->locationval(),deltat);
}


void Sphere::makegrob()
{
  // tell the grob how to make us
  grobp->makeSphereList(radius);
  grobp->setdetail(meddetail);
}




// ------------------------------------------------------------------------------------------
// -- Cube
// ------------------------------------------------------------------------------------------

Cube::Cube (percision len1,percision len2,percision len3)
  : Shape()
{
  xlength=len1;
  ylength=len2;
  zlength=len3;
  makegrob();
  makecontactpoints();
}


void Cube::makegrob()
{
  // tell the grob how to make us
  grobp->makeCubeList(xlength,ylength,zlength);
  grobp->setdetail(meddetail);
}





// ------------------------------------------------------------------------------------------
// -- RoundCube
// ------------------------------------------------------------------------------------------

RoundCube::RoundCube (percision len1,percision len2,percision len3,percision inshift)
  : Shape()
{
  xlength=len1;
  ylength=len2;
  zlength=len3;
  shift=inshift;
  makegrob();
  makecontactpoints();
}


void RoundCube::makegrob()
{
  // tell the grob how to make us
  grobp->makeRoundCubeList(xlength,ylength,zlength,shift);
  grobp->setdetail(meddetail);
}







// ------------------------------------------------------------------------------------------
// -- Legcone
// ------------------------------------------------------------------------------------------

Legcone::Legcone (percision pr, percision dr, percision l, percision wthickness=0,int insleeve=0, percision inoffst=0, int inpointed=0)
  : Shape()
{
  // intialize
  proximal_radius=pr;
  distal_radius=dr;
  length=l;
  wall_thickness=wthickness;
  sleeve=insleeve;
  offst=inoffst;
  pointed=inpointed;
  makegrob();
  makecontactpoints();
}


percision Legcone::radiusat(percision pos)
{
  // Calculate the approx. radius of a cone, at the specified offset,
  //  given that we know proximal_radius, distal_radius, and length
  //  of the cone
  percision retradius;
  percision offset;
  percision deltarad;
  
  // simple to calculate, just think of radius as a function of offset:
  // y=f(x)+b, calculate the slope and then just take radius=f(position)+prox_rad
  offset=pos/length;
  deltarad=distal_radius-proximal_radius;
  
  retradius=proximal_radius + (offset*deltarad);
  return retradius;
}


percision Legcone::volumeat(percision pr, percision dr, percision len)
{
  // Calculate the approx. % volume of a section of a cone
  // [Anton ,3rd ed., p.378 volumes by slicing]
  // Volume is SUM from x=0->len, of PI*f(x)^2
  //  where f(x)=radius at point along axis from 0 to len
  percision vol;
  
  // IMPORTANT!!!! for expediency we use legsim mode of cylinder volume!!!
  // JESSE this should be changed to a real volume of partial cone
  //  5/1/96
  
  vol=M_PI*pow(pr,2)*len;
  return (vol);
}


void Legcone::makecontactpoints()
{
  // allocate and register the potential contact points on the shape
  // make one point at the distal end
  perVectorR3 comdist;

  comdist(0)=0.0; comdist(1)=0.0; comdist(2)=0.0;
  addpoint(comdist);
}


void Legcone::updatecontactlocs(Physicalobject *thisobj,percision deltat)
{
  // update the contact location points
  perVectorR3 sizevec;
  perVectorR3 temp3;
  perVectorR3 distloc;
  
  temp3(0)=length; temp3(1)=0.0;temp3(2)=0.0;
  sizevec=(thisobj->orientationval())*(perVector)temp3;
  distloc=(thisobj->locationval())+sizevec;

  // new test
  endlocation=distloc;

  point[0]->updatepoint(distloc,deltat);
}


void Legcone::makegrob()
{
  // tell the grob how to make us
  grobp->makeLegconeList(proximal_radius,distal_radius,length,wall_thickness,sleeve,offst,pointed);
  grobp->setdetail(meddetail);
}




// ------------------------------------------------------------------------------------------
// -- Triangle
// ------------------------------------------------------------------------------------------


Triangle::Triangle(perVectorR3 inP1,perVectorR3 inP2,perVectorR3 inP3,float inthickness)
  : Shape()
{
  // created a new triangle
  P1=inP1;
  P2=inP2;
  P3=inP3;
  //  N=inN;
  //  k=ink;
  thickness=inthickness;
  plane_equation();
  makegrob();
  makecontactpoints();
}


void Triangle::plane_equation()
{
  // Compute values as per. Jan's plane_equation in impact2 document [Jan 1995]
  
  perVectorR3 D1;		//3x1 vector edge of triangle
  perVectorR3 D2;		//3x1 other vector edge of triangle
  perVectorR3 ia;         //3x1 tangential perpendicular1 (x axis)
  perVectorR3 ja;         //3x1 tangential perpendicular2 (y axis)
  perVectorR3 ka;         //3x1 tangential perpendicular2 (z axis)
  
  // Jans N and k
  D1 = P2 - P1;           //subtraction overloaded operator
  D2 = P3 - P2;   	//subtraction overloaded operator
  N = JessCross(D1,D2); 	//cross product overloaded operator
  k = N.dot(P1);   	//dot product overloaded operator
}


int Triangle:: linecontact(perVectorR3 linep1, perVectorR3 linep2, perVectorR3 &Intersect)
{
  // Compute whether the line touches the triangle tile,
  //  and return YES or NO.
  // Uses Jan's segment_plane_intersect() code from his impact document[Jan 1995]
  
  perVectorR3 B;	//3x1 Base Vector to P1
  perVectorR3 D;	//3x1 Direction Vector from P1 to P2
  percision mu,mu1,mu2;
  perVectorR3 P,Q;

  // The endpoints of the line
  P=linep1;
  Q=linep2;
  
  B = P;
  D = Q - P;
  if (D.dot(N) == 0.0)
    {
      //parallel case
      Intersect.fill(0.0);
      return NO;
    }
  
  mu1=B.dot(N);
  mu2=D.dot(N);
  mu = (k - mu1) / mu2;
  
  Intersect = B + mu * D;

  if (mu < 0.0 || mu > 1.0)	//intersection outside of segment case
    return NO;
  return YES;
}



  
void Triangle::normalframe(perVectorR3 contactloc,perMatrix &orientation)
{
  // Given the unit basis vectors i,j,k, compute the normal,
  //  which is the rotation matrix from World frame to Object frame, which
  //  may have dif. implications for dif. objects.
  // For triangle tiles, the object frame has Z perpendicular OUT of object
  //  and x,y not important.
  // From Spong (1989) p.34
  perVectorR3 i1;
  perVectorR3 j1;
  perVectorR3 k1;
  perVectorR3 i0(1,0,0);
  perVectorR3 j0(0,1,0);
  perVectorR3 k0(0,0,1);
  perVectorR3 D1,D2,N1,tempv;

  orientation=perMatrix(3,3);

  //  cout << "in triangle normalframe.\n";

  // Given the three points (in the world frame) that compose the
  //  triangle tile, compute the object frame, which is:
  // The object frame has Z perpendicular OUT of object and x,y not
  //  important.  Based in part on Spong(1989) p.34
  D1 = P2 - P1;           //subtraction overloaded operator
  D2 = P3 - P2;   	//subtraction overloaded operator
  N1 = JessCross(D1,D2); 	//cross product overloaded operator
  
  // ka (z) axis is just the perpendicular N
  k1=N1;
  // ia and ja axes (x and y) are just any *mutually* perpendicular other axis
  i1=JessCross(N1,D1);
  j1=JessCross(N1,i1);
  
  // Now we have i1,j1,k1 all perpendicular axes, but to get a basis set
  //  we must normalize them.
  i1.normalize_self();
  j1.normalize_self();
  k1.normalize_self();
  
  // Now we have a valid basis set, and can call setorientation to compute
  //  The transformation rotation matrix using the basis set.
  //     printf("Basis set for triangle tile is:\n");
  //     i1.print();
  //     j1.print();
  //     k1.print();
  
  // We want the rotation matrix from WORLD to OBJECT (from frame 0 to 1)
  // Taken directly from Spong (1989) p.34
  orientation(0,0)=i0.dot(i1);
  orientation(1,0)=i0.dot(j1);
  orientation(2,0)=i0.dot(k1);
  orientation(0,1)=j0.dot(i1);
  orientation(1,1)=j0.dot(j1);
  orientation(2,1)=j0.dot(k1);
  orientation(0,2)=k0.dot(i1);
  orientation(1,2)=k0.dot(j1);
  orientation(2,2)=k0.dot(k1);

  //  cout << "okay returning:\n";
  //  orientation.print();
}



void Triangle::makegrob()
{
  // tell the grob how to make us
  grobp->makeTriangletileList(P1(0),P1(1),P1(2),P2(0),P2(1),P2(2),P3(0),P3(1),P3(2),thickness);
  grobp->setdetail(meddetail);
}



void Physicalobject::drawforce(class GrobVideoCamera *cam)
{
  // Draw a force line extending from com of the object
  // external force applied at distal end (already in world frame)
  perVectorR3 sizevec;
  perVectorR3 len;
  perVectorR3 loc;
  perVectorR3 off;
  percision forcescale=.02*forcezoom;
  
  if (!showforce)
    return;

  // get shape ENDPOINT (ie tip of leg or center of sphere)
  loc=shape->getendlocation();
  //  loc(0)=shape->getlocx();
  //  loc(1)=shape->getlocy();
  //  loc(2)=shape->getlocz();

  len=extF+userextF;
  len=len*forcescale;

  if ((len(0)!=0.0 || len(1)!=0.0 || len(2)!=0.0))
  {
    // Now draw a force vector at distal end
    GrobGlHelper::drawLine(loc(0),loc(1),loc(2),loc(0)+len(0),loc(1)+len(1),loc(2)+len(2));
  }

}

















// ------------------------------------------------------------------------------------------
// -- Mesh
// ------------------------------------------------------------------------------------------


Mesh::Mesh(float isx,float iex,float isy,float iey,float isz,float iez)
  : Shape()
{
  // create a new Mesh
  sx=isx;
  ex=iex;
  sy=isy;
  ey=iey;
  sz=isz;
  ez=iez;

  // initialize heights;
  heights=NULL;

  // make the grob
  makedefaultgrob();
}


Mesh::~Mesh()
{
  // free up the height buffer
  if (heights!=NULL)
    delete heights;
}


void Mesh::setdims(int igridsx,int igridsy)
{
  gridsx=igridsx;
  gridsy=igridsy;
}


void Mesh::loadheights(float *inheights,int entries)
{
  // load in grid data
  int x,y;
  float h;

  // do calculations based on dimensions
  if (gridsx<=1 || gridsy<=1)
    {
      cout << "ERROR: surfacemesh must be minimally of size 2x2.\n";
      exit(-1);
    }
  // calculate xdist and ydist (spacing along the grid to conform to sx,ex,sy,ey)
  xdist=(ex-sx)/(gridsx-1);
  ydist=(ey-sy)/(gridsy-1);

  // allocate the buffer
  if (heights!=NULL)
    delete [] heights;
  heights=new float[gridsx*gridsy];

  lowest=999999;
  highest=-999999;

  if (entries!=gridsx*gridsy && entries!=-1)
    {
      cout << "ERROR: incoming surfacegrid is of different dimensions then allocated originally.\n";
      exit(-1);
    }

  //  cout << "Loading Mesh heights.\n";

  for (x=0;x<gridsx;++x)
    {
      for (y=0;y<gridsy;++y)
	{
	  h=heights[index(x,y)]=inheights[index(x,y)];
	  if (h<lowest)
	    lowest=h;
	  if (h>highest)
	    highest=h;
	}
    }

  //  cout << "Filled Mesh.\n";

  // now normalize it to sx,ex,sy,ey,sz,ez
  scale();

  //  cout << "Scaled Mesh.\n";

  // now remake the grob
  makegrob();

  //  cout << "made Grob Mesh.\n";
}


void Mesh::scale()
{
  // go through height data and scale it to sz,ez
  float dist,scale;
  int x,y;
  float h;

  dist=highest-lowest;
  if (dist==0)
    scale=0;
  else
    scale=(ez-sz)/dist;

  for (x=0;x<gridsx;++x)
    {
      for (y=0;y<gridsy;++y)
	{
	  h=heights[index(x,y)];
	  heights[index(x,y)]=sz+(scale*(h-lowest));
	}
    }

  // now readjust lowest and highest
  lowest=sz;
  highest=ez;
}


int Mesh:: linecontact(perVectorR3 linep1, perVectorR3 linep2, perVectorR3 &Intersect)
{
  // Compute whether the line touches the Mesh tile,
  //  and return YES or NO, and set Intersect to the point of contact on mesh
  float x,y,z,h;

  // first check to see if the cntact point is greater than the lowest height on surface, if not,
  // then there is def. no contact and no need to do more sophisticated checking.

  //  printf("In mesh linecontact.\n");

  // for mesh contact, we need only check to see if the current endpoint (linep1) is below surface
  //  (and we ignore linep2 which is the last location of the endpoint).

  // details of endpoint
  x=linep1(0);
  y=linep1(1);
  z=linep1(2);

  if (z<lowest)
    {
      // can't contact since it is higher than the highest point on surface
      //      printf("Mesh GUARUNTEE.\n");
      return NO;
    }

  // there might be contact, so get height at that point
  h=heightat(x,y);

  //  printf ("Distance = %g\n",z-h);

  if (z>h)
    {
      // the endpoint is below the surface (ie z of endpoint is large)
      //      printf("Mesh contact at height %g (surface is %g).\n",z,h);
      Intersect(0)=x;
      Intersect(1)=y;
      Intersect(2)=z;
      return YES;
    }

  //  printf("NO Mesh contact.\n");
  return NO;
}




float Mesh::heightat(float x,float y)
{
  // calculate the height of the surface at the specified x & y coordinates
  // calculations based on McMillan's superior Dynamechs contact routines
  int xi,yi;
  float xof,yof;
  float height;

  // get the cell of the mesh the point is in, and how far into that mesh
  gridcoords(x,y,xi,yi,xof,yof);

  if (xof>yof)
    {
      // we are interested in upper right half of mesh triangle pair
      height=heights[index(xi+1,yi)];
      height-=(1.0-xof)*(heights[index(xi+1,yi)]-heights[index(xi,yi)]);
      height+=yof*(heights[index(xi+1,yi+1)]-heights[index(xi+1,yi)]);
    }
  else
    {
      // we are interested in lower left half of mesh triangle pair
      height=heights[index(xi,yi+1)];
      height+=xof*(heights[index(xi+1,yi+1)]-heights[index(xi,yi+1)]);
      height-=(1.0-yof)*(heights[index(xi,yi+1)]-heights[index(xi,yi)]);
    }


  //  printf("height evaluation at %d,%d (%g,%g).\n",xi,yi,xof,yof);

  //  printf("Point at (%g,%g) mapped to hieght %g  on  grid (%d,%d)  with offsets (%g,%g)\n",x,y,height,xi,yi,xof,yof);
  return height;
}




void Mesh::normalframe(perVectorR3 contactloc,perMatrix &orientation)
{
  // calculate the normal at a given x,y point
  // calculations based on McMillan's superior Dynamechs contact routines
  int xi,yi;
  float xof,yof;
  float x,y;
  float norm;
  perVectorR3 v1;
  perVectorR3 v2;
  perVectorR3 normalv;
  perVectorR3 i1;

  // get the cell of the mesh the point is in, and how far into that mesh
  x=contactloc(0);
  y=contactloc(1);
  gridcoords(x,y,xi,yi,xof,yof);

  //  printf("normal evaluatin at %d,%d (%g,%g).\n",xi,yi,xof,yof);

  if (xof>yof)
    {
      // we are interested in upper right half of mesh triangle pair
      v1(0)=xdist;
      v1(1)=0.0;
      v1(2)=-1*(heights[index(xi,yi)]-heights[index(xi+1,yi)]);
      v2(0)=0.0;
      v2(1)=ydist;
      v2(2)=heights[index(xi+1,yi+1)]-heights[index(xi+1,yi)];

      normalv=JessCross(v1,v2);
      i1=JessCross(normalv,v1);
    }
  else
    {
      // we are interested in lower left half of mesh triangle pair
      v1(0)=0;
      v1(1)=-ydist;
      v1(2)=heights[index(xi,yi)]-heights[index(xi,yi+1)];
      v2(0)=xdist;
      v2(1)=0.0;
      v2(2)=heights[index(xi+1,yi+1)]-heights[index(xi,yi+1)];
      normalv=JessCross(v1,v2);
      i1=JessCross(normalv,v1);
    }


  // if we just were getting a normal 3vector, then all we would do is:
  // return normalv.
  //  // now normalize the force normal vector
  //  norm=sqrt(normalv(0)*normalv(0)+normalv(1)*normalv(1)+normalv(2)*normalv(2));
  //  normalv(0)=normalv(0)/norm;
  //  normalv(1)=normalv(1)/norm;
  //  normalv(2)=normalv(2)/norm;

  // BUT we really want a full orientation vector (why?)
  perVectorR3 j1;
  perVectorR3 i0(1,0,0);
  perVectorR3 j0(0,1,0);
  perVectorR3 k0(0,0,1);

  // ka (z) axis is just the perpendicular N
  // ia and ja axes (x and y) are just any *mutually* perpendicular other axis
  j1=JessCross(normalv,i1);
  // Now we have i1,j1,k1 all perpendicular axes, but to get a basis set
  //  we must normalize them.
  i1.normalize_self();
  j1.normalize_self();
  normalv.normalize_self();
  // We want the rotation matrix from WORLD to OBJECT (from frame 0 to 1)
  // Taken directly from Spong (1989) p.34
  orientation=perMatrix(3,3);
  orientation(0,0)=i0.dot(i1);
  orientation(1,0)=i0.dot(j1);
  orientation(2,0)=i0.dot(normalv);
  orientation(0,1)=j0.dot(i1);
  orientation(1,1)=j0.dot(j1);
  orientation(2,1)=j0.dot(normalv);
  orientation(0,2)=k0.dot(i1);
  orientation(1,2)=k0.dot(j1);
  orientation(2,2)=k0.dot(normalv);
}



void Mesh::gridcoords(float x, float y, int &xi, int &yi, float &xof, float &yof)
{
  // calculate which cell of the mesh the point is in, and how far into that mesh
  float xif,yif;

  if (x<sx)
    {
      // they are off the left hand side of surface, so use leftmost value
      xi=0;
      xof=0;
    }
  else if (x>=ex)
    {
      // they are off the left hand side of surface, so use leftmost value
      xi=gridsx-2;
      xof=1.0;
    }
  else
    {
      xif=((x-sx)/(ex-sx)) * (gridsx-1);
      xi=(int)xif;
      xof=xif-xi;
    }

  if (y<sy)
    {
      // they are off the top of surface, so use topmost value
      yi=0;
      yof=0.0;
    }
  else if (y>=ey)
    {
      // they are off the bottom of surface, so use lowest value
      yi=gridsy-2;
      yof=1.0;
    }
  else
    {
      yif=((y-sy)/(ey-sy)) * (gridsy-1);
      yi=(int)yif;
      yof=yif-yi;
    }
}


  
void Mesh::makegrob()
{
  // make the real grob from the mesh
  //  if (grobp!=NULL)
  //    delete grobp;
  //  printf("Calling meshlist with (%d,%d,%g,%g).\n",gridsx,gridsy,xdist,ydist);
  grobp->makeMeshList(gridsx,gridsy,xdist,ydist,heights);
  grobp->setdetail(meddetail);
  grobp->setlocation(sx,sy,0);
}


void Mesh::makedefaultgrob()
{
  // make a temporary grob
  grobp->makeCubeList(30,30,.2);
  grobp->setdetail(meddetail);
}
