//  Jesse Reichler, January 1996
//  Fred Delcomyn
//
//  world.C
//  This file contains the class methods for the World class.
//   see description in header (.h) file.
//

#include "misc.def"
#include "world.h"
#include "physobject.h"
#include "robot.h"
#include "contact.h"

#include <stdlib.h>
#include <time.h>
#include <string.h>



// test
bool NODYNAMICS=false;




World::World(const char *worldname)
{
  // Construct the world specified
  
  // initialize
  strcpy(name,worldname);
  things=0;
  contacts=0;
  ground=(Ground *)NULL;
  
  // Start the runtime clock, and start world PAUSED
  running=false;
  stepping=false;
  worldtime=0.0;
  runtime=0.0;
  starttime=time(NULL);
  needsrecamera=NO;
  movieloading=false;

  // test defaults
  worldgravity=9.81;
  deltat=0.01;
}



void World::done_construct()
{
  // we have added all objects

  // make all contacts
  makecontacts();
}



































World::~World()
{
  // Deconstruct the world by deconstructing the attached objects
  // This is a little tricky.  There are objects attached to the world
  //  that we are not responsible for deleting (such as links).
  // BUT, after we delete some others (such as robots), the things we
  //  aren't supposed to delete are not available for querying whether
  //  we should delete them!!
  // We handle this by FIRST removing from our attached list any things
  //  that we should not delete, then delete the rest.
  int count;
  
  for (count=0;count<things;++count)
    {
      if (thing[count]->worlddelete())
	;
      else
	thing[count]=NULL;  // mark it not to be deleted
    }
  
  // delete the ground that is attached
  if (ground != (Ground *)NULL)
    delete ground;
  
  // Now cfpely delete the things that are attached, ignoring
  //  those that have been previously nulled out as 'not our responsib.'
  for (count=0;count<things;++count)
    {
      if (thing[count]==NULL)
	continue;
      delete thing[count];
      thing[count]=NULL;
    }
  
  // delete the contacts that are attached
  for (count=0;count<contacts;++count)
    delete contact[count];
}




bool World::step()
{
  // Iterate the world for delta time seconds
  int count;
  
  if (!running)
    {
      // return false to signify we haven't done anything
      return false;
    }

  if (movieloading)
    {
      // Tell any videocamera listeners that the world has updated
      // 2/4/99 trying without updatefordraw
      //      updatefordraw();
      cameranotify();
      return true;
    }

  static int pcount=0;
  //  if (++pcount>=1000)
  //    {
  //      cout << "In world step time=" << worldtime << ".\n";
  //      pcount=0;
  //    }

  if (NODYNAMICS)
    {
      // TEST - no dyamics
      updatefordraw();
      cameranotify();
      worldtime+=deltat;
      return true;
    }


  // Note that we must first completely iterate and then
  //  completely update, in that order.
  // The reason for this is that we want to simulate simultaneous
  //  parallel iteration, so we don't update positions until
  //  all are done iterating.  This is standard practice.
  
  //  cout << "In world step: checking contacts.\n";

  // now compute any contact forces
  checkcontacts(deltat);
  
  //    cout << "In world step: iterating objects.\n";

  // next, iterate each object in the world (calc. accelerations)
  for (count=0;count<things;++count)
    {
      if (thing[count]->worlditerate())
	thing[count]->iterate(deltat);
    }
  
  //  cout << "In world step: updating things.\n";

  // now update each object's position/velocity, given new accel
  for (count=0;count<things;++count)
    {
      if (thing[count]->worldupdate())
	thing[count]->update(deltat);
    }
  
  // Update the world clock
  worldtime+=deltat;

  //  cout << "In world step: updating for draw.\n";

  // Tell any videocamera listeners that the world has updated
  cameranotify();

  // return saying that we have updated
  return true;
}




void World::updatefordraw()
{
  // reprepare all objects for drawing
  int count;
  
  // now update each object's position/velocity, given new accel
  for (count=0;count<things;++count)
    {
      if (thing[count]->worldupdate())
	thing[count]->updatefordraw();
    }
}




void World::grobvidcam_draw(GrobVideoCamera *cam)
{
  // now draw each object to the specified camera
  int count;
  
  for (count=0;count<things;++count)
    {
      if (thing[count]->worlddraw())
	thing[count]->draw(cam);
    }
}



void World::addthing(Thing *athing)
{
  // Attach the object 'it' to the world.  This means that the
  //  object is now considered to be part of the world, and will
  //  be iterated, drawn, checked for contact with, etc...
  
  // First, make sure we can add it
  if (things>=MAX_THINGS_PERWORLD)
    {
      printf("$$$ERROR: too many things (%d) already attached to world '%s'!\n",things,name);
      return;
    }
  
  // Now add it
  thing[things]=athing;
  ++things;

}


void World::makeobjcontacts(Physicalobject *obj1)
{
  // add all contacts between this object and the rest of the world objects
  Physicalobject *obj2;
  int shapetype1,shapetype2;
  int count;

  shapetype1=obj1->shapetype();
  if (shapetype1!=SHAPE_LEGCONE && shapetype1==SHAPE_SPHERE)
    {
      // currently we add only (legcone|sphere)->triangletile contacts
      return;
    }

  //  for (count=0;count<things-1;++count)
  for (count=0;count<things;++count)
    {
      shapetype2=thing[count]->shapetype();
      if (shapetype2==SHAPE_NOSHAPE)
	continue;
      
      obj2=(Physicalobject *)thing[count];
      if (obj2->thingtype()==THING_TRITILE)
	{
	  // add only legcone,triangletile contacts
	  obj1->addcontacts(obj2);
	}
      else if (obj2->thingtype()==THING_GROUNDMESH)
	{
	  // and legcone, mesh contacts
	  obj1->addcontacts(obj2);
	}
    }
}



void World::makecontacts()
{
  // this should be called after all objects are ceated with shapes
  Physicalobject *obj1;
      Thing *athing;
  int shapetype1;
  int count;
  
  //  for (count=0;count<things-1;++count)
  for (count=0;count<things;++count)
    {
      athing=thing[count];
      shapetype1=athing->shapetype();   // shape of thing being added
      if (shapetype1!=SHAPE_NOSHAPE)
	{
	  obj1=(Physicalobject *)athing;
	  makeobjcontacts(obj1);
	}
    }
}




Thing* World::lookupthing(char *thingname)
{
  // Find an object attached to the world with the name specified
  //  and retunr a pointer to it
  for (int count=0;count<things;++count)
    {
      if (strcmp(thing[count]->name,thingname)==0)
	return(thing[count]);
    }
  return (Thing *)NULL;
}


Physicalobject* World::lookupphysobject(char *objname)
{
  // Find an object attached to the world with the name specified
  //  make sure it is a physobject, then return a pointer to it
  for (int count=0;count<things;++count)
    {
      if (strcmp(thing[count]->name,objname)==0)
	{
	  if (thing[count]->isphys()==YES)
	    return((Physicalobject *)thing[count]);
	  else if (thing[count]->thingtype()==THING_ROBOT)
	    {
	      // user wants to focus on a physical thing, so
	      //  return pointer to robot base
	      return ((Physicalobject *)((Robot *)thing[count])->base);
	    }
	  
	}
    }
  return (Physicalobject *)NULL;
}




void World::addcontact(Contact *icontact)
{
  // Attach the contact instance to the world.
  if (contacts>=MAX_CONTACTS_PERWORLD)
    {
      printf("$$$ERROR: too many contacts (%d) already attached to world '%s'!\n",contacts,name);
      return;
    }
  contact[contacts]=icontact;
  ++contacts;
}



void World::clearforces()
{
  // Clear the external forces on all attached physical objects
  Physicalobject *objp;
  
  for (int count=0;count<things;++count)
    {
      // clear any other forces for the object
      
      if (thing[count]->shapetype()!=SHAPE_NOSHAPE)
	{
	  objp=(Physicalobject *)thing[count];
	  objp->clearextforce();
	  objp->clearotherforces();
	}
    }
}



void World::checkcontacts(double deltat)
{
  // Check all the possible contacts, which fill in external forces
  int count;
  
  for (count=0;count<contacts;++count)
    contact[count]->update(deltat);
}




void World::setpause(percision dummyv)
{
  // we have received a pause signal from the user, so toggle pause
  // Toggle the pause conidition of the simulation
  //  and update the real runtime counter as well, which doesn't
  //  want to count time that the simulation is paused.
  
  if (!running)
    {
      // On unpausing, we start the elapsed runtime counter going
      running=true;
      starttime=time(NULL);
    }
  else
    {
      // Since we are now pausing, update our account of runtime
      //  note this has to be done prior to setting ispaused,
      //  as per fixupruntime() documentation.
      fixupruntime();
      running=false;
    }
}




void World::fixupruntime()
{
  // If the simulation is running at the current time,
  //  to get an accurate account of time running we must add the
  //  elapsed time since the last update.  Especially useful when
  //  the program ends.
  
  if (running)
    {
      endtime=time(NULL);
      runtime+=difftime(endtime,starttime);
    }
  else
    {
      // If the simulation is already paused, no update is nesc.
      //  as this was done at time of pausing.
    }
}



int World::movie_load(FILE *movie_playfp)
{
  // load and play a movie frame from file movie_playfp
  // if we get to the end of the file, just restart from the begining of
  //  the file.
  percision oldtime;
  int count;
  int retv;
  float tempf;
  int temp;

  // save oldtime
  oldtime=runtime;

  // load time
  retv=fread(&worldtime,sizeof(worldtime),1,movie_playfp);

  if (feof(movie_playfp) || retv==0)
    {
      // 2/29/99
      return 0;

      // we hit the end of the file, so automatically start over
      fseek(movie_playfp,0L,SEEK_SET);
      // read over sim and version version
      fread(&tempf,sizeof(tempf),1,movie_playfp);
      fread(&temp,sizeof(temp),1,movie_playfp);
      
      // now reset
      worldtime=oldtime;
      oldtime=0.0;
      return (0);
    }
  
  fread(&worldgravity,sizeof(worldgravity),1,movie_playfp);
  
  // now load each object from the file
  for (count=0;count<things;++count)
    {
      if (thing[count]->thingtype()==THING_ROBOT)
	{
	  thing[count]->movie_load(movie_playfp);
	}
    }

  // calculate deltat before leaving
  deltat=worldtime-oldtime;

  return (1);
}



void World::movie_save(FILE *movie_recordfp)
{
  // save a movie frame to file movie_recordfp
  int count;

  // save time
  fwrite(&worldtime,sizeof(worldtime),1,movie_recordfp);
  fwrite(&worldgravity,sizeof(worldgravity),1,movie_recordfp);
  
  // now save all objects
  for (count=0;count<things;++count)
    {
      if (thing[count]->thingtype()==THING_ROBOT)
	thing[count]->movie_save(movie_recordfp);
    }
}

