//  datafactory.h
//  This file contains the class definition for the datafactory classes, which are top level supervisors
//   of reading and writing data to and from accessors (ie saving data or routing input data to targets).
//

// forward declarations
class DatStore;
class Datafactory;
class DatafactoryIn;
class DatafactoryOut;
class StreamFilter;

#ifndef datafactoryh_defined
#define datafactoryh_defined

#include "cfpaccessor.h"
#include "datastream.h"

//#include <time.h>
//#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <vector>
#include <fstream>
#include <iostream>





// ---------------------------------------------------------------------------------------
// -- Typedefs
// ---------------------------------------------------------------------------------------

//: SampleInType enumeration, for modes of sampling from input files
typedef enum SampleOutType
{
  current_sampomode,
  average_sampomode,
  sum_sampomode
};

//: Data Interpolation modes for *reading* data files
typedef enum SampleInType
{
  step_sampimode,
  punctuated_sampimode,
  interpolate_sampimode
};


//: FilterReturnType for methods for file filters
typedef enum FilterReturnType
{
  success_retval,
  error_retval,
  end_retval
};






// ---------------------------------------------------------------------------------------
// -- DatStore
// ---------------------------------------------------------------------------------------
//: The virtual base class for DatInStore and DatOutStore, classes used for
//  managing input and output variables.

class DatStore
{
protected:
  CfpVAccessor *vac;
  int index;
  bool skip;
  Datafactory *dfactp;
public:
  DatStore(CfpVAccessor *invac,Datafactory *indfactp,int inindex=-1) {vac=invac;skip=false;dfactp=indfactp;index=inindex;};
public:
  CfpVAccessor *getvacp(){return vac;};
  CfpVal getval(){return vac->get(index);};
public:
  void setskip(bool inskip=true){skip=inskip;};
  bool getskip(){return skip;};
};







// ---------------------------------------------------------------------------------------
// -- Datafactory
// ---------------------------------------------------------------------------------------
//:The base virtual class for DatafactoryOut and DatafactoryIn, classes used for reading
// and writing data sources (files)

class Datafactory
{
  friend class DatStore;
  friend class StreamFilter;
protected:
  Configpp *configppp;
  Datastream *streamp;
  StreamFilter *filterp;
  vector<DatStore *> datlist;
protected:
  int dtcounter;
  int samprate;
  bool done;
public:
  Datafactory(Configpp *cfpp, int insamprate);
  ~Datafactory();
public:
  virtual string gettype() {return "virtual";};
  int vacs() {return datlist.size();};
  CfpVAccessor *vac(int index) {return datlist[index]->getvacp();};
  CfpVal val(int index) {return vac(index)->get();};
public:
  void setstreamp(Datastream *instreamp) {streamp=instreamp;};
  void setfilterp(StreamFilter *infilterp) {filterp=infilterp;};
  int getsamprate() {return samprate;};
  long getrowcount(); //{return filterp->getrowcount();};
public:
  virtual void start();
  virtual void finish();
  void adddat(DatStore *dstorep) {datlist.push_back(dstorep);};
  void error(string errstring,string hinstr="");
  virtual void update() {;};
};






// ---------------------------------------------------------------------------------------
// -- StreamFilter
// ---------------------------------------------------------------------------------------
//: The base virtual class used handle the low-level formating of outgoing and incoming
//  data.  It is a helper class for DataOutFile and DataInFile.   It is subclassed to handle
//  Matlab, ascii, etc.

class StreamFilter
{
  friend class Datafactory;
protected:
  Configpp *configppp;
  Datafactory *dfactp;
  Datastream *streamp;
  bool usecomments;
protected:
  long rowcount;
public:
  virtual void start();
  void finish() {;};
public:
  StreamFilter(Configpp *cfpp);
public:
  virtual FilterReturnType writeheader() {return error_retval;};
  virtual FilterReturnType writetrailer() {return error_retval;};
  virtual FilterReturnType writerow() {return error_retval;};
  virtual FilterReturnType readheader() {return error_retval;};
  virtual FilterReturnType readtrailer() {return error_retval;};
  virtual FilterReturnType readrow(int readindex) {return error_retval;};
public:
  void setfactory(Datafactory *indfactp) {dfactp=indfactp;};
  void setstream(Datastream *instreamp) {streamp=instreamp;};
  void set_usecomments(bool incomments) {usecomments=incomments;};
  long getrowcount() {return rowcount;};
  void resetrowcount() {rowcount=0;};
protected:
  int vacs() {return dfactp->vacs();};
  CfpVAccessor *vac(int index) {return dfactp->vac(index);};
  CfpVal val(int index) {return dfactp->val(index);};
  void error(string errstring,string hinstr="");
};







#endif
