// cfpmovie.C

#include "cfpmovie.h"
#include <cstdio>




cfpMovie::cfpMovie(string fname)
{
  // just initialize the movie structure
  filename=fname;
  filep=NULL;
  refreshrate=1;
  filestartpos=0;
  framepos=0;
  lastframepos=0;
  nextframepos=0;
  curframe=0;
  framecount=0;
  refreshcount=0;
  cyclemode=loop_cycle;
  moviemode=invalidmovie;
  usercomment="";
}

cfpMovie::~cfpMovie()
{
  // close the movie if it is currently open
  closefile();

  // free the list of pointers to objects
}


void cfpMovie::addobj(CfpObject *sobj)
{
  // add a new object to our list of those we save/load
  obj.push_back(sobj);

  if (moviemode==playmovie)
    {
      // now if this is a movie playback, we need to inform target objects that they are being set from a movie,
      //  in case they want to do something differently (like forgoe internal updating which is pointless if they are loaded)
      sobj->setmovieplaybackflag();
    }
}

void cfpMovie::removeobj(CfpObject *sobj)
{
  // what we have in mind here is to search for the object in our list of those we save or load
  // and if we find it, remove it.  of course the tricky part is still being able to load and save dynamically changing lists...
  // we'll worry about it after the basics work.
}




void cfpMovie::update()
{
  // save or load objects to file
  if (!filep)
    {
      // error: no file is open
      if (moviemode!=errormovie)
	cerr << "Movie Error: trying to update movie to file '"<<filename<<"'.  file is not open.\n";
      movieerror();
      return;
    }

  if (moviemode==playmovie)
    update_load();
  else
    update_save();
}



void cfpMovie::update_load()
{
  // load objects from file
  bool bretval;

  if (refreshrate==0)
    return;

  // first go to the frame we want to read from (we come into this func at the frame after the last one read)
  if (!skipframes(refreshrate-1))
    return;

  //  cout << "Reading frame "<<curframe+1<<" of "<<framecount<<".\n";

  // now read the frame header
  read_frameheader();

  // now have each object load itself
  list<CfpObject *>::iterator p=obj.begin();
  while (p!=obj.end())
    {
      bretval=(*p)->movieload(filep);
      if (!bretval)
	{
	  cerr << "Movie Error: loading object '"<<(*p)->prettyname()<<"' from file '"<<filename<<"' at frame "<<curframe<<" (near file position "<<ftell(filep)<<")\n";
	  closefile();
	  return;
	}
      ++p;
    }

  // load the frame trailer
  read_frametrailer();

  // advance the frame counter
  ++curframe;
}


void cfpMovie::update_save()
{
  // write each object on our list to the file
  bool bretval;

  if (++refreshcount<refreshrate)
    return;
  else
    refreshcount=0;

  // first write the fram header
  write_frameheader();

  // now have each object save itself
  list<CfpObject *>::iterator p=obj.begin();
  while (p!=obj.end())
    {
      bretval=(*p)->moviesave(filep);
      if (!bretval)
	{
	  cerr << "Movie Error: saving object '"<<(*p)->prettyname()<<"' from file '"<<filename<<"' at frame "<<framepos<<".\n";
	  movieerror();
	  return;
	}
      ++p;
    }

  // now write the fram header
  write_frametrailer();

  // increment our counters
  ++curframe;
  ++framecount;
}




bool cfpMovie::skipframes(long frameskip)
{
  // skip frames (maybe 0) AND check to see if we have reached the end (or gone backward past begining)
  // return true on success, false to stop
  long targetframe=curframe+frameskip;

  if (targetframe>=framecount || targetframe<0)
    {
      // we have gone beyond the boundary
      switch (cyclemode)
	{
	case end_cycle:
	  // just stop here at the end
	  refreshrate=0;
	  return false;
	  break;
	case loop_cycle:
	  if (refreshrate>0)
	    {
	      // we were goping forward, so just go back to begining (we can go directly there)
	      fseek(filep,firstframepos,SEEK_SET);
	      curframe=0;
	      return true;
	    }
	  else
	    {
	      // we need to seek to go to the last frame
	      targetframe=framecount-1;
	      frameskip=targetframe-curframe;
	    }
	  break;
	case reverse_cycle:
	  // start going in reverse
	  refreshrate*=-1;
	  targetframe=curframe+(refreshrate-1);
	  if (targetframe>=framecount || targetframe<0)
	    {
	      refreshrate=0;
	      closefile();
	      return false;
	    }
	  frameskip=targetframe-curframe;
	  break;
	}
    }
  else if (frameskip==0)
    return true;

  if (frameskip>0)
    {
      while (curframe<targetframe && moviemode!=errormovie)
	{
	  // jump to the next frame
	  // read the header so we can get pointer to next
	  read_frameheader();
	  fseek(filep,nextframepos,SEEK_SET);
	  ++curframe;
	}
    }
  else if (frameskip<0)
    {
      fseek(filep,framepos,SEEK_SET);
      --curframe;
      while (curframe>targetframe && moviemode!=errormovie)
	{
	  // jump to the next frame
	  // read the header so we can get pointer to next
	  read_frameheader();
	  fseek(filep,lastframepos,SEEK_SET);
	  --curframe;
	}
    }

  if (moviemode==errormovie)
    return false;
  return true;
}





bool cfpMovie::loadopen()
{
  // open the file for loading, return true on success
  moviemode=playmovie;

  if ((filep=fopen(filename.c_str(),"rb"))==(FILE *)NULL)
    {
      // We failed to open the file
      cerr << "Movie Error: failed to open movie input file '"<<filename<<"'.\n";
      return false;
    }

  // read the initial movie header
  read_movieheader();

  // we need to inform any added objects that they are being set from a movie,
  //  in case they want to do something differently (like forgoe internal updating which is pointless if they are loaded)
  list<CfpObject *>::iterator p=obj.begin();
  while (p!=obj.end())
    {
      (*p)->setmovieplaybackflag();
      ++p;
    }
}

bool cfpMovie::saveopen()
{
  // open the file for saving, return true on success
  moviemode=recordmovie;

  if ((filep=fopen(filename.c_str(),"wb"))==(FILE *)NULL)
    {
      // We failed to open the file
      cerr << "Movie Error: failed to open movie output file '"<<filename<<"'.\n";
      return false;
    }

  // write the initial movie header
  write_movieheader();
}



void cfpMovie::movieerror()
{
  // just close the movie file and change mode to indicate an error ocurred
  closefile();
  moviemode=errormovie;
}

void cfpMovie::closefile()
{
  // just close the movie file if it is open
  if (filep)
    {
      if (moviemode==recordmovie)
	{
	  // write the movie trailer
	  write_movietrailer();
	}
      fclose(filep);
      filep=NULL;
      refreshrate=0;
      moviemode=invalidmovie;
    }
}






void cfpMovie::write_movieheader()
{
  // write the info that goes at the start of every movie

  // first the version and date
  starttime=time(NULL);
  fprintf(filep,"Configpp version %s using Movie File format version %s.\n",SAF_VERSIONSTRING,writeversionstring().c_str());
  fprintf(filep,"File '%s' created %s",filename.c_str(),ctime(&starttime));
  fprintf(filep,"NOTE: this is a binary file whose contents may be unreadable in a text editor.\n");

  // now any user comments
  if (usercomment!="")
    {
      fputs("\n",filep);
      fputs(usercomment.c_str(),filep);
      fputs("\n",filep);
    }

  // now show what objects are being saved
  if (obj.size()>0)
    {
      fputs("\n",filep);
      fprintf(filep,"Recording state of the following objects (initial refreshrate=%d):\n",refreshrate);
      list<CfpObject *>::iterator p=obj.begin();
      while (p!=obj.end())
	{
	  fprintf(filep,"  %s\n",((*p)->prettyname()).c_str());
	  ++p;
	}
      fputs("\n",filep);
    }

  // now the signal that the movie data comes next
  fputs("\n------------------------------------------------------------------\n",filep);
  fputs("BEGIN MOVIEDATA\n",filep);
  fputs("------------------------------------------------------------------\n",filep);

  // record the start of data
  filestartpos=ftell(filep);

  // write a dummy framecount which will be replaced when movie is completed
  fwrite(&framecount,sizeof(framecount),1,filep);

  // record the location of the first frame (for fast restarting)
  firstframepos=ftell(filep);
}


void cfpMovie::read_movieheader()
{
  // read the info that goes at the start of every movie
  char dummytext[255];

  // the first line should be a version string
  fgets(readversionstring,254,filep);
  // the rest we don't know about, maybe user comments, maybe stuff from a new version and we are stupid.
  // we read '\n' terminated lines until we see the keywords "BEGIN MOVIEDATA".
  for (;;)
    {
      if (feof(filep))
	  {
	    closefile();
	    return;
	  }
      fgets(dummytext,254,filep);
      if (strncmp(dummytext,"BEGIN MOVIEDATA",15)==0)
	{
	  // consume the extra separator line and then break
	  fgets(dummytext,254,filep);
	  break;
	}
      else
	usercomment+=string(dummytext);
    }

  // read the total number of frames framecount
  fread(&framecount,sizeof(framecount),1,filep);

  // record the location of the first frame (for fast restarting)
  firstframepos=ftell(filep);
}


void cfpMovie::write_movietrailer()
{
  // write the info that goes at the end of every movie

  // now write some nice stuff at the end
  time_t endtime;
  double dtime;
  endtime=time(NULL);
  dtime=difftime(endtime,starttime);
  fputs("\n\n------------------------------------------------------------------\n",filep);
  fputs("END MOVIEDATA\n",filep);
  fputs("------------------------------------------------------------------\n",filep);
  fprintf(filep,"Frames Recorded:          %ld.\n",framecount);
  fprintf(filep,"Actual Recording Time:    %.2f minutes.\n",dtime/60.0);


  // lastly, we need to dart back to the begining of the file and write the total number of frames in file
  fseek(filep,filestartpos,SEEK_SET);
  fwrite(&framecount,sizeof(framecount),1,filep);
}

void cfpMovie::read_movietrailer()
{
  // read the info that goes at the end of every movie
}





void cfpMovie::write_frameheader()
{
  // write the info that goes before every fame

  // we store the file pos we are at for later updating
  framepos=ftell(filep);

  // first we write a string that we can check for during reading to make sure we stay in sync and don't get lost
  fwrite("---FRAME---",11,1,filep);
 
  // write the pos of the previous frame
  // and write a dummy value for the location of next frame, which we will correct in trailer
  fwrite(&lastframepos,sizeof(lastframepos),1,filep);
  fwrite(&nextframepos,sizeof(nextframepos),1,filep);
}

void cfpMovie::read_frameheader()
{
  // read the info that goes before every frame
  char dummytext[25];

  // we store the file pos we are at in case we want to go backwards
  framepos=ftell(filep);

  fread(dummytext,11,1,filep);
  if (strncmp(dummytext,"---FRAME---",11)!=0)
    {
      // error, we have gotten out of sync with frames
      cerr << "Movie Error: out of sync with file data while reading file '"<<filename<<"' at frame "<<framepos<<" (near file position "<<ftell(filep)<<")\n";
      movieerror();
    }

  // now read the position of the previous and next frame in case we want to skip somewhere immediately
  fread(&lastframepos,sizeof(lastframepos),1,filep);
  fread(&nextframepos,sizeof(nextframepos),1,filep);
  //  cout << "read header.  lastpos="<<lastframepos<<"   nextpos="<<nextframepos<<".\n";
}


void cfpMovie::write_frametrailer()
{
  // write the info that goes after every frame
  long curpos;

  // lastly, we want to dart back and fill in the position for the NEXT frame (so it can be easily skipped too if desired)
  curpos=ftell(filep);
  fseek(filep,framepos+11+sizeof(lastframepos),SEEK_SET);
  fwrite(&curpos,sizeof(curpos),1,filep);
  fseek(filep,curpos,SEEK_SET);

  // and save the startposition of the frame we just wrote as the 'lastframepos' for the next frame
  lastframepos=framepos;
}


void cfpMovie::read_frametrailer()
{
  // read the info that goes after every frame
}






