// plot.C

#include "plot.h"

#include <cmath>
#include <iostream>



//-------------------------------------------------------------------------------
//-- Datapoint
//-------------------------------------------------------------------------------

void Datapoint::allocate(char indims)
{
  // allocate space for
  data=new float[indims];
  dims=indims;
  psourcenum=2;
}

Datapoint::~Datapoint()
{
  // set values
  if (data)
    delete [] data;
}

void Datapoint::set(float *indata)
{
  // set values (dimensions better be equal)
  for(int count=0;count<dims;++count)
    data[count]=indata[count];
}


void Datapoint::print()
{
  // print values (for debugging)
  for(int count=0;count<dims;++count)
    cout << data[count] << " ";
  cout << "\n";
}


void Datapoint::setloc(float *loc,int x, int y, int z)
{
  // set a 3vec location based on indexes x,y,z

  if (x!=-1)
    loc[0]=data[x];
  if (y!=-1)
    loc[1]=data[y];
  if (z!=-1)
    loc[2]=data[z];
}






//-------------------------------------------------------------------------------
//-- PointSource
//-------------------------------------------------------------------------------

PointSource::PointSource(string inlabel)
{
  label=inlabel;
}



//-------------------------------------------------------------------------------
//-- Plot
//-------------------------------------------------------------------------------

Plot::Plot()
  :GrobVideoCameraDrawer()
{
  dims=0;
  needsadjust=false;
  needsredraw=true;
  newpointer=points.end();
  lastpoint=NULL;
  newpoints=0;
  refreshcount=0;
  storecount=0;
  storerate=10;
  refreshrate=1;

  // test
  storerate=12;
  refreshrate=2;

  savepoints=100;
  //  savepoints=10;
  seenpoints=0;
  symbols=0;
  storageaxis=xaxis=yaxis=zaxis=-1;
  supportsinc=false;
  connect=true;
  colorscale=colorless_scale;

  makesymbols();
  make_colormaps();
}


Plot::~Plot()
{
  // clear vectors and lists
  int count;

  axes.clear();
  psources.clear();
  clearpoints();

  // clear grob symbols
  for (count=0;count<symbols;++count)
    {
      if (grobsymbol[count])
	delete grobsymbol[count];
    }
}


void Plot::clearpoints()
{
  // remove and delete all stored points
  Datapoint *dp;
  while (points.size()>0)
    {
      // we need to get rid of one
      dp=points.front();
      points.pop_front();
      delete dp;
    }
  newpointer=points.end();
  lastpoint=NULL;
  newpoints=0;
}


void Plot::complete()
{
  // initialize min and max vals
  minpoint.allocate(dims);
  maxpoint.allocate(dims);

  // now run any local (virtual) complete
  virt_complete();

  // now register the graphing axes pointers
  if (xaxis!=-1)
    xaxisp=axes[xaxis];
  if (yaxis!=-1)
    yaxisp=axes[yaxis];
  if (zaxis!=-1)
    zaxisp=axes[zaxis];
}



void Plot::forceredraw()
{
  // make the camera redraw us
  needsredraw=true;
  cameranotify_axes();

  // trick to make it update the next time it is called
  refreshcount=refreshrate*storerate;
}



void Plot::addpoint(Datapoint *indatap)
{
  // add a new point and adjust range
  float v;
  Datapoint *dp;
  Datapoint *datap;

  // first see if we want to store it
  if (++storecount<storerate)
    return;
  else
    storecount=0;

  datap=new Datapoint(dims,indatap);
  //  datap->print();

  if (++seenpoints==1)
    {
      // this is the first point we've ever seen
      minpoint.copy(datap);
      maxpoint.copy(datap);
      needsadjust=true;
    }

  // add point to history
  points.push_back(datap);
  if (newpoints==0)
    {
      newpointer=points.end();
      --newpointer;
    }
  ++newpoints;

  // keep point list small
  if (points.size()>savepoints)
    {
      // we need to get rid of one
      dp=points.front();
      points.pop_front();
      if (*newpointer==dp)
	{
	  // we popped off our stored memory of the oldest nondrawn point, so we need to correct
	  --newpoints;
	  newpointer=points.begin();
	}
      if (lastpoint==dp)
	{
	  // we popped off our stored memory of the last drawn point, so we need to correct (this should rarely happen)
	  lastpoint=NULL;
	  //	  lastpoint=*newpointer;
	}
      delete dp;
    }

  // now adjust min&max values
  for(int count=0;count<dims;++count)
    {
      if (count==storageaxis)
	{
	  // this axis is used by table plots, and a points value here does not reflect the axis range
	  continue;
	}
      v=datap->valat(count);
      if (v<minpoint.valat(count))
	{
	  minpoint.setvalat(count,v);
	  needsadjust=true;
	  needsredraw=true;
	}
      if (v>maxpoint.valat(count))
	{
	  maxpoint.setvalat(count,v);
	  needsadjust=true;
	  needsredraw=true;
	}
    }

  if (!needsadjust)
    {
      // let any table plot see it, providing we have not already decided to readjust the table
      incorporate_point(indatap);
    }
}


void Plot::addaxis(VisibleAxis *axisp)
{
  // add a new axis
  axes.push_back(axisp);

  // guess implicit graphing axes (can be overridden explicitly);
  if (axisp->getname()=="x" && xaxis==-1)
    xaxis=dims;
  else if (axisp->getname()=="y" && yaxis==-1)
    yaxis=dims;
  else if (axisp->getname()=="z" && zaxis==-1)
    zaxis=dims;

  ++dims;
}

int Plot::lookupaxis(string axisname)
{
  // lookup axis number or return -1 if not found
  for (int count=0;count<dims;++count)
    {
      if (axes[count]->getname()==axisname)
	return count;
    }
  return -1;
}

void Plot::addpsource(PointSource *psourcep)
{
  // add a new pointsource
  psources.push_back(psourcep);
}


bool Plot::adjustout(bool force)
{
  // adjust all the axes, using the minimum and maximum values.  the strategy we use for adjusting axes
  //  is to, rather than adjust every axis every time we get a new datapoint, we keep track of min and max
  //  data values as datapoints are added, and then every so often (like when we redraw), we adjust the
  //  axes based on min/max value, which saves some time.
  bool waschanged=false;

  for(int count=0;count<axes.size();++count)
    {
      if (axes[count]->adjustout(minpoint.valat(count)))
	{
	  waschanged=true;
	  minpoint.setvalat(count,axes[count]->getmin());
	}
      if (axes[count]->adjustout(maxpoint.valat(count)))
	{
	  waschanged=true;
	  maxpoint.setvalat(count,axes[count]->getmax());
	}
      if (count !=storageaxis)
	{
	  minpoint.setvalat(count,axes[count]->getmin());
	  maxpoint.setvalat(count,axes[count]->getmax());
	}
    }
  
  // clear adjust flag and set redraw flag
  needsadjust=false;

  if (waschanged || force)
    {
      axeschange();
      scalesymbols();
      cameranotify_axes();
    }
}

void Plot::updatedraw()
{
  // decide how to updatet he visual image of the plot

  if (++refreshcount<refreshrate*storerate)
    return;
  else
    refreshcount=0;

  // do we need to adjust axes because a value fell out of range?
  if (needsadjust)
    adjustout();
  else if (newpoints==0)
    return;

  // do we need to redraw the whole thing, or can we just draw the new points?
  if (!needsredraw)
    {
      // try incremental redraw (this may force complete redraw if it fails)
      // tell any cameras watching us that we want an incrental draw order
      cameranotify_inc();
    }
  else
    {
      // tell any cameras watching us that we want a complete redraw order
      cameranotify();
    }
}







void Plot::makesymbols()
{
  // make some default grob symbols for the plot, all should be UNIT size (they will be stretched/scaled elsewhere)

  // a simple point
  grobsymbol[symbols]=new Grob();
  grobsymbol[symbols]->makePointList();
  symbolscale[symbols]=1;
  ++symbols;

  // a cube
  grobsymbol[symbols]=new Grob();
  grobsymbol[symbols]->makeBarList(1.0,1.0,1.0);
  symbolscale[symbols]=1;
  ++symbols;

  // a sphere
  grobsymbol[symbols]=new Grob();
  grobsymbol[symbols]->makeSphereList(0.5);
  // sneak and replace wireframe cubelist with a point
  grobsymbol[symbols]->replaceLow();
  // sneak and replace wireframe fineres sphere with highres
  grobsymbol[symbols]->replaceHigh();
  symbolscale[symbols]=1;
  ++symbols;

  // a cone
  grobsymbol[symbols]=new Grob();
  grobsymbol[symbols]->makeConeList(0.5,1);
  symbolscale[symbols]=1;
  ++symbols;
}



void Plot::scalesymbols()
{
  // scale symbols so that they display properly in the scaled axes
  float scalevec[3];
  int count;

  scalevec[0]=scalevec[1]=scalevec[2]=1.0;

  // get scales from axes
  if (xaxis!=-1)
    {
      if (!axes[xaxis]->getstretch())
	scalevec[0]=axes[xaxis]->symbolsize()/axes[xaxis]->getscale();
    }

  if (yaxis!=-1)
    {
      if (!axes[yaxis]->getstretch())
	scalevec[1]=axes[yaxis]->symbolsize()/axes[yaxis]->getscale();
    }
  if (zaxis!=-1)
    {
      if (!axes[zaxis]->getstretch())
	scalevec[2]=axes[zaxis]->symbolsize()/axes[zaxis]->getscale();
    }

  // now scale the unit-sized grobs
  for (count=0;count<symbols;++count)
    grobsymbol[count]->setscale(scalevec);
}




void Plot::connectpoints(Datapoint *sdp,Datapoint *edp)
{
  // darw a line between the two points
  float locs[3],loce[3];
  locs[0]=locs[1]=locs[2]=0.0;
  loce[0]=loce[1]=loce[2]=0.0;

  // get locations of points
  sdp->setloc(locs,xaxis,yaxis,zaxis);
  edp->setloc(loce,xaxis,yaxis,zaxis);

  // now ask static grob to draw a line
  Grob::drawline(locs[0],locs[1],locs[2],loce[0],loce[1],loce[2]);
}









void Plot::make_colormaps()
{
  // create some color ramps for plot colors
  int count;

  // first null them all out so that we don't try to free bad memory
  for (count=0;count<GROBCOLORSCALES;++count)
    rgbmap[count]=NULL;

  // okay now create the ones we want
  rgbmap[(int)hot_scale]=GrobGlHelper::rgbramp_hot();
  rgbmap[(int)cool_scale]=GrobGlHelper::rgbramp_cool();
  rgbmap[(int)jet_scale]=GrobGlHelper::rgbramp_jet();
  rgbmap[(int)hsv_scale]=GrobGlHelper::rgbramp_hsv();
  rgbmap[(int)copper_scale]=GrobGlHelper::rgbramp_copper();
  rgbmap[(int)bone_scale]=GrobGlHelper::rgbramp_bone();
  rgbmap[(int)pink_scale]=GrobGlHelper::rgbramp_pink();
  rgbmap[(int)gray_scale]=GrobGlHelper::rgbramp_gray();
}



void Plot::free_colormaps()
{
  // free up any memory used by color maps
  int count;

  for (count=0;count<GROBCOLORSCALES;++count)
    {
      if (rgbmap[count])
	{
	  delete rgbmap[count];
	  rgbmap[count]=NULL;
	}
    }
}


GrobSpacerType Plot::wantspacer()
{
  // some plots, like barcharts, need spacers around them because they use grob symbols which are centered at drawing loc.
  //  while others, like surface plots.
  switch(displaymode)
    {
    case scatter_display:
      return box_spacer;
    case bars_display:
      return none_spacer;
      return xy_spacer;
    case surface_display:
      return none_spacer;
      return xy_spacer;
      return none_spacer;
    case contour_display:
      return none_spacer;
    case flatgrid_display:
      return none_spacer;
    }
  return box_spacer;
}
