// plot.h
// This file defines the Plot classes


// forward delcarations
class Datapoint;
class PointSource;
class Plot;
class PointsPlot;
class PlotTable;
class TablePlot;


#ifndef ploth_defined
#define ploth_defined

#include "cfpaxis.h"
#include "JRinterpolater.h"

#include <list>
#include <vector>
#include <string>

// Camera stuff for drawing
#include "grob.h"
#include "grobcamera.h"

#include "grobinline.h"
#include "grobcamerainline.h"


//-------------------------------------------------------------------------------
//-- Enums for plotting
//-------------------------------------------------------------------------------


// unnesc. large number of max dimensions for any datapoint
#define MAX_TABLE_DIMENSIONS 20

typedef enum PlotSampleMode
{
  current_mode,
  average_mode,
  sum_mode,
  count_mode
};

typedef enum PlotDisplayMode
{
  scatter_display,
  bars_display,
  surface_display,
  contour_display,
  flatgrid_display
};


typedef enum PlotConnectorType
{
  no_connect,
  x_connect,
  y_connect,
  xy_connect
};







//-------------------------------------------------------------------------------
//-- Datapoint
//-------------------------------------------------------------------------------

class Datapoint
{
  friend class Plot;
private:
  char dims;										// number of dims
  char psourcenum;									// pointsourcenum in plot
  float weight;										// weight of datapoint
  float *data;										// vector value (size=dims)
public:
  Datapoint() {dims=0;data=NULL;};							// construct w/o allocat
  Datapoint(char indims) {allocate(indims);};						// allocate space
  Datapoint(char indims,float *indata) {allocate(indims); set(indata);};		// allocate and assign
  Datapoint(char indims,Datapoint *indatap) {allocate(indims); copy(indatap);};		// allocate and copy
  ~Datapoint();										// free space
public:
  void allocate(char indims);
public:
  float *datap(int index) {return &data[index];};					// pointer to component
  float valat(int index) {return data[index];};						// component accessor
  void setvalat(int index,float val) {data[index]=val;};				// component accessor
  void copy(Datapoint *indatap) {set(indatap->data);};					// make equal to copy
  void set(float *indata);								// set to vals
  void setloc(float loc[3],int x,int y,int z);
  int getsourcenum() {return (int)psourcenum;};						// accessor
  void print();										// debug print
};








//-------------------------------------------------------------------------------
//-- PointSource
//-------------------------------------------------------------------------------
//: PointSource is used so that a plot can have multiple datapoint sources (like overlaid plots)
// and each datapoint source (symbol)

class PointSource
{
private:
  string label;										// label for the symbol
  string symbol;									// shape (?)
  float color;										// color of symbol
  float scale;										// scale of symbol
public:
  PointSource(string inlabel);
  ~PointSource() {;};
public:
  float getscale() {return scale;};
};













//-------------------------------------------------------------------------------
//-- Plot
//-------------------------------------------------------------------------------

class Plot : public GrobVideoCameraDrawer
{
protected:
  JRinterpolater *rgbmap[GROBCOLORSCALES];
protected:
  GrobColorScaleType colorscale;
  PlotDisplayMode displaymode;
  int dims;										// number of dimensions
  string title;										// a big title
  string subtitle;									// a subtitle
  int refreshrate;									// how often to redraw
  int refreshcount;									//  
  int storerate;									// how often to store new
  int storecount;									//  
  int savepoints;									// number of points to save
  long seenpoints;									// number of points seen
  bool supportsinc;
  bool connect;
protected:
  int xaxis,yaxis,zaxis;								// currently used indexes
  int coloraxis,symbolaxis;								//  for drawing
  int storageaxis;									// index of storage axis
protected:
  bool needsadjust;
  bool needsredraw;
  bool show_axes;									//
  bool show_legend;									//
  bool show_titles, show_labels;							// whether to show these
  bool show_minorticks, show_majorticks;						//
  bool show_walls;									//
  Datapoint minpoint,maxpoint;								// min,max values
  list<Datapoint *>::iterator newpointer;						// pointer to first new
  Datapoint *lastpoint;						                        // pointer to last point drawn
  int newpoints;									// how many new
  int symbols;
  Grob *grobsymbol[10];									// dif symbols for plotting
  float symbolscale[10];								// scale of symbol
protected:
  list<Datapoint *> points;								// history of points
  vector<VisibleAxis *> axes;								// list of axes
  vector<PointSource *> psources;							// list of pointsources
public:
  Plot();
  ~Plot();
  void clearpoints();									// remove stored points
private:
  void make_colormaps();
  void free_colormaps();
public:
  void complete();									// plot is done building
  void addpoint(Datapoint *datap);							// add new point
  void addaxis(VisibleAxis *axisp);							// add a new axis
  int lookupaxis(string axisname);							// lookup axis number
  void addpsource(PointSource *psourcep);						// add a new axis
  bool adjustout(bool force=false);							// expand or shift if nesc.
  void updatedraw();									// 
  void makesymbols();									// some default symbols
  void scalesymbols();									// scale symbols
  void connectpoints(Datapoint *sdp,Datapoint *edp);
  void forceredraw();
public:
  void draw_axes() {;};									//
  void draw_legend() {;};								//
  void draw_titles() {;};								// draw various components
  void draw_labels() {;};								//
  void draw_ticks() {;};								//
  void draw_walls() {;};								//
public:
  void setdisplaymode(int inmode){displaymode=(PlotDisplayMode)inmode;forceredraw();};
  void setcolorscale(int incscale){if (incscale<GROBCOLORSCALES) {colorscale=(GrobColorScaleType)incscale;forceredraw();};};
public:
  GrobSpacerType wantspacer();
  bool grobvidcam_supportsincdraw() {return supportsinc;};   	// can inc draw to cam?
  bool grobvidcam_canincdraw() {if (!needsredraw && supportsinc) return true; else return false;};   	// can inc draw to cam?
  virtual void virt_complete() {;};							// plot is complete
  virtual void incorporate_point(Datapoint *datap) {;};					// incorporate new point
  virtual void axeschange() {needsredraw=true;}						// axes have changed
  virtual void exportfile(string filename,string format) {;};				// save data to file
  virtual string drawertype(){return "Plot";};
};





//-------------------------------------------------------------------------------
//-- PointsPlot
//-------------------------------------------------------------------------------

class PointsPlot : public Plot
{
public:
  PointsPlot();
  ~PointsPlot() {;};
public:
  void grobvidcam_incdraw(GrobVideoCamera *cam);	// inc draw to cam
  void grobvidcam_draw(GrobVideoCamera *cam);		// draw to cam
};



//-------------------------------------------------------------------------------
//-- TablePlot
//-------------------------------------------------------------------------------

class TablePlot : public Plot
{
private:
  PlotSampleMode sampmode;								// sampling mode
  PlotConnectorType tableconnector;
private:
  int *dimsize;										// vector of dim. sizes
  int *dimmults;									// vector of dim. indexs
  int elementcount;									// table size
  unsigned long *tablecounts;								// matrix of counts
  float *tabledata;									// matrix of tabledata
  float minval,maxval;
  float heightskew;
  bool testcolormap;
  bool normalizecolors;
  bool minclip;
  bool emptyclip;
  float minclipval;
  bool rebuilding;
  bool needsreminmax;
  bool useheights;
public:
  TablePlot();										// construct w/o allocate
  ~TablePlot();										// free space
public:
  void reallocate();
public:
  void clear();										// reset table enties
  void incorporate_point(Datapoint *datap);						// incorporate new point
  float valueat(int *index) {return tabledata[eindex(index)];};				// return indexed value
  unsigned long countat(int *index) {return tablecounts[eindex(index)];};		// return count at index
  int eindex(int *index);								// get index into table
  float valueat(int x,int y) {return tabledata[x*dimmults[xaxis]+y*dimmults[yaxis]];};             // inline table val
  unsigned long countat(int x,int y) {return tablecounts[x*dimmults[xaxis]+y*dimmults[yaxis]];};   // inline table count
public:
  void virt_complete();									// plot is built
  void complete_redraw();								// redraw plot completely
  void axeschange();								// axes have changed
  void grobvidcam_draw(GrobVideoCamera *cam);						// draw to cam
public:
  void setclip(bool inminclip=true){minclip=inminclip;forceredraw();};
  void setconnectorstyle(int incon){tableconnector=(PlotConnectorType) incon;forceredraw();};
public:
  void reverse_index(int *tindex,float *rindex);
private:
  void allocate();
  void deallocate();
  void findminmax();
private:
  void drawgvc_points(GrobVideoCamera *cam) {;};
  void drawgvc_bars(GrobVideoCamera *cam);
  void drawgvc_surface(GrobVideoCamera *cam);
  void drawgvc_flatgrid(GrobVideoCamera *cam) {;};
  void drawgvc_contour(GrobVideoCamera *cam) {;};
};
















#endif
