// tableplot.C
// Contains the methods for the derived TablePlot class, used for building histograms, barcharts, surfaceplots, etc.
// see plot.h for header

#include "plot.h"

#include <cmath>
#include <iostream>







//-------------------------------------------------------------------------------
//-- TablePlot
//-------------------------------------------------------------------------------

TablePlot::TablePlot()
  :Plot()
{
  // constructor
  dims=0;
  dimsize=NULL;
  tabledata=NULL;
  tablecounts=NULL;
  supportsinc=false;
  minval=maxval=0;
  xaxis=-1;yaxis=-1;
  displaymode=bars_display;
  colorscale=jet_scale;
  heightskew=1.0;
  minclip=false;
  minclipval=0.0;
  testcolormap=false;
  emptyclip=false;
  tableconnector=xy_connect;
  normalizecolors=true;
  needsreminmax=false;
  rebuilding=false;
  useheights=true;

  // test
  displaymode=surface_display;
  sampmode=count_mode;
  storerate=1;
  refreshrate=5;
  testcolormap=true;
  //  useheights=false;
}

TablePlot::~TablePlot()
{
  // destructor
  deallocate();
}


void TablePlot::deallocate()
{
  if (tabledata)
    {
      delete [] tabledata;
      tabledata=NULL;
    }
  if (tablecounts)
    {
      delete [] tabledata;
      tablecounts=NULL;
    }
  if (dimsize)
    {
      delete [] dimsize;
      delete [] dimmults;
      dimsize=NULL;
    }
}


void TablePlot::allocate()
{
  // allocate the required space
  int count;
  int mindex;

  // number of dimensions=number of axes
  dims=axes.size();
  if (dims==0)
    return;

  // allocate dimsize,dimmults space
  dimsize=new int[dims];
  dimmults=new int[dims];

  if (storageaxis==-1)
    {
      cout << "No storage axis defined, assuming last axis.\n";
      storageaxis=dims-1;
    }

  // initialize
  mindex=1;
  for (count=0;count<dims;++count)
    {
      dimsize[count]=axes[count]->getdiscretes();
      if (count==storageaxis)
	{
	  // since this is the storage axis, we don't include it in our table
	  dimmults[count]=0;
	}
      else
	{
	  dimmults[count]=mindex;
	  mindex*=dimsize[count];
	}
    }

  // calculate table size
  elementcount=1;
  for (count=0;count<dims;++count)
    {
      if (count!=storageaxis)
	elementcount*=dimsize[count];
    }

  // allocate table data space
  tablecounts=new unsigned long[elementcount];
  tabledata=new float[elementcount];

  if (!tablecounts || !tabledata)
    {
      cerr << "PLOT ERROR: couldn't get enough memory for "<<elementcount<<" table elements.\n";
      dims=0;
      elementcount=0;
    }

  //  cout << "Table allocated.  Elements="<<elementcount<<".\n";
  //  for (count=0;count<dims;++count)
  //    cout << "Dimension "<<count<< " size="<<dimsize[count]<<".\n";
  //  cout << "x dimension="<<xaxis<<"  dimsize="<<dimsize[xaxis]<<".\n";
  //  cout << "y dimension="<<yaxis<<"  dimsize="<<dimsize[yaxis]<<".\n";
  //  cout << "storage dim="<<storageaxis<<"  dimsize="<<storageaxis<<".\n";

  // clear the table
  clear();
}

void TablePlot::clear()
{
  // reset table entries (don't remove just zero out)
  for (int count=0;count<elementcount;++count)
    {
      tablecounts[count]=0;
      tabledata[count]=0.0;
    }
  newpoints=0;
  newpointer=points.end();
  minval=maxval=0;
}


void TablePlot::reallocate()
{
  // the table dimensions have changed, so we want to free and then reallocate the table
  deallocate();
  allocate();
}


void TablePlot::incorporate_point(Datapoint *datap)
{
  // incorporate datapoint to the table data
  int ind=0;
  int axind;
  float val,v,oldv;
  float minp,maxp;

  // now add it to our table
  // first we calculate where in table the datapoint falls
  for (int count=0;count<dims;++count)
    {
      if (count==storageaxis)
	{
	  // get the value from the datapoint that we use for storage
	  val=datap->valat(storageaxis);
	}
      else
	{
	  // get and add discrete axis index
	  axind=axes[count]->discreteind(datap->valat(count));
	  if (axind==-1)
	    {
	      // falls outside table, how did this happen
	      cout << "value out of axis range for table plot.\n";
	      return;
	    }
	  ind+=axind*dimmults[count];
	}
    }

  //  cout << "putting point in entry "<<ind<<"  with val="<<val<<".\n";

  // increment table count
  ++tablecounts[ind];
  oldv=tabledata[ind];

  // now we incorporate datapoint
  switch (sampmode)
    {
    case current_mode:
      v=tabledata[ind]=val;
      break;
    case average_mode:
      // store sum and divide later
      tabledata[ind]+=val;
      v=tabledata[ind]/tablecounts[ind];
      break;
    case sum_mode:
      tabledata[ind]+=val;
      v=tabledata[ind];
      break;
    case count_mode:
      v=++tabledata[ind];
      break;
    }

  if (rebuilding==false)
    {
      // now decide whether this new point might force a change of the storageaxis scale/range
      if (v<minval)
	minval=v;
      if (v>maxval)
	maxval=v;

      minp=minpoint.valat(storageaxis);
      maxp=maxpoint.valat(storageaxis);

      if (v<minp)
	{
	  minpoint.setvalat(storageaxis,v);
	  needsadjust=true;
	  needsredraw=true;
	}
      else if (minp==oldv && v>oldv)
	{
	  // we *may* have just replaced the min value, so we need to recheck everyone
	  needsreminmax=true;
	}

      if (v>maxp)
	{
	  maxpoint.setvalat(storageaxis,v);
	  needsadjust=true;
	  needsredraw=true;
	}
      else if (maxp==oldv && v<oldv)
	{
	  // we *may* have just replaced the max value, so we need to recheck everyone
	  needsreminmax=true;
	}
    }
}


void TablePlot::findminmax()
{
  // go through the whole table and recalc min and max
  float minp,maxp;
  float v;

  //  cout << "refinding min&max.\n";

  minval=maxval=tabledata[0];
  for (int count=1;count<elementcount;++count)
    {
      v=tabledata[count];
      if (v<minval)
	minval=v;
      if (v>maxval)
	maxval=v;
    }

  minp=minpoint.valat(storageaxis);
  maxp=maxpoint.valat(storageaxis);
  if (minp!=minval)
    {
      //      cout << "minmax FORCE min, was "<<minp<<" but should be "<<minval<<".\n";
      minpoint.setvalat(storageaxis,minval);
      needsadjust=needsredraw=true;
    }
  if (maxp!=maxval)
    {
      //      cout << "minmax FORCE max, was "<<maxp<<" but should be "<<maxval<<".\n";
      maxpoint.setvalat(storageaxis,maxval);
      needsadjust=needsredraw=true;
    }

  needsreminmax=false;
}



int TablePlot::eindex(int *index)
{
  // calculate index into table
  int ind=0;
  for (int count=0;count<dims;++count)
    ind+=index[count]*dimmults[count];
  return ind;
}



void TablePlot::axeschange()
{
  // when table axes change, we have to refill the table completely, not just redraw, because
  //  if the bins that datappoints now fall into may be dif. than what where they used to fall
  // NOTE: this does not yet support changing the number of bins per axis
  int len;

  cout << "Remaking table.\n";
  rebuilding=true;

  // clear the table
  clear();
  
  // reprocess all stored datapoints (if we haven't stored them all, we lose some old info)
  newpointer=points.begin();
  while (newpointer!=points.end())
    {
      incorporate_point(*newpointer);
      ++newpointer;
    }

  // now set redraw flag
  rebuilding=false;
  needsreminmax=true;
  needsredraw=true;
}



void TablePlot::virt_complete()
{
  // the plot is built

  // test
  axes[xaxis]->setdiscretes(13);
  axes[yaxis]->setdiscretes(13);

  // test stuff
  axes[xaxis]->setrange(-40,40);
  axes[yaxis]->setrange(-40,40);
  axes[zaxis]->setrange(0,40);

  // allocate space for table
  allocate();
  
  // force the plot to adjust to it's axes
  adjustout(true);
}



void TablePlot::reverse_index(int *tindex,float *rindex)
{
  // given a table position,calculate the inverse equiv. location in dataspace
  for (int count=0;count<dims;++count)
    {
      if (count!=storageaxis)
	rindex[count]=axes[count]->reverseind(tindex[count]);
    }
}





void TablePlot::grobvidcam_draw(GrobVideoCamera *cam)
{
  // completely redraw the plot
  // first see if we need to adjust axes because a value fell out of range before the camera had a chance to redraw
  //  this is a little tricky since the axis may already be on the screen!  we assume camera will deal with this sensibly.
  if (needsreminmax)
    findminmax();
  if (needsadjust)
    adjustout();
  if (needsreminmax)
    {
      // okay, the chain of a modified min/max of table has caused axis rescaling,
      // which has now caused a desire to once again find the new min&max
      findminmax();
    }
  
  if (yaxis==-1 || xaxis==-1)
    {
      // must be 2d to plot (ie have an x and a y dimension) ?
      // this restriction should be lifted!
      cout << "Error, table must be 2d to plot.\n";
      return;
    }

  // call the appropriate drawing function
  switch(displaymode)
    {
    case scatter_display:
      drawgvc_points(cam);
      break;
    case bars_display:
      drawgvc_bars(cam);
      break;
    case surface_display:
      drawgvc_surface(cam);
      break;
    case flatgrid_display:
      drawgvc_flatgrid(cam);
      break;
    case contour_display:
      drawgvc_contour(cam);
      break;
    }

  // done
  needsredraw=false;
}



































void TablePlot::drawgvc_bars(GrobVideoCamera *cam)
{
  // Draw a bar chart
  int x,y;
  float fval,hval;
  int symnum;
  Grob *grb;
  float loc[3];
  float scale;
  float xmin,xmax,ymin,ymax,xper,yper,zmin,zmax;
  float axismin,axismax,axisdif,axisscale,axisdifscale,symsize;
  float sx,sy,ex,ey;
  int xdim,ydim,xdimm,ydimm;
  float czmin,czmax,czdif,colorpos;
  bool colorbars;
  int colormapnum;
  GLfloat barColor[3];
  GLfloat whiteColor[] = {1, 1, 1, 1};
  float colorv[3];
  GrobDetail camdetail=cam->getdetail();

  // test - which symbol to use for display
  symnum=1;

  // get the min,max
  xdim=dimsize[xaxis];
  ydim=dimsize[yaxis];
  xmin=axes[xaxis]->getmin();
  xmax=axes[xaxis]->getmax();
  ymin=axes[yaxis]->getmin();
  ymax=axes[yaxis]->getmax();
  zmin=axes[zaxis]->getmin();
  zmax=axes[zaxis]->getmax();

  // color range
  if (testcolormap)
    {
      // a useful test to force grid through colormap
      czmin=0;
      czmax=(ydim-1)*(xdim-1);
    }
  else if (normalizecolors)
    {
      // normalize the color scale (to get full range of colors regardless of axis bufferspace) [see surface plot]
      czmin=minval;
      czmax=maxval;
    }
  else
    {
      czmin=zmin;
      czmax=zmax;
    }

  if (xmax-xmin<=0 || ymax-ymin<=0 || czmax-czmin<=0 || xdim==0 || ydim==0)
    {
      cout << "one or more axes is vacant; not plotting.\n";
      return;
    }

  // shorthands
  xdimm=xdim-1; ydimm=ydim-1;
  axismin=axes[storageaxis]->getmin();
  axismax=axes[storageaxis]->getmax();
  axisdif=axismax-axismin;
  czdif=czmax-czmin;
  axisscale=axes[storageaxis]->getscale();
  symsize=axes[storageaxis]->symbolsize()/axisscale;
  axisdifscale=axisdif*axisscale;
  xper=(xmax-xmin)/xdimm;
  yper=(ymax-ymin)/ydimm;

  // initialize
  loc[0]=loc[1]=loc[2]=0.0;

  // deicde which symbol to plot
  grb=grobsymbol[symnum];

  // are we using color bars, if so enable coloring
  colormapnum=(int)colorscale;
  if (colorscale!=colorless_scale && rgbmap[colormapnum]!=NULL)
    {
      colorbars=true;
      glEnable(GL_COLOR_MATERIAL);
      glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
      glLineWidth(1);
    }
  else
    colorbars=false;

  // now draw the bars
  sy=ymin;
  for (y=0;y<ydim;++y)
    {
      ey=sy+yper;
      sx=xmin;
      for (x=0;x<xdim;++x)
	{
	  ex=sx+xper;

	  // get the value at this table entry
	  hval=fval=valueat(x,y);

	  if (minclip)
	    {
	      // if activated, only neighborhoods above a minimum height get shown
	      // useful for viewing large sparse plots (speeds it up tremendously)
	      if (fval<=minclipval)
		{
		  sx=ex;
		  continue;
		}
	    }
	  else if (emptyclip)
	    {
	      // don't display table entries which are 'empty'
	      // a little more expensive to check for, but more generally applicable
	      if (countat(x,y)<=0)
		{
		  sx=ex;
		  continue;
		}
	    }

	  if (!useheights)
	    {
	      // we do not show z dimension height, all bars are 0 height
	      hval=axismin;
	    }

	  // stretch/scale z axis of symbol to fraction of plot z axis height
	  // this can be used to make "bars" out of any symbol, will look okay for cone,box, bad for sphere
	  scale=(hval-axismin)/axisdifscale;
	  grobsymbol[symnum]->setscale(2,scale);

	  // now form and set the location in plot of the symbol
	  loc[0]=sx;
	  loc[1]=sy;
	  // the symbol z location needs to be midway between bottom and value, so when stretched the symbol touches ground and val
	  loc[2]=(hval+axismin)/2.0;
	  grb->setlocation(loc);

	  if (colorbars)
	    {
	      // color it, if appropr, using normalized color value
	      if (testcolormap)
		rgbmap[colormapnum]->interpolate((y*xdim+x)/czdif,colorv);
	      else
		rgbmap[colormapnum]->interpolate((fval-czmin)/czdif,colorv);
	      // now set the gl colors
	      barColor[0]=colorv[0];  barColor[1]=colorv[1];  barColor[2]=colorv[2];
	      glColor3fv(barColor);

	      if (camdetail==lowdetail)
		cam->grobdraw(grb,lowdetail);
	      else
		{
		  // in order to color surface-textured grobs, we need to draw the specific grob which has no
		  // surface colors in it, then it will take on our colors
		  cam->grobdraw(grb,paintdetail);
		}
	    }
	  else
	    {
	      // draw it regularly
	      cam->grobdraw(grb);
	    }

	  // advance x position
	  sx=ex;
	}
      // advance y position
      sy=ey;
    }

  // restore state
  cam->setlinewidth();
  if (colorbars)
    {
      glDisable(GL_COLOR_MATERIAL);
      // for some reason, disabling GL_COLOR_MATERIAL still can leave residual color effects, at least in Mesa 2.5, so we kuldge:
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, whiteColor);
    }
}












void TablePlot::drawgvc_surface(GrobVideoCamera *cam)
{
  // Draw a surface chart
  //  GLfloat specular[] = {.2, .2, .2, .2};
  //  GLfloat shininess[] = {256};
  //  GLfloat specular[] = {.2, .2, .2, .2};
  GLfloat specular[] = {.2, .2, .2, 1};
  GLfloat shininess[] = {256};
  GLfloat humpColor[] = {0.8, 0.6, .1, 1};
  GLfloat triColor2[] = {1, 1, 1, 1};
  GLfloat whiteColor[] = {1, 1, 1, 1};
  int x,y;
  float fval;
  float avgz,normz;
  int count;
  GrobDetail camdetail=cam->getdetail();
  int xdim,ydim,xdimm,ydimm;
  float xmin,xmax,ymin,ymax,zmin,zmax,xper,yper,zdif;
  float lx,ly,lxper,lyper;
  int xps,yps;
  int xcurvepoints,ycurvepoints;
  float sx,sy,ex,ey;
  GLfloat glmapdata[2][2][3];
  GLfloat glcolordata[2][2][4];
  GLfloat *glmapdatap;
  GLfloat *glcolordatap;
  GLfloat vul,vur,vll,vlr;
  float colorv[3];
  int colormapnum;
  bool showlines,showsolid,showcliff,showcolor,showxlines,showylines;
  bool smoothsolid,usecolormap,colorlines;
  float maxnormz=-1;
  float curheightskew=1;

  // get the min,max
  xdim=dimsize[xaxis];
  ydim=dimsize[yaxis];
  xmin=axes[xaxis]->getmin();
  xmax=axes[xaxis]->getmax();
  ymin=axes[yaxis]->getmin();
  ymax=axes[yaxis]->getmax();

  // color range
  // NOTE: we normally use real min and max of z for calculating colors.  this means that we don't
  //  care how high the actual axis boundaries for z are, only the range of the data.
  //  this is useful for when the axes are big, but the data fall in a more restricted range,
  //  because it always displays a chart with the full z range, rather than a muted chart.
  //  this wouldn't be an issue with static charts, but for dynamic charts, it is signif,
  //  since data may chage over time and thus fall out of full color range, etc.
  //  the only drawback is that if data accumulate, existing patches will change color
  //  as the actual z range changes, even though the actual axes aren't changing. this may
  //  bother some people, so it should be an option
  if (testcolormap)
    {
      // a useful test to force grid through colormap
      zmin=0;
      zmax=(ydim-1)*(xdim-1);
    }
  else if (normalizecolors)
    {
      zmin=minval;
      zmax=maxval;
      curheightskew=heightskew;
    }
  else
    {
      zmin=axes[zaxis]->getmin();
      zmax=axes[zaxis]->getmax();
    }

  if (xmax-xmin<=0 || ymax-ymin<=0 || zmax-zmin<=0 || xdim<=1 || ydim<=1)
    {
      cout << "one or more axes is vacant; not plotting.\n";
      return;
    }

  // initialize
  glcolordata[0][0][3]=1;
  glcolordata[0][1][3]=1;
  glcolordata[1][0][3]=1;
  glcolordata[1][1][3]=1;

  // univ. defaults
  showlines=true;

  // connecting line styles
  switch (tableconnector)
    {
    case no_connect:
      showxlines=showylines=false;
      break;
    case x_connect:
      showxlines=true;
      showylines=false;
      break;
    case y_connect:
      showxlines=false;
      showylines=true;
      break;
    case xy_connect:
      showxlines=true;
      showylines=true;
      break;
    }

  // set some optional display niceties
  switch(camdetail)
    {
    case lowdetail:
      showlines=true;
      showcliff=false;
      xcurvepoints=1;
      ycurvepoints=1;
      colorlines=true;
      showcolor=false;
      showsolid=false;
      if (showlines==false)
	{
	  // there has to be something to look at !
	  showxlines=showylines=true;
	}
      break;
    case meddetail:
      showlines=true;
      showsolid=true;
      showcliff=false;
      xcurvepoints=min(3,60/xdim);
      ycurvepoints=min(3,60/ydim);
      showcolor=true;
      smoothsolid=false;
      colorlines=false;
      break;
    case highdetail:
      showlines=true;
      showsolid=true;
      showcliff=false;
      xcurvepoints=min(7,70/xdim);
      ycurvepoints=min(7,70/ydim);
      showcolor=true;
      smoothsolid=false;
      colorlines=false;
      //      smoothsolid=true;
      break;
    case finedetail:
      showlines=true;
      showsolid=true;
      showcliff=true;
      showcolor=true;
      smoothsolid=true;
      colorlines=false;
      showlines=true;
      xcurvepoints=min(15,100/xdim);
      ycurvepoints=min(15,100/ydim);
      break;
    }

  if (showxlines || showylines)
    showlines=true;
  else
    {
      if (showsolid)
	showlines=false;
      else
	{
	  // there must be something to show!
	  showxlines=showylines=showlines=true;
	}
    }

  if (useheights==false)
    {
      xcurvepoints/=2;
      ycurvepoints/=2;
    }

  // check curvepoints
  if (xcurvepoints<1)
    xcurvepoints=1;
  if (ycurvepoints<1)
    ycurvepoints=1;

  // set some camera drawing properties
  // at certain dimensions thick lines is bad, while for small plots its nice
  if (xdim*ydim>500)
    glLineWidth(1);
  else if (camdetail==finedetail)
    glLineWidth(2);

  // is there a color scale?
  colormapnum=(int)colorscale;
  if (colorscale==colorless_scale || rgbmap[colormapnum]==NULL)
    {
      showcolor=false;
      smoothsolid=false;
      colorlines=false;
    }

  // shorthands
  xdimm=xdim-1; ydimm=ydim-1;
  xper=(xmax-xmin)/xdimm;
  yper=(ymax-ymin)/ydimm;
  glmapdatap=&glmapdata[0][0][0];
  glcolordatap=&glcolordata[0][0][0];
  lxper=xper/xcurvepoints;
  lyper=yper/ycurvepoints;
  xps=xcurvepoints; yps=ycurvepoints;
  zdif=zmax-zmin;

  // prepare for the GL map
  glEnable(GL_MAP2_VERTEX_3);
  GrobGlHelper::depthoverwrite();

  // set material properties
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

  // are we using the color evaluators
  if (smoothsolid || colorlines)
    {
      glEnable(GL_MAP2_COLOR_4);
      usecolormap=true;
      glEnable(GL_COLOR_MATERIAL);
      glColorMaterial(GL_FRONT,GL_AMBIENT_AND_DIFFUSE);
      //      glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
    }
  else
    {
      usecolormap=false;
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, humpColor);
      glColor4fv(whiteColor);
    }

  if (showsolid)
    glEnable(GL_AUTO_NORMAL);

  if (colorlines)
    {
      // color lines require smooth shading to work
      glPushAttrib(GL_LIGHTING);
      glShadeModel(GL_SMOOTH);
    }


  // okay now draw the surface plot
  sy=ymin;
  for (y=0;y<ydimm;++y)
    {
      ey=sy+yper;
      sx=xmin;
      for (x=0;x<xdimm;++x)
	{
	  ex=sx+xper;

	  // get values at corners
	  vul=valueat(x,y);
	  vur=valueat(x+1,y);
	  vll=valueat(x,y+1);
	  vlr=valueat(x+1,y+1);

	  // test
	  //	  vur=vll=zmax;

	  if (minclip)
	    {
	      // if activated, only neighborhoods above a minimum height get shown
	      // useful for viewing large sparse plots (speeds it up tremendously)
	      if (vul<=minclipval && vur<=minclipval && vll<=minclipval && vlr<=minclipval)
		{
		  sx=ex;
		  continue;
		}
	    }
	  else if (emptyclip)
	    {
	      // don't display table entries which are 'empty'
	      // a little more expensive to check for, but more generally applicable
	      if (countat(x,y)<=0 && countat(x+1,y)<=0 && countat(x,y+1)<=0 && countat(x+1,y+1)<=0)
		{
		  sx=ex;
		  continue;
		}
	    }

	  // set upper left control point
	  glmapdata[0][0][0]=sx;
	  glmapdata[0][0][1]=sy;
	  // set upper right control point
	  glmapdata[0][1][0]=ex;
	  glmapdata[0][1][1]=sy;
	  // set lower left control point
	  glmapdata[1][0][0]=sx;
	  glmapdata[1][0][1]=ey;
	  // set lower right control point
	  glmapdata[1][1][0]=ex;
	  glmapdata[1][1][1]=ey;

	  if (useheights)
	    {
	      glmapdata[0][0][2]=vul;
	      glmapdata[0][1][2]=vur;
	      glmapdata[1][0][2]=vll;
	      glmapdata[1][1][2]=vlr;
	    }
	  else
	    {
	      // we do not show z dimension height, all bars are 0 height
	      glmapdata[0][0][2]=zmin;
	      glmapdata[0][1][2]=zmin;
	      glmapdata[1][0][2]=zmin;
	      glmapdata[1][1][2]=zmin;
	    }

	  // okay now make gridmap out of neighborhood
	  glMap2f(GL_MAP2_VERTEX_3, sx, ex, 3, 2, sy, ey, 2*3, 2, glmapdatap);

	  if (testcolormap)
	    {
	      // make the chart display the full color range, regardless of heights
	      vul=y*xdim+x;
	      vur=y*xdim+(x+1);
	      vll=(y+1)*xdim+x;
	      vlr=(y+1)*xdim+x+1;
	    }

	  // set the color neighborhood, if activated
	  if (usecolormap)
	    {
	      // set upper left control point
	      normz=(vul-zmin)/zdif;
	      rgbmap[colormapnum]->interpolate(normz,colorv);
	      glcolordata[0][0][0]=colorv[0];
	      glcolordata[0][0][1]=colorv[1];
	      glcolordata[0][0][2]=colorv[2];
	      
	      // set upper right control point
	      normz=(vur-zmin)/zdif;
	      rgbmap[colormapnum]->interpolate(normz,colorv);
	      glcolordata[0][1][0]=colorv[0];
	      glcolordata[0][1][1]=colorv[1];
	      glcolordata[0][1][2]=colorv[2];
	      
	      // set lower left control point
	      normz=(vll-zmin)/zdif;
	      rgbmap[colormapnum]->interpolate(normz,colorv);
	      glcolordata[1][0][0]=colorv[0];
	      glcolordata[1][0][1]=colorv[1];
	      glcolordata[1][0][2]=colorv[2];
	      
	      // set lower right control point
	      normz=(vlr-zmin)/zdif;
	      rgbmap[colormapnum]->interpolate(normz,colorv);
	      glcolordata[1][1][0]=colorv[0];
	      glcolordata[1][1][1]=colorv[1];
	      glcolordata[1][1][2]=colorv[2];

	      glMap2f(GL_MAP2_COLOR_4, sx, ex, 4, 2, sy, ey, 2*4, 2, glcolordatap);
	    }

	  // tell gl that we are going to be querying color and vertex from the 2d grid
	  glMapGrid2f(xcurvepoints, sx, ex, ycurvepoints, sy, ey);

	  // now render it
	  if (showsolid)
	    {
	      // draw solid render
	      if (smoothsolid)
		{
		  // use color evaluator for color
		  glEvalMesh2(GL_FILL, 0, xcurvepoints, 0, ycurvepoints);
		}
	      else
		{
		  if (showcolor)
		    {
		      // set a flat hump color based on height
		      // first get the avg normalized height  (from 0 to 1)
		      avgz=(vul+vur+vll+vlr)/4;
		      normz=(avgz-zmin)/zdif;
		      if (normz>maxnormz)
			maxnormz=normz;
		      // now ask the color ramp for what rgb values to use at this normalized height
		      rgbmap[colormapnum]->interpolate(normz*curheightskew,colorv);
		      // now set the gl colors
		      humpColor[0]=colorv[0];  humpColor[1]=colorv[1];  humpColor[2]=colorv[2];
		    }
		  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, humpColor);
		  glEvalMesh2(GL_FILL, 0, xcurvepoints, 0, ycurvepoints);
		}
	    }

	  if (showlines)
	    {
	      // draw border lines
	      // We could just call evalmesh2 with GL_LINE, as below, BUT it lines every
	      //  subsampled/interpolate division in curvepoints>1, which is too many, so we do it manually
	      //	      glEvalMesh2(GL_LINE, 0, xcurvepoints, 0, ycurvepoints);
	      //	      glEvalMesh2(GL_FILL, 0, xcurvepoints, 0, ycurvepoints);
	      //	      sx=ex;
	      //	      continue;
	      
	      if (smoothsolid)
		{
		  glDisable(GL_COLOR_MATERIAL);
		  if (!colorlines)
		    {
		      // turn off the automatic color evaluator
		      glDisable(GL_MAP2_COLOR_4);
		    }
		}

	      if (!colorlines)
		{
		  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor2);
		  glColor4fv(whiteColor);
		}

	      if (showxlines)
		{
		  glBegin(GL_LINE_STRIP);
		  for (count=0,lx=sx;count<=xps;lx+=lxper,++count)
		    glEvalCoord2f(lx,sy);
		  glEnd();
		  glBegin(GL_LINE_STRIP);
		  for (count=0,lx=ex;count<=xps;lx-=lxper,++count)
		    glEvalCoord2f(lx,ey);
		  glEnd();
		}
	      if (showylines)
		{
		  glBegin(GL_LINE_STRIP);
		  for (count=0,ly=sy;count<=yps;ly+=lyper,++count)
		    glEvalCoord2f(ex,ly);
		  glEnd();
		  glBegin(GL_LINE_STRIP);
		  for (count=0,ly=ey;count<=yps;ly-=lyper,++count)
		    glEvalCoord2f(sx,ly);
		  glEnd();
		}

	      if (smoothsolid)
		{
		  // turn back on color materials
		  glEnable(GL_COLOR_MATERIAL);
		  if (!colorlines)
		    {
		      // turn back on auto color evaluator
		      glEnable(GL_MAP2_COLOR_4);
		    }
		}

	    }
	  
	  // advance x position
	  sx=ex;
	}
      // advance y position
      sy=ey;
    }

  // a hack for dealing with spikey plots where the max norm height is dragged << 1.0
  // this value will be used for next plot as a gain on color value.  the effect of this is
  // to ensure that at least one patch on the plot will have the highest colorramp value.
  if (maxnormz>0)
    heightskew=1.0/maxnormz;
  
  // restore state
  if (colorlines)
    {
      // color lines push lighting bit, so we restore here.
      glPopAttrib();
    }
  GrobGlHelper::depthnormal();
  glDisable(GL_MAP2_VERTEX_3);
  if (showsolid)
    glDisable(GL_AUTO_NORMAL);
  if (usecolormap)
    {
      glDisable(GL_MAP2_COLOR_4);
      glDisable(GL_COLOR_MATERIAL);
    }
  cam->setlinewidth();
  // for some reason, disabling GL_COLOR_MATERIAL still can leave residual color effects, at least in Mesa 2.5, so we kuldge:
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, whiteColor);
  glColor4fv(whiteColor);
}


