// cfptimer.h
// This file describes the Configpp timer classes, which are used to invoke actions/objects
//  at regular intervals.

// forward declarations
class CfpTimer;
class CfpTimerManager;





#ifndef INC_cfptimerh
#define INC_cfptimerh

#include "cfpinc.h"

//Not implemented in gcc yet (2.8.1)
//namespace Configpp{






//:CfpstdTimer class is used to create clocks/timers which regularly activate other objects/accessors
// Useful for registering the GUI which should be invoked regularly, and for physical/agent simulations.
// Timers can be set to activate on intervals of program cycles, or *near* regular clock intervals.
//  note that when a timer requests to be activated at regular wallclock interval (in seconds),
//  such activation is only approximate, and may occur significantly less often if system/program load is high.
// The niceness value is a heuristic meant to represent a scale from 1 (highest priority) to 19 (lowest
// priority), such that timers with low priority may miss cycles in order to accomodate timers with low priority.
// The lower the niceness, the more often the timer will be polled for triggering.
//  (the gui should be lowest=1). valus are multiplicative (niceness=2 checked 1/2 as freq. as nice 1)
// Note that multiple timers can coexist, and multiple callbacks can be attached to each timer.

class CfpTimer
{
  friend class CfpTimerManager;
private:
  string name;
  CfpTimerType timertype;		// wallclockType, cycleclockType, or simclockType
  Configpp *configppp;			// pointer to the main configpp class
  CfpTimerManager *managerp;		// pointer to the timer manager
  int niceness;
  long activations;			// count of activations so far
  double wall_time;			// current wall time
  double wall_elapsedt;			// elapsed wall time since last activation
  double wall_deltat_activate;		// delta clock time between activations
  double wall_delay;			// avg. diff. between desired deltat and invoked deltat
  long cycle_time;			// current cycle count
  long cycle_elapsedt;			// elapsed cycles since last activation
  long cycle_deltat_activate;		// delta cycles bewteen activations
  double cycle_delay;			// avg. dif between desired deltacycles and invoked deltacycles
  double sim_time;			// current internal simulation time
  double sim_timestep;			// simulated simtime per cycle
  double sim_elapsedt;			// elapsed simulation time since last activation
  double sim_deltat_activate;		// delta simulation time bewteen activations
  double sim_delay;			// avg. dif between desired deltasimcycles and invoked deltasimcycles
  int deathcount;			// timer dies after certain number of activations (0=eternal)
  CfpValAccessor *acctimer;
  CfpVal acval;
public:
  CfpTimer(string intimername,Configpp *inconfigppp, int innice=5,int indeathcount=0);
  ~CfpTimer();
public:
  string getname() {return name;};						// accessor
  CfpValAccessor *getaccessorp() {return acctimer;};				// accessor
  CfpVAccstub *addlistener(CfpVAccessor *listenerp,int toindex=-1);				// accessor
  void simclock(double insimstep,double inseimact);
  void wallclock(double intimeact);
  void cycleclock(long intimeact);
public:
  virtual int cycle();								// decide whether to activate
  virtual int trigger();							// trigger the timer
};






//:The sole timer manager for the program, which invokes all other times.
// This class contains and manages a single clock, and passes it on to the other timers.
// Ideally, it should watch the individual timers and invoke some less often than others
//  based on their niceness values.  Right now we invoke every time every cycle.

class CfpTimerManager
{
  friend class CfpTimer;
private:
  list<CfpTimer *>::iterator tbegin;
  list<CfpTimer *>::iterator tend;
  list<CfpTimer *>::iterator ti;
  double wall_time;
  long cycle_time;
  int timercount;
  bool timers_modified;
  bool wantstodie;
  list<CfpTimer *> timers;
  Configpp *configppp;
  int min_nice,max_nice;
public:
  CfpTimerManager(Configpp *inconfigppp);
  ~CfpTimerManager();
public:
  double get_wall_time(){return wall_time;};
  double get_cycle_time(){return cycle_time;};
public:
  void addtimer(CfpTimer *newtimerp);
  void takeover();
  void cycle();
  void update_clock();
  void update_niceties();
public:
  void kill() {wantstodie=true;};
};



























//}
// Namespace close-brace
#endif
