// Datastream.C

#include "datastream.h"




Datastream::Datastream(Configpp *cfpp)
{
  // just initialize
  configppp=cfpp;
  curline=0;
  numlines=0;
  realstart=0;
  realend=-1;
  done=false;
}



Datastream::~Datastream()
{
  // cleanup before death
  if (!done)
    finish();
}


void Datastream::finish()
{
  // close any open files, and clean up
  closefile();
  done=true;
}


void Datastream::openfile()
{
  // open file for input
  filep=configppp->inFilestream(name,name,".dat");
}


void Datastream::closefile()
{
  if (filep)
    {
      filep->close();
      filep=NULL;
    }
}


void Datastream::error(string errstring,string hintstr="")
{
  // display an appropriate generic error for the Datastream
  if (!filep)
    throw CfpError(errstring,hintstr,io,true,configppp);
  else
    throw CfpError(errstring,hintstr,io,true,name,filep);
}



bool Datastream::ateof()
{
  // return true if we are at eof, otherwise false
  if (!filep)
    return true;
  else if (filep->eof())
    return true;
  else if (getendpos()>0 && filep->tellg()>getendpos())
    {
      // we have a limit inside this file beyond which we consider eof
      return true;
    }
  return false;
}





string Datastream::readline()
{
  // read a line and return as a string
  char str[MAX_DF_STRLEN];

  filep->getline(str,MAX_DF_STRLEN);
  return (string)str;
}

string Datastream::readword()
{
  // read a token and return as string
  string str;
  (*filep) >> str;
  return str;
}

StreamtokenType Datastream::readtoken(string &word,bool skipcomments=true)
{
  // read a token and return as string
  // here we go again with rewriting this type of function, we really should find a reusable solution
  // set string and return one of the following:
  //  StreamTokeType_CR,
  //  StreamTokeType_EOF,
  //  StreamTokeType_ERROR,
  //  StreamTokeType_number,
  //  StreamTokeType_alpha
  long fpos;
  char c;
  bool heof;
  bool inlinecomment=false;
  int commentblockdepth=0;

  word="";
  fpos=filep->tellg();
  if (fpos>realend)
    return StreamTokeType_EOF;

  for (;;)
    {
      // read a character from the stream
      filep->read(&c,1);
      ++fpos;
      if (c=='\n' && word=="")
	{
	  // a CR on its own just gets returned, nothing fancy
	  return StreamTokeType_CR;
	}
      else if ((c==' ' || c=='\t' || c=='\n' || fpos>realend) && word.length()>0)
	{
	  // hit a whitespace after a word
	  if (c=='\n')
	    {
	      // a little tricky, we want to push the CR back on the stack
	      //  so that we can reread it and return it next time.
	      --fpos;
	      filep->seekg(fpos,ios::beg);
	    }
	  if (word[0]>='0' && word[0]<='9')
	    return StreamTokeType_number;
	  else
	    return StreamTokeType_alpha;
	}
      else if (c!=' ' && c!='\t')
	{
	  // it's a charcter we want to add to the word
	  word+=c;
	}
      else
	{
	  // its a space or tab at the begining of a word, so ignore it.
	}
    }
  return StreamTokeType_ERROR;
}


bool Datastream::readline(char *str,int maxlen)
{
  // read a line into str and return true if succesfull, false if eof or error
  filep->getline(str,maxlen);
  return true;
}


void Datastream::reset()
{
  // move the file pointer back to begining of file
  filep->seekg(realstart,ios::beg);
}





DatastreamIn::DatastreamIn(Configpp *cfpp)
  : Datastream(cfpp)
{
  // nothing to do in the local constructor
  int age=23;
}



void DatastreamOut::openfile()
{
  // open file for appending, overwriting, or input
  if (fmode==append_writemode)
    filep=configppp->outappFilestream(name,name,".dat");
  else if (fmode==overwrite_writemode)
    filep=configppp->outFilestream(name,name,".dat");
}



void DatastreamOut::error(string errstring,string hintstr="")
{
  // display an appropriate generic error for the Datastream
  throw CfpError(errstring,hintstr,io,true,configppp);
}



void DatastreamOut::write(string str)
{
  // just write the line out
  filep->write(str.c_str(),str.length());
}

void DatastreamOut::write(char *str)
{
  // just write the line out
  filep->write(str,strlen(str));
}





void DatastreamInline::openfile()
{
  // open file for input (ignore writemode)
  filep=configppp->inFilestream(name,name,".dat");
  // now move it to where the data starts
  reset();
}





bool Datastream::readfloatarray(float **arraypp,long &rows,long &cols,bool weallocate=true,bool fullrows=true,bool formatted=true,bool skipcomments=true,bool complain=true)
{
  // read in an array of floats from a streamp
  // 1. IF weallocate==true, then we allocate space here for the array, user will deallocate later
  //      user should deallocate with  "delete [] arrayp;"
  // 2. columns and rows are the expected size of the table (if either are -1, then size will be adjusted to fit datastream).
  //     if either columns or rows ==-1, then we fill in the observed value.
  // 3. IF fullrows==true then it is an error (terminate, return false) if data stream has cols of unequal sizes
  //     if ==false then we pad short rows at end with some constant.
  // 4. IF formatted==true then it is an error if each row is not terminated by a CR
  // 5. IF skipcomments==true then we skip over comments (// starting a line and /* */ around blocks)
  // 6. IF complain==true then we throw a parsing error if we find an error.
  // Returns true on success, returns false on failure (and deallocates array!)
  long currow,mincols,maxcols,curcol;
  StreamtokenType ttype;
  string word;
  float ftemp;
  long ltemp;
  int index;
  long datacount;
  float padnumber=0.0;
  long readcurcol;
  float *arrayp;

  // shorthand notation copy
  arrayp=*arraypp;

  //  cout << "... About to try to read an array of data from stream '"+getname()+"'.\n";

  // a quick check for parameter consistency
  if (rows==-1 && cols==-1 && !formatted)
    {
      // we consider this an error since we have no way to figure out where rows start and end
      if (complain)
	error("Class programmer error.  Attempt was made to read an array from a file, but since column size was ommitted and data was indicated to be in free for, we have no way of identifying row segmentation.");
      return false;
    }
  if (cols==-1 && !fullrows)
    {
      // we consider this an error since kthe user has specified contradictory column sizes
      if (complain)
	error("Class programmer error.  Attempt was made to read an array from a file, but while a specific column size was specified, it was also specified that rows could be of differing column length.");
      return false;
    }

  // first we want to allocate space for the array.  if either row or columns ==-1, then we don't know
  //  the size until we read some data.  we could dynamically readjust the size of the array during reading, 
  //  but instead we make two passes through the data, first to check for errors and get the proper size,
  //  then the next to actually read the data.

  // first pass, determine (or verify) the size of the data stream, starting from begining of stream
  currow=curcol=0;
  mincols=maxcols=0;
  datacount=0;
  reset();
  for (;;)
    {
      ttype=readtoken(word,skipcomments);
      //      cout << "...  Token returned from stream: '"+word+"'\n.";
      if ((ttype==StreamTokeType_CR || ttype==StreamTokeType_EOF) && curcol>0)
	{
	  // the row of data has ended
	  if (curcol>maxcols || currow==0)
	    maxcols=curcol;
	  if (curcol<mincols || currow==0)
	    mincols=curcol;
	  curcol=0;
	  ++currow;
	}
      else if (ttype==StreamTokeType_number)
	{
	  // we got a valid piece of data
	  ++datacount;
	  ++curcol;
	}
      else if (ttype==StreamTokeType_alpha)
	{
	  // we read an alpha string value, not a number, so this is an error
	  if (complain)
	    error("Encountered a non-numeric value while reading a data table.");
	  return false;
	}
      if (ttype==StreamTokeType_EOF || ttype==StreamTokeType_ERROR)
	break;
    }

  // Okay, now we know about the data in the file.  Let's figure out if it is valid
  if (ttype==StreamTokeType_ERROR)
    {
      if (complain)
	error("File i/o error while reading a table of data from stream.");
      return false;
    }

  // first, if the user has told us that the data rows may not be CR terminated, then
  //  we can't just check cols and rows, so we have to be a little clever

  // if user cares, check to see if table has fixed col size
  if (mincols != maxcols && fullrows)
    {
      if (complain)
	error("Table data in stream has rows of varying column sizes (from "+JRtostr(mincols)+" to "+JRtostr(maxcols)+"), but we we were expecting rows with fixed column sizes.");
      return false;
    }

  if (formatted)
    {
      // The user has tld us that each row was CR terminated, so it's easy to validate.

      // check number of rows for acceptability
      if (rows==-1)
	rows=currow;
      else if (rows!=currow)
	{
	  if (complain)
	    error("Table data in stream has "+JRtostr(currow)+" rows, but we were expecting "+JRtostr(rows)+" rows.");
	  return false;
	}
      
      // check column sizes for acceptability
      if (cols==-1)
	cols=maxcols;
      else if (cols!=maxcols)
	{
	  if (complain)
	    error("Table data in stream has rows with "+JRtostr(maxcols)+" columns, but we were expecting rows with "+JRtostr(cols)+" columns.");
	  return false;
	}
    }
  else
    {
      // The data may not have CRs at the end of rows, which makes life a little more interesting
      if (rows==-1 && cols==-1)
	{
	  rows=currow;
	  cols=curcol;
	}
      else if (rows==-1)
	{
	  // try to infer rows from the # of columns the user said there were.
	  ftemp=(float)datacount/(float)cols;
	  ltemp=(long)ftemp;
	  if (ftemp != (float)ltemp)
	    {
	      // problem - if we use users column size, we do not end up with integer # of rows
	      if (complain)
		error("Table data in stream was assumed to be unformatted and with exactly "+JRtostr(cols)+" columns, but we found "+JRtostr(datacount)+" elements, leaving incomplete colums.");
	      return false;
	    }
	  rows=ltemp;
	}
      else if (cols==-1)
	{
	  // try to infer cols from the # of rows the user said there were.
	  ftemp=(float)datacount/(float)rows;
	  ltemp=(long)ftemp;
	  if (ftemp != (float)ltemp)
	    {
	      // problem - if we use users row size, we do not end up with integer # of cols
	      if (complain)
		error("Table data in stream was assumed to be unformatted and with exactly "+JRtostr(rows)+" rows, but we found "+JRtostr(datacount)+" elements, leaving incomplete rows.");
	      return false;
	    }
	  cols=ltemp;
	}
      else
	{
	  if (cols*rows!=datacount)
	    {
	      if (complain)
		error("Table data in stream was assumed to be unformatted and with exactly "+JRtostr(rows)+" rows and "+JRtostr(cols)+" cols, but we found "+JRtostr(datacount)+" elements.");
	      return false;
	    }
	}
    }

  // Okay, whew, the data passes the test, and we have valid row,col counts
  //  cout << "Datastream data is valid and of size "+JRtostr(rows)+" rows and "+JRtostr(cols)+" columns.\n";

  // allocate the array space if requested
  if (weallocate)
    arrayp=new float[rows*cols];

  // pad it out if this is an array with uneven column lengths
  if (!fullrows)
    {
      for (long count=0;count<rows*cols;++count)
	arrayp[count]=padnumber;
    }

  // second pass: load the data into the array, starting from begining of stream
  currow=curcol=readcurcol=0;
  reset();

  for (;;)
    {
      ttype=readtoken(word,skipcomments);
      if (ttype==StreamTokeType_CR || ttype==StreamTokeType_EOF)
	{
	  // the phsyical row of data has ended
	  if (readcurcol==0)
	    {
	      // this is a blank line (2 CRs), so we don't consider it to be a row specification
	    }
	  else if (formatted)
	    {
	      // CRs mean end of data row
	      ++currow;
	      curcol=0;
	    }
	  readcurcol=0;
	}
      else if (ttype==StreamTokeType_number)
	{
	  // we got a valid piece of data
	  index=currow*cols+curcol;
	  if (index>datacount)
	    {
	      error("Error while reading table data.  Somehow we miscalculated the table size, or datastream has been modified after verification.");
	      if (weallocate)
		delete [] arrayp;
	      return false;
	    }
	  arrayp[index]=(float)(strtod(word.c_str(),NULL));
	  ++readcurcol;
	  ++curcol;
	  if (!formatted)
	    {
	      // data rows are not CR formatted, so we decide when a row is done manually
	      if (curcol>cols)
		{
		  ++currow;
		  curcol=0;
		}
	    }
	}
      else if (ttype==StreamTokeType_alpha)
	{
	  // we should have picked this up during verification, but it can't hurt to repeat it here.
	  if (complain)
	    error("Somehow encountered a non-numeric value after while reading a data table *afer* verification.");
	  if (weallocate)
	    delete [] arrayp;
	  return false;
	}
      if (ttype==StreamTokeType_EOF || ttype==StreamTokeType_ERROR)
	break;
    }

  if (ttype==StreamTokeType_ERROR)
    {
      if (complain)
	error("File i/o error while reading a table of data from stream.");
      if (weallocate)
	delete [] arrayp;
      return false;
    }

  //  cout << "Data array succesfully read!\n";
  // modify the pointer to point to the array we allocated
  *arraypp=arrayp;
  return true;
}











