#ifndef COORD_VECTORS_HDR
#define COORD_VECTORS_HDR

#include <math.h>
#include <DMatrix.h>
#include <DVector.h>
#include <FMatrix.h>
#include <FVector.h>

// some handy defines
#define TMP_NEW new
#define ELT(a) ((*this)(a))
#define NON(v,a) (((DVector&)(v))(a))
#define VIRTUAL 

class DVectorR2;     // standard cartesian 2 and 3 space
class DVectorR3;
class DVectorHR2;    // homogeneous 2 and 3 space
class DVectorHR3; 
class DPolarCoords;     // polar coordinates (2d)
class DSphericalCoords; // spherical coords (3d) 
class DCylindricalCoords;// cylindrical coords (3d)

//================================================== 2-D DVectors
class DVectorR2:public VIRTUAL DVector {
public:
 double& X;
 double& Y;

 // ctors
 DVectorR2();
 DVectorR2(const DVector& v);
 DVectorR2(const DVectorR2& v);
 DVectorR2(const DVectorHR2& v);
 DVectorR2(double _x,double _y);

 // dtor
 virtual ~DVectorR2();

 // assignment
 DVectorR2& operator =(const DVector& v);
 DVectorR2& operator =(const DVectorR2& v);
 DVectorR2& operator =(const DVectorHR2& v);

 // optimizations over regular vectors
 double Length();
 double dot(const DVectorR2& v);
 double dot(const DVectorHR2& v);

 // additional
 double Angle();
 DVectorR2& Rotate_Self(double theta);
 DVectorR2& Rotate(double theta);
 
 // conversions
 DPolarCoords& DPolar();
};


//================================================== 3-D DVectors
class DVectorR3:public VIRTUAL DVector {
public:
 double& X;
 double& Y;
 double& Z;

 // ctors
 DVectorR3();
 DVectorR3(const DVector& v);
 DVectorR3(const DVectorR3& v);
 DVectorR3(const DVectorHR3& v);
 DVectorR3(double _x,double _y,double _z);

 // dtor
 virtual ~DVectorR3();

 // assignment
 DVectorR3& operator =(const DVector& v);
 DVectorR3& operator =(const DVectorR3& v);
 DVectorR3& operator =(const DVectorHR3& v);
friend DVectorR3 operator *(double s,const DVectorR3& v) {return v*s; }

 // optimizations over regular vectors
 double Length();
 double dot(const DVectorR3& v);
 double dot(const DVectorHR3& v);

 // additional
 DVectorR3& Rotate_Self_Around_X(double theta);
 DVectorR3& Rotate_Self_Around_Y(double theta);
 DVectorR3& Rotate_Self_Around_Z(double theta);
 DVectorR3& Rotate_Around_X(double theta);
 DVectorR3& Rotate_Around_Y(double theta);
 DVectorR3& Rotate_Around_Z(double theta);
 DVectorR3& cross(const DVectorR3& v);
 DVectorR3& cross(const DVectorHR3& v);
 
 // conversions
 DCylindricalCoords& DCylindrical();
 DSphericalCoords&   DSpherical();
};


//================================================== 2-D Homogeneous Coord DVectors
class DVectorHR2: public VIRTUAL DVector {
public:
 double& X;
 double& Y;

 // ctors
 DVectorHR2();
 DVectorHR2(const DVector& v);
 DVectorHR2(const DVectorR2& v);
 DVectorHR2(const DVectorHR2& v);
 DVectorHR2(double _x,double _y);

 // dtor;
 // JESSE modified 7/1/98;
 virtual ~DVectorHR2() {};

 // assignment
 DVectorHR2& operator =(const DVector& v);
 DVectorHR2& operator =(const DVectorR2& v);
 DVectorHR2& operator =(const DVectorHR2& v);

 // optimizations over regular vectors
 double Length();
 double dot(const DVectorR3& v);
 double dot(const DVectorHR3& v);

 // additional
 double Angle();
 DVectorHR2&   Rotate_Self(double theta);
 DVectorHR2&   Rotate(double theta);

 // conversions
 DPolarCoords& DPolar();
};

//================================================== 3-D Homogeneous Coord DVectors
class DVectorHR3: public VIRTUAL DVector {
public:
 double& X;
 double& Y;
 double& Z;

 // ctors
 DVectorHR3();
 DVectorHR3(const DVector& v);
 DVectorHR3(const DVectorR3& v);
 DVectorHR3(const DVectorHR3& v);
 DVectorHR3(double _x,double _y,double _z);

 // dtor
 virtual ~DVectorHR3() {};

 // assignment
 DVectorHR3& operator =(const DVector& v);
 DVectorHR3& operator =(const DVectorR3& v);
 DVectorHR3& operator =(const DVectorHR3& v);

 // optimizations over regular vectors
 double Length();
 double dot(const DVectorR3& v);
 double dot(const DVectorHR3& v);

 // additional
 DVectorHR3&   Rotate_Self_Around_X(double theta);
 DVectorHR3&   Rotate_Self_Around_Y(double theta);
 DVectorHR3&   Rotate_Self_Around_Z(double theta);
 DVectorHR3&   Rotate_Around_X(double theta);
 DVectorHR3&   Rotate_Around_Y(double theta);
 DVectorHR3&   Rotate_Around_Z(double theta);
 DVectorHR3&   cross(const DVectorR3& v);
 DVectorHR3&   cross(const DVectorHR3& v);

 // conversions
 DCylindricalCoords& DCylindrical();
 DSphericalCoords&   DSpherical();
};


//================================================== DPolar Coordinates
class DPolarCoords: public VIRTUAL DVector {
public:
 double& Radius;
 double& Theta;

 // ctors
 DPolarCoords();
 DPolarCoords(const DVector& v);
 DPolarCoords(const DPolarCoords& v);
 DPolarCoords(double _r,double _theta);

 // dtor
 virtual ~DPolarCoords();

 // assignment
 DPolarCoords& operator =(const DVector& v);
 DPolarCoords& operator =(const DPolarCoords& v);

 // optimizations over regular vectors
 double Length();
 double dot(const DPolarCoords& v);

 // additional
 double Angle();
 DPolarCoords& Rotate_Self(double theta);
 DPolarCoords& Rotate(double theta);
 
 // conversions
 DVectorR2& Cartesian();
 DVectorHR2&   Homogeneous();
};


//================================================== DCylindrical Coordinates
class DCylindricalCoords: public VIRTUAL DVector {
public:
 double& Radius;
 double& Theta;
 double& Z;

 // ctors
 DCylindricalCoords();
 DCylindricalCoords(const DVector& v);
 DCylindricalCoords(const DCylindricalCoords& v);
 DCylindricalCoords(double _radius,double _theta,double _z);

 // dtor
 virtual ~DCylindricalCoords();

 // assignment
 DCylindricalCoords& operator =(const DVector& v);
 DCylindricalCoords& operator =(const DCylindricalCoords& v);

 // optimizations over regular vectors
 double Length();

 // conversions
 DVectorR3& Cartesian();
 DVectorHR3&   Homogeneous();
};


//================================================== 3-D Homogeneous Coord DVectors
class DSphericalCoords: public VIRTUAL DVector {
public:
 double& Radius;
 double& Elevation;
 double& Azimuth;

 // ctors
 DSphericalCoords();
 DSphericalCoords(const DVector& v);
 DSphericalCoords(const DSphericalCoords& v);
 DSphericalCoords(double _r,double _e,double _a);

 // dtor
 virtual ~DSphericalCoords();

 // assignment
 DSphericalCoords& operator =(const DVector& v);
 DSphericalCoords& operator =(const DSphericalCoords& v);

 // optimizations over regular vectors
 double Length();

 // conversions
 DVectorR3& Cartesian();
 DVectorHR3&   Homogeneous();
};



//================================================================================
//================================================================================
//
//
//
inline DVectorR2::DVectorR2():DVector(2,ColV),X(ELT(0)),Y(ELT(1))
   { X=0.0; Y=0.0; }
inline DVectorR2::DVectorR2(const DVector& v):DVector(2,ColV),X(ELT(0)),Y(ELT(1))
   { X=NON(v,0); Y=NON(v,1); }
inline DVectorR2::DVectorR2(const DVectorR2& v):DVector(2,ColV),X(ELT(0)),Y(ELT(1))
   { X=v.X; Y=v.Y; }
inline DVectorR2::DVectorR2(const DVectorHR2& v):DVector(2,ColV),X(ELT(0)),Y(ELT(1))
   { X=v.X; Y=v.Y; }
inline DVectorR2::DVectorR2(double _x,double _y):DVector(2,ColV),X(ELT(0)),Y(ELT(1))
   { X=_x; Y=_y; }

inline DVectorR2::~DVectorR2() {};

inline DVectorR2& DVectorR2::operator =(const DVector& v)
   { X=v(0); Y=v(1); return (*this); }
inline DVectorR2& DVectorR2::operator =(const DVectorR2& v)
   { X=v.X; Y=v.Y; return (*this); }
inline DVectorR2& DVectorR2::operator =(const DVectorHR2& v)
   { X=v.X; Y=v.Y; return (*this); }
 
inline double DVectorR2::Length()
   { return hypot(X,Y); }
inline double DVectorR2::dot(const DVectorR2& v)
   { return ((X*v.X)+(Y*v.Y)); }
inline double DVectorR2::dot(const DVectorHR2& v)
   { return ((X*v.X)+(Y*v.Y)); }

inline double DVectorR2::Angle()       { return atan2(Y,X); }
inline DVectorR2& DVectorR2::Rotate_Self(double theta) {
   register double costh=cos(theta); register double sinth=sin(theta);
   register double tmp_x=costh*X-sinth*Y; Y=sinth*X+costh*Y; X=tmp_x;
   return (*this);
   }
inline DVectorR2& DVectorR2::Rotate(double theta)
   { return (TMP_NEW DVectorR2(*this))->Rotate_Self(theta); }
 
inline DPolarCoords& DVectorR2::DPolar()
   { return (*(TMP_NEW DPolarCoords(Length(),Angle()))); }


//================================================================================
//================================================================================
//
//
//
inline DVectorR3::DVectorR3():DVector(3,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=0.0; Y=0.0; Z=0.0; }
inline DVectorR3::DVectorR3(const DVector& v):DVector(3,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=v(0); Y=v(1); Z=v(2); }
inline DVectorR3::DVectorR3(const DVectorR3& v):DVector(3,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=v.X; Y=v.Y; Z=v.Z; }
inline DVectorR3::DVectorR3(const DVectorHR3& v):DVector(3,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=v.X; Y=v.Y; Z=v.Z; }
inline DVectorR3::DVectorR3(double _x,double _y,double _z):DVector(3,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=_x; Y=_y; Z=_z; }

inline DVectorR3::~DVectorR3() {};

inline DVectorR3& DVectorR3::operator =(const DVector& v)
   { X=v(0); Y=v(1); Z=v(2); return (*this); }
inline DVectorR3& DVectorR3::operator =(const DVectorR3& v)
   { X=v.X; Y=v.Y; Z=v.Z; return (*this); }
inline DVectorR3& DVectorR3::operator =(const DVectorHR3& v)
   { X=v.X; Y=v.Y; Z=v.Z; return (*this); }

inline double DVectorR3::Length()
   { return sqrt(X*X+Y*Y+Z*Z); }
inline double DVectorR3::dot(const DVectorR3& v)
   { return ((X*v.X)+(Y*v.Y)+(Z*v.Z)); }
inline double DVectorR3::dot(const DVectorHR3& v)
   { return ((X*v.X)+(Y*v.Y)+(Z*v.Z)); }

inline DVectorR3& DVectorR3::Rotate_Self_Around_X(double theta) {
   register double costh=cos(theta); register double sinth=sin(theta);
   register double tmp_y=costh*Y-sinth*Z; Z=sinth*Y+costh*Z; Y=tmp_y;
   return (*this); }

inline DVectorR3& DVectorR3::Rotate_Self_Around_Y(double theta) {
   register double costh=cos(theta); register double sinth=sin(theta);
   register double tmp_x=costh*X-sinth*Z; Z=sinth*X+costh*Z; X=tmp_x;
   return (*this); }

inline DVectorR3& DVectorR3::Rotate_Self_Around_Z(double theta) {
   register double costh=cos(theta); register double sinth=sin(theta);
   register double tmp_x=costh*X-sinth*Y; Y=sinth*X+costh*Y; X=tmp_x;
   return (*this); }

inline DVectorR3& DVectorR3::Rotate_Around_X(double theta)
   { return (TMP_NEW DVectorR3(*this))->Rotate_Self_Around_X(theta); }
inline DVectorR3& DVectorR3::Rotate_Around_Y(double theta)
   { return (TMP_NEW DVectorR3(*this))->Rotate_Self_Around_Y(theta); }
inline DVectorR3& DVectorR3::Rotate_Around_Z(double theta)
   { return (TMP_NEW DVectorR3(*this))->Rotate_Self_Around_Z(theta); }
inline DVectorR3& DVectorR3::cross(const DVectorR3& v)
   { return *(TMP_NEW DVectorR3((Y*v.Z)-(Z*v.Y),(Z*v.X)-(X*v.Z),(X*v.Y)-(Y*v.X))); }
inline DVectorR3& DVectorR3::cross(const DVectorHR3& v)
   { return *(TMP_NEW DVectorR3((Y*v.Z)-(Z*v.Y),(Z*v.X)-(X*v.Z),(X*v.Y)-(Y*v.X))); }

inline  DCylindricalCoords&   DVectorR3::DCylindrical()
   { return (*(TMP_NEW DCylindricalCoords(Length(),atan2(Y,X),Z))); }
inline DSphericalCoords&   DVectorR3::DSpherical()
   { return (*(TMP_NEW DSphericalCoords(Length(),atan2(sqrt(X*X+Y*Y),Z),atan2(Y,X)))); }


//======================================================================
//======================================================================
//
//
//
inline DVectorHR2::DVectorHR2():DVector(3,ColV),X(ELT(0)),Y(ELT(1))
   { X=0.0; Y=0.0; ELT(2)=1.0; }
inline DVectorHR2::DVectorHR2(const DVector& v):DVector(3,ColV),X(ELT(0)),Y(ELT(1))
   { X=v(0); Y=v(1); ELT(2)=1.0; }
inline DVectorHR2::DVectorHR2(const DVectorR2& v):DVector(3,ColV),X(ELT(0)),Y(ELT(1))
   { X=v.X; Y=v.Y; ELT(2)=1.0; }
inline DVectorHR2::DVectorHR2(const DVectorHR2& v):DVector(3,ColV),X(ELT(0)),Y(ELT(1))
   { X=v.X; Y=v.Y; ELT(2)=1.0; }
inline DVectorHR2::DVectorHR2(double _x,double _y):DVector(3,ColV),X(ELT(0)),Y(ELT(1))
   { X=_x; Y=_y; ELT(2)=1.0; }

// JESSE 7/1/98
//line DVectorHR2::~DVectorHR2() {};

inline DVectorHR2& DVectorHR2::operator =(const DVector& v)
   { X=v(0); Y=v(1); return (*this); }
inline DVectorHR2& DVectorHR2::operator =(const DVectorR2& v)
   { X=v.X; Y=v.Y; return (*this); }
inline DVectorHR2& DVectorHR2::operator =(const DVectorHR2& v)
   { X=v.X; Y=v.Y; return (*this); }

inline double DVectorHR2::Length()
   { return hypot(X,Y); }
inline double DVectorHR2::dot(const DVectorR3& v)
   { return ((X*v.X)+(Y*v.Y)); }
inline double DVectorHR2::dot(const DVectorHR3& v)
   { return ((X*v.X)+(Y*v.Y)); }

inline double DVectorHR2::Angle()
   { return atan2(Y,X); }
inline DVectorHR2&   DVectorHR2::Rotate_Self(double theta) {
   register double costh=cos(theta); register double sinth=sin(theta);
   register double tmp_x=costh*X-sinth*Y; Y=sinth*X+costh*Y; X=tmp_x;
   return (*this);
   }
inline DVectorHR2&   DVectorHR2::Rotate(double theta)
   { return (TMP_NEW DVectorHR2(*this))->Rotate_Self(theta); }
 
inline DPolarCoords& DVectorHR2::DPolar()
   { return (*(TMP_NEW DPolarCoords(Length(),Angle()))); }


//======================================================================
//======================================================================
//
//
//
inline DVectorHR3::DVectorHR3():DVector(4,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=0.0; Y=0.0; Z=0.0; ELT(3)=1.0; }
inline DVectorHR3::DVectorHR3(const DVector& v):DVector(4,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=v(0); Y=v(1); Z=v(2); ELT(3)=1.0; }
inline DVectorHR3::DVectorHR3(const DVectorR3& v):DVector(4,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=v.X; Y=v.Y; Z=v.Z; ELT(3)=1.0; }
inline DVectorHR3::DVectorHR3(const DVectorHR3& v):DVector(4,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=v.X; Y=v.Y; Z=v.Z; ELT(3)=1.0; }
inline DVectorHR3::DVectorHR3(double _x,double _y,double _z):DVector(4,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=_x; Y=_y; Z=_z; ELT(3)=1.0; }


// Jesse commented 7/1/98
//line DVectorHR3::~DVectorHR3() {};

inline DVectorHR3& DVectorHR3::operator =(const DVector& v)
   { X=v(0); Y=v(1); Z=v(2); return (*this); }
inline DVectorHR3& DVectorHR3::operator =(const DVectorR3& v)
   { X=v.X; Y=v.Y; Z=v.Z; return (*this); }
inline DVectorHR3& DVectorHR3::operator =(const DVectorHR3& v)
   { X=v.X; Y=v.Y; Z=v.Z; return (*this); }


inline double DVectorHR3::Length()
   { return sqrt(X*X+Y*Y+Z+Z); }
inline double DVectorHR3::dot(const DVectorR3& v)
   { return ((X*v.X)+(Y*v.Y)+(Z*v.Z)); }
inline double DVectorHR3::dot(const DVectorHR3& v)
   { return ((X*v.X)+(Y*v.Y)+(Z*v.Z)); }

inline DVectorHR3&   DVectorHR3::Rotate_Self_Around_X(double theta) {
   register double costh=cos(theta); register double sinth=sin(theta);
   register double tmp_y=costh*Y-sinth*Z; Z=sinth*Y+costh*Z; Y=tmp_y;
   return (*this); }
inline DVectorHR3&   DVectorHR3::Rotate_Self_Around_Y(double theta) {
   register double costh=cos(theta); register double sinth=sin(theta);
   register double tmp_x=costh*X-sinth*Z; Z=sinth*X+costh*Z; X=tmp_x;
   return (*this); }
inline DVectorHR3&   DVectorHR3::Rotate_Self_Around_Z(double theta) {
   register double costh=cos(theta); register double sinth=sin(theta);
   register double tmp_x=costh*X-sinth*Y; Y=sinth*X+costh*Y; X=tmp_x;
   return (*this); }
inline DVectorHR3&   DVectorHR3::Rotate_Around_X(double theta)
   { return (TMP_NEW DVectorHR3(*this))->Rotate_Self_Around_X(theta); }
inline DVectorHR3&   DVectorHR3::Rotate_Around_Y(double theta)
   { return (TMP_NEW DVectorHR3(*this))->Rotate_Self_Around_Y(theta); }
inline DVectorHR3&   DVectorHR3::Rotate_Around_Z(double theta)
   { return (TMP_NEW DVectorHR3(*this))->Rotate_Self_Around_Z(theta); }
inline DVectorHR3&   DVectorHR3::cross(const DVectorR3& v)
   { return *(TMP_NEW DVectorHR3((Y*v.Z)-(Z*v.Y),(Z*v.X)-(X*v.Z),(X*v.Y)-(Y*v.X))); }
inline DVectorHR3&   DVectorHR3::cross(const DVectorHR3& v)
   { return *(TMP_NEW DVectorHR3((Y*v.Z)-(Z*v.Y),(Z*v.X)-(X*v.Z),(X*v.Y)-(Y*v.X))); }
 
inline DCylindricalCoords& DVectorHR3::DCylindrical()
   { return (*(TMP_NEW DCylindricalCoords(Length(),atan2(Y,X),Z))); }
inline DSphericalCoords& DVectorHR3::DSpherical()
   { return (*(TMP_NEW DSphericalCoords(Length(),atan2(sqrt(X*X+Y*Y),Z),atan2(Y,X)))); }

//================================================================================
//================================================================================
//
//
//
inline DPolarCoords::DPolarCoords():
   DVector(2),Radius(ELT(0)),Theta(ELT(1))
   { Radius=0.0; Theta=0.0; }
inline DPolarCoords::DPolarCoords(const DVector& v):
   DVector(2),Radius(ELT(0)),Theta(ELT(1))
   { Radius=v(0); Theta=v(1); }
inline DPolarCoords::DPolarCoords(const DPolarCoords& v):
   DVector(2),Radius(ELT(0)),Theta(ELT(1))
   { Radius=v.Radius; Theta=v.Theta; }
inline DPolarCoords::DPolarCoords(double _r,double _theta):
   DVector(2),Radius(ELT(0)),Theta(ELT(1))
   { Radius=_r; Theta=_theta; }

inline DPolarCoords::~DPolarCoords() {}

inline DPolarCoords& DPolarCoords::operator =(const DVector& v)
   { Radius=v(0); Theta=v(1); return (*this); }
inline DPolarCoords& DPolarCoords::operator =(const DPolarCoords& v)
   { Radius=v.Radius; Theta=v.Theta; return (*this); }

inline double DPolarCoords::Length()
   { return Radius; }
inline double DPolarCoords::dot(const DPolarCoords& v)
   { return Radius*v.Radius*cos(fabs(Theta-v.Theta)); }

inline double DPolarCoords::Angle()
   { return Theta; }
inline DPolarCoords& DPolarCoords::Rotate_Self(double theta)
   { Theta+=theta; return (*this); }
inline DPolarCoords& DPolarCoords::Rotate(double theta)
   { return (TMP_NEW DPolarCoords(*this))->Rotate_Self(theta); }
 
inline DVectorR2& DPolarCoords::Cartesian()
   { return *(TMP_NEW DVectorR2(Radius*cos(Theta),Radius*sin(Theta))); }
inline DVectorHR2&   DPolarCoords::Homogeneous()
   { return *(TMP_NEW DVectorHR2(Radius*cos(Theta),Radius*sin(Theta))); }

//================================================================================
//================================================================================
//
//
//

inline DCylindricalCoords::DCylindricalCoords():
   DVector(3),Radius(ELT(0)),Theta(ELT(1)),Z(ELT(2))
   { Radius=0.0; Theta=0.0; Z=0.0; }
inline DCylindricalCoords::DCylindricalCoords(const DVector& v):
   DVector(3),Radius(ELT(0)),Theta(ELT(1)),Z(ELT(2))
   { Radius=v(0); Theta=v(1); Z=v(2); }
inline DCylindricalCoords::DCylindricalCoords(const DCylindricalCoords& v):
   DVector(3),Radius(ELT(0)),Theta(ELT(1)),Z(ELT(2))
   { Radius=v.Radius; Theta=v.Theta; Z=v.Z; }
inline DCylindricalCoords::DCylindricalCoords(double _radius,double _theta,double _z):
   DVector(3),Radius(ELT(0)),Theta(ELT(1)),Z(ELT(2))
   { Radius=_radius; Theta=_theta; Z=_z; }

inline DCylindricalCoords::~DCylindricalCoords() {}

inline DCylindricalCoords& DCylindricalCoords::operator =(const DVector& v)
   { Radius=v(0); Theta=v(1); Z=v(2); return (*this); }
inline DCylindricalCoords& DCylindricalCoords::operator =(const DCylindricalCoords& v)
   { Radius=v.Radius; Theta=v.Theta; Z=v.Z; return (*this); }

inline double DCylindricalCoords::Length()
   { return Radius; }

inline DVectorR3& DCylindricalCoords::Cartesian()
   { return *(TMP_NEW DVectorR3(Radius*cos(Theta),Radius*sin(Theta),Z)); };

inline DVectorHR3& DCylindricalCoords::Homogeneous()
   { return *(TMP_NEW DVectorHR3(Radius*cos(Theta),Radius*sin(Theta),Z)); };


//================================================================================
//================================================================================
//
//
//


inline DSphericalCoords::DSphericalCoords():
   DVector(3),Radius(ELT(0)),Elevation(ELT(1)),Azimuth(ELT(2))
   { Radius=0.0; Elevation=0.0; Azimuth=0.0; }
inline DSphericalCoords::DSphericalCoords(const DVector& v):
   DVector(3),Radius(ELT(0)),Elevation(ELT(1)),Azimuth(ELT(2))
   { Radius=v(0); Elevation=v(1); Azimuth=v(2); }
inline DSphericalCoords::DSphericalCoords(const DSphericalCoords& v):
   DVector(3),Radius(ELT(0)),Elevation(ELT(1)),Azimuth(ELT(2))
   { Radius=v.Radius; Elevation=v.Elevation; Azimuth=v.Azimuth; }
inline DSphericalCoords::DSphericalCoords(double _r,double _e,double _a):
   DVector(3),Radius(ELT(0)),Elevation(ELT(1)),Azimuth(ELT(2))
   { Radius=_r; Elevation=_e; Azimuth=_a; }

inline DSphericalCoords::~DSphericalCoords() {}

inline DSphericalCoords& DSphericalCoords::operator =(const DVector& v)
   { Radius=v(0); Elevation=v(1); Azimuth=v(2); return (*this); }
inline DSphericalCoords& DSphericalCoords::operator =(const DSphericalCoords& v)
   { Radius=v.Radius; Elevation=v.Elevation; Azimuth=v.Azimuth; return (*this); }

inline double DSphericalCoords::Length()
   { return Radius; }

inline DVectorR3& DSphericalCoords::Cartesian()
   { return *(TMP_NEW DVectorR3(
                              Radius*sin(Elevation)*cos(Azimuth),
                              Radius*sin(Elevation)*sin(Azimuth),
                              Radius*cos(Elevation))); }
inline DVectorHR3&   DSphericalCoords::Homogeneous()
   { return *(TMP_NEW DVectorHR3(
                              Radius*sin(Elevation)*cos(Azimuth),
                              Radius*sin(Elevation)*sin(Azimuth),
                              Radius*cos(Elevation))); }

#undef TMP_NEW
#undef ELT
#undef NON
#undef VIRTUAL 
#define TMP_NEW new
#define ELT(a) ((*this)(a))
#define NON(v,a) (((FVector&)(v))(a))
#define VIRTUAL 

class FVectorR2;     // standard cartesian 2 and 3 space
class FVectorR3;
class FVectorHR2;    // homogeneous 2 and 3 space
class FVectorHR3; 
class FPolarCoords;     // polar coordinates (2d)
class FSphericalCoords; // spherical coords (3d) 
class FCylindricalCoords;// cylindrical coords (3d)

//================================================== 2-F FVectors
class FVectorR2:public VIRTUAL FVector {
public:
 float& X;
 float& Y;

 // ctors
 FVectorR2();
 FVectorR2(const FVector& v);
 FVectorR2(const FVectorR2& v);
 FVectorR2(const FVectorHR2& v);
 FVectorR2(float _x,float _y);

 // dtor
 virtual ~FVectorR2();

 // assignment
 FVectorR2& operator =(const FVector& v);
 FVectorR2& operator =(const FVectorR2& v);
 FVectorR2& operator =(const FVectorHR2& v);

 // optimizations over regular vectors
 float Length();
 float dot(const FVectorR2& v);
 float dot(const FVectorHR2& v);

 // additional
 float Angle();
 FVectorR2& Rotate_Self(float theta);
 FVectorR2& Rotate(float theta);
 
 // conversions
 FPolarCoords& FPolar();
};


//================================================== 3-F FVectors
class FVectorR3:public VIRTUAL FVector {
public:
 float& X;
 float& Y;
 float& Z;

 // ctors
 FVectorR3();
 FVectorR3(const FVector& v);
 FVectorR3(const FVectorR3& v);
 FVectorR3(const FVectorHR3& v);
 FVectorR3(float _x,float _y,float _z);

 // dtor
 virtual ~FVectorR3();

 // assignment
 FVectorR3& operator =(const FVector& v);
 FVectorR3& operator =(const FVectorR3& v);
 FVectorR3& operator =(const FVectorHR3& v);
friend inline FVectorR3 operator *(float s,const FVectorR3& v) { return v*s; }

 // optimizations over regular vectors
 float Length();
 float dot(const FVectorR3& v);
 float dot(const FVectorHR3& v);

 // additional
 FVectorR3& Rotate_Self_Around_X(float theta);
 FVectorR3& Rotate_Self_Around_Y(float theta);
 FVectorR3& Rotate_Self_Around_Z(float theta);
 FVectorR3& Rotate_Around_X(float theta);
 FVectorR3& Rotate_Around_Y(float theta);
 FVectorR3& Rotate_Around_Z(float theta);
 FVectorR3& cross(const FVectorR3& v);
 FVectorR3& cross(const FVectorHR3& v);
 
 // conversions
 FCylindricalCoords& FCylindrical();
 FSphericalCoords&   FSpherical();
};


//================================================== 2-F Homogeneous Coord FVectors
class FVectorHR2: public VIRTUAL FVector {
public:
 float& X;
 float& Y;

 // ctors
 FVectorHR2();
 FVectorHR2(const FVector& v);
 FVectorHR2(const FVectorR2& v);
 FVectorHR2(const FVectorHR2& v);
 FVectorHR2(float _x,float _y);

 // dtor
 virtual ~FVectorHR2();

 // assignment
 FVectorHR2& operator =(const FVector& v);
 FVectorHR2& operator =(const FVectorR2& v);
 FVectorHR2& operator =(const FVectorHR2& v);

 // optimizations over regular vectors
 float Length();
 float dot(const FVectorR3& v);
 float dot(const FVectorHR3& v);

 // additional
 float Angle();
 FVectorHR2&   Rotate_Self(float theta);
 FVectorHR2&   Rotate(float theta);

 // conversions
 FPolarCoords& FPolar();
};

//================================================== 3-F Homogeneous Coord FVectors
class FVectorHR3: public VIRTUAL FVector {
public:
 float& X;
 float& Y;
 float& Z;

 // ctors
 FVectorHR3();
 FVectorHR3(const FVector& v);
 FVectorHR3(const FVectorR3& v);
 FVectorHR3(const FVectorHR3& v);
 FVectorHR3(float _x,float _y,float _z);

 // dtor
 virtual ~FVectorHR3() {}

 // assignment
 FVectorHR3& operator =(const FVector& v);
 FVectorHR3& operator =(const FVectorR3& v);
 FVectorHR3& operator =(const FVectorHR3& v);

 // optimizations over regular vectors
 float Length();
 float dot(const FVectorR3& v);
 float dot(const FVectorHR3& v);

 // additional
 FVectorHR3&   Rotate_Self_Around_X(float theta);
 FVectorHR3&   Rotate_Self_Around_Y(float theta);
 FVectorHR3&   Rotate_Self_Around_Z(float theta);
 FVectorHR3&   Rotate_Around_X(float theta);
 FVectorHR3&   Rotate_Around_Y(float theta);
 FVectorHR3&   Rotate_Around_Z(float theta);
 FVectorHR3&   cross(const FVectorR3& v);
 FVectorHR3&   cross(const FVectorHR3& v);

 // conversions
 FCylindricalCoords& FCylindrical();
 FSphericalCoords&   FSpherical();
};


//================================================== FPolar Coordinates
class FPolarCoords: public VIRTUAL FVector {
public:
 float& Radius;
 float& Theta;

 // ctors
 FPolarCoords();
 FPolarCoords(const FVector& v);
 FPolarCoords(const FPolarCoords& v);
 FPolarCoords(float _r,float _theta);

 // dtor
 virtual ~FPolarCoords();

 // assignment
 FPolarCoords& operator =(const FVector& v);
 FPolarCoords& operator =(const FPolarCoords& v);

 // optimizations over regular vectors
 float Length();
 float dot(const FPolarCoords& v);

 // additional
 float Angle();
 FPolarCoords& Rotate_Self(float theta);
 FPolarCoords& Rotate(float theta);
 
 // conversions
 FVectorR2& Cartesian();
 FVectorHR2&   Homogeneous();
};


//================================================== FCylindrical Coordinates
class FCylindricalCoords: public VIRTUAL FVector {
public:
 float& Radius;
 float& Theta;
 float& Z;

 // ctors
 FCylindricalCoords();
 FCylindricalCoords(const FVector& v);
 FCylindricalCoords(const FCylindricalCoords& v);
 FCylindricalCoords(float _radius,float _theta,float _z);

 // dtor
 virtual ~FCylindricalCoords();

 // assignment
 FCylindricalCoords& operator =(const FVector& v);
 FCylindricalCoords& operator =(const FCylindricalCoords& v);

 // optimizations over regular vectors
 float Length();

 // conversions
 FVectorR3& Cartesian();
 FVectorHR3&   Homogeneous();
};


//================================================== 3-F Homogeneous Coord FVectors
class FSphericalCoords: public VIRTUAL FVector {
public:
 float& Radius;
 float& Elevation;
 float& Azimuth;

 // ctors
 FSphericalCoords();
 FSphericalCoords(const FVector& v);
 FSphericalCoords(const FSphericalCoords& v);
 FSphericalCoords(float _r,float _e,float _a);

 // dtor
 virtual ~FSphericalCoords();

 // assignment
 FSphericalCoords& operator =(const FVector& v);
 FSphericalCoords& operator =(const FSphericalCoords& v);

 // optimizations over regular vectors
 float Length();

 // conversions
 FVectorR3& Cartesian();
 FVectorHR3&   Homogeneous();
};



//================================================================================
//================================================================================
//
//
//
inline FVectorR2::FVectorR2():FVector(2,ColV),X(ELT(0)),Y(ELT(1))
   { X=0.0; Y=0.0; }
inline FVectorR2::FVectorR2(const FVector& v):FVector(2,ColV),X(ELT(0)),Y(ELT(1))
   { X=NON(v,0); Y=NON(v,1); }
inline FVectorR2::FVectorR2(const FVectorR2& v):FVector(2,ColV),X(ELT(0)),Y(ELT(1))
   { X=v.X; Y=v.Y; }
inline FVectorR2::FVectorR2(const FVectorHR2& v):FVector(2,ColV),X(ELT(0)),Y(ELT(1))
   { X=v.X; Y=v.Y; }
inline FVectorR2::FVectorR2(float _x,float _y):FVector(2,ColV),X(ELT(0)),Y(ELT(1))
   { X=_x; Y=_y; }

inline FVectorR2::~FVectorR2() {}

inline FVectorR2& FVectorR2::operator =(const FVector& v)
   { X=v(0); Y=v(1); return (*this); }
inline FVectorR2& FVectorR2::operator =(const FVectorR2& v)
   { X=v.X; Y=v.Y; return (*this); }
inline FVectorR2& FVectorR2::operator =(const FVectorHR2& v)
   { X=v.X; Y=v.Y; return (*this); }
 
inline float FVectorR2::Length()
   { return (float)hypot(X,Y); }
inline float FVectorR2::dot(const FVectorR2& v)
   { return ((X*v.X)+(Y*v.Y)); }
inline float FVectorR2::dot(const FVectorHR2& v)
   { return ((X*v.X)+(Y*v.Y)); }

inline float FVectorR2::Angle()        { return (float)atan2(Y,X); }
inline FVectorR2& FVectorR2::Rotate_Self(float theta) {
   register float costh=(float)cos(theta); register float sinth=(float)sin(theta);
   register float tmp_x=costh*X-sinth*Y; Y=sinth*X+costh*Y; X=tmp_x;
   return (*this);
   }
inline FVectorR2& FVectorR2::Rotate(float theta)
   { return (TMP_NEW FVectorR2(*this))->Rotate_Self(theta); }
 
inline FPolarCoords& FVectorR2::FPolar()
   { return (*(TMP_NEW FPolarCoords(Length(),Angle()))); }


//================================================================================
//================================================================================
//
//
//
inline FVectorR3::FVectorR3():FVector(3,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=0.0; Y=0.0; Z=0.0; }
inline FVectorR3::FVectorR3(const FVector& v):FVector(3,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=v(0); Y=v(1); Z=v(2); }
inline FVectorR3::FVectorR3(const FVectorR3& v):FVector(3,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=v.X; Y=v.Y; Z=v.Z; }
inline FVectorR3::FVectorR3(const FVectorHR3& v):FVector(3,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=v.X; Y=v.Y; Z=v.Z; }
inline FVectorR3::FVectorR3(float _x,float _y,float _z):FVector(3,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=_x; Y=_y; Z=_z; }

inline FVectorR3::~FVectorR3() {}

inline FVectorR3& FVectorR3::operator =(const FVector& v)
   { X=v(0); Y=v(1); Z=v(2); return (*this); }
inline FVectorR3& FVectorR3::operator =(const FVectorR3& v)
   { X=v.X; Y=v.Y; Z=v.Z; return (*this); }
inline FVectorR3& FVectorR3::operator =(const FVectorHR3& v)
   { X=v.X; Y=v.Y; Z=v.Z; return (*this); }

inline float FVectorR3::Length()
   { return sqrt(X*X+Y*Y+Z*Z); }
inline float FVectorR3::dot(const FVectorR3& v)
   { return ((X*v.X)+(Y*v.Y)+(Z*v.Z)); }
inline float FVectorR3::dot(const FVectorHR3& v)
   { return ((X*v.X)+(Y*v.Y)+(Z*v.Z)); }

inline FVectorR3& FVectorR3::Rotate_Self_Around_X(float theta) {
   register float costh=(float)cos(theta); register float sinth=(float)sin(theta);
   register float tmp_y=costh*Y-sinth*Z; Z=sinth*Y+costh*Z; Y=tmp_y;
   return (*this); }

inline FVectorR3& FVectorR3::Rotate_Self_Around_Y(float theta) {
   register float costh=(float)cos(theta); register float sinth=(float)sin(theta);
   register float tmp_x=costh*X-sinth*Z; Z=sinth*X+costh*Z; X=tmp_x;
   return (*this); }

inline FVectorR3& FVectorR3::Rotate_Self_Around_Z(float theta) {
   register float costh=(float)cos(theta); register float sinth=(float)sin(theta);
   register float tmp_x=costh*X-sinth*Y; Y=sinth*X+costh*Y; X=tmp_x;
   return (*this); }

inline FVectorR3& FVectorR3::Rotate_Around_X(float theta)
   { return (TMP_NEW FVectorR3(*this))->Rotate_Self_Around_X(theta); }
inline FVectorR3& FVectorR3::Rotate_Around_Y(float theta)
   { return (TMP_NEW FVectorR3(*this))->Rotate_Self_Around_Y(theta); }
inline FVectorR3& FVectorR3::Rotate_Around_Z(float theta)
   { return (TMP_NEW FVectorR3(*this))->Rotate_Self_Around_Z(theta); }
inline FVectorR3& FVectorR3::cross(const FVectorR3& v)
   { return *(TMP_NEW FVectorR3((Y*v.Z)-(Z*v.Y),(Z*v.X)-(X*v.Z),(X*v.Y)-(Y*v.X))); }
inline FVectorR3& FVectorR3::cross(const FVectorHR3& v)
   { return *(TMP_NEW FVectorR3((Y*v.Z)-(Z*v.Y),(Z*v.X)-(X*v.Z),(X*v.Y)-(Y*v.X))); }

inline  FCylindricalCoords&   FVectorR3::FCylindrical()
   { return (*(TMP_NEW FCylindricalCoords(Length(),(float)atan2(Y,X),Z))); }
inline FSphericalCoords&   FVectorR3::FSpherical()
   { return (*(TMP_NEW FSphericalCoords(Length(),(float)atan2(sqrt(X*X+Y*Y),Z),(float)atan2(Y,X)))); }


//======================================================================
//======================================================================
//
//
//
inline FVectorHR2::FVectorHR2():FVector(3,ColV),X(ELT(0)),Y(ELT(1))
   { X=0.0; Y=0.0; ELT(2)=1.0; }
inline FVectorHR2::FVectorHR2(const FVector& v):FVector(3,ColV),X(ELT(0)),Y(ELT(1))
   { X=v(0); Y=v(1); ELT(2)=1.0; }
inline FVectorHR2::FVectorHR2(const FVectorR2& v):FVector(3,ColV),X(ELT(0)),Y(ELT(1))
   { X=v.X; Y=v.Y; ELT(2)=1.0; }
inline FVectorHR2::FVectorHR2(const FVectorHR2& v):FVector(3,ColV),X(ELT(0)),Y(ELT(1))
   { X=v.X; Y=v.Y; ELT(2)=1.0; }
inline FVectorHR2::FVectorHR2(float _x,float _y):FVector(3,ColV),X(ELT(0)),Y(ELT(1))
   { X=_x; Y=_y; ELT(2)=1.0; }

inline FVectorHR2::~FVectorHR2() {};

inline FVectorHR2& FVectorHR2::operator =(const FVector& v)
   { X=v(0); Y=v(1); return (*this); }
inline FVectorHR2& FVectorHR2::operator =(const FVectorR2& v)
   { X=v.X; Y=v.Y; return (*this); }
inline FVectorHR2& FVectorHR2::operator =(const FVectorHR2& v)
   { X=v.X; Y=v.Y; return (*this); }

inline float FVectorHR2::Length()
   { return (float)hypot(X,Y); }
inline float FVectorHR2::dot(const FVectorR3& v)
   { return ((X*v.X)+(Y*v.Y)); }
inline float FVectorHR2::dot(const FVectorHR3& v)
   { return ((X*v.X)+(Y*v.Y)); }

inline float FVectorHR2::Angle()
   { return (float)atan2(Y,X); }
inline FVectorHR2&   FVectorHR2::Rotate_Self(float theta) {
   register float costh=(float)cos(theta); register float sinth=(float)sin(theta);
   register float tmp_x=costh*X-sinth*Y; Y=sinth*X+costh*Y; X=tmp_x;
   return (*this);
   }
inline FVectorHR2&   FVectorHR2::Rotate(float theta)
   { return (TMP_NEW FVectorHR2(*this))->Rotate_Self(theta); }
 
inline FPolarCoords& FVectorHR2::FPolar()
   { return (*(TMP_NEW FPolarCoords(Length(),Angle()))); }


//======================================================================
//======================================================================
//
//
//
inline FVectorHR3::FVectorHR3():FVector(4,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=0.0; Y=0.0; Z=0.0; ELT(3)=1.0; }
inline FVectorHR3::FVectorHR3(const FVector& v):FVector(4,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=v(0); Y=v(1); Z=v(2); ELT(3)=1.0; }
inline FVectorHR3::FVectorHR3(const FVectorR3& v):FVector(4,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=v.X; Y=v.Y; Z=v.Z; ELT(3)=1.0; }
inline FVectorHR3::FVectorHR3(const FVectorHR3& v):FVector(4,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=v.X; Y=v.Y; Z=v.Z; ELT(3)=1.0; }
inline FVectorHR3::FVectorHR3(float _x,float _y,float _z):FVector(4,ColV),X(ELT(0)),Y(ELT(1)),Z(ELT(2))
   { X=_x; Y=_y; Z=_z; ELT(3)=1.0; }


// JESSE 7/1/98
//inline FVectorHR3::~FVectorHR3() {};

inline FVectorHR3& FVectorHR3::operator =(const FVector& v)
   { X=v(0); Y=v(1); Z=v(2); return (*this); }
inline FVectorHR3& FVectorHR3::operator =(const FVectorR3& v)
   { X=v.X; Y=v.Y; Z=v.Z; return (*this); }
inline FVectorHR3& FVectorHR3::operator =(const FVectorHR3& v)
   { X=v.X; Y=v.Y; Z=v.Z; return (*this); }


inline float FVectorHR3::Length()
   { return sqrt(X*X+Y*Y+Z+Z); }
inline float FVectorHR3::dot(const FVectorR3& v)
   { return ((X*v.X)+(Y*v.Y)+(Z*v.Z)); }
inline float FVectorHR3::dot(const FVectorHR3& v)
   { return ((X*v.X)+(Y*v.Y)+(Z*v.Z)); }

inline FVectorHR3&   FVectorHR3::Rotate_Self_Around_X(float theta) {
   register float costh=(float)cos(theta); register float sinth=(float)sin(theta);
   register float tmp_y=costh*Y-sinth*Z; Z=sinth*Y+costh*Z; Y=tmp_y;
   return (*this); }
inline FVectorHR3&   FVectorHR3::Rotate_Self_Around_Y(float theta) {
   register float costh=(float)cos(theta); register float sinth=(float)sin(theta);
   register float tmp_x=costh*X-sinth*Z; Z=sinth*X+costh*Z; X=tmp_x;
   return (*this); }
inline FVectorHR3&   FVectorHR3::Rotate_Self_Around_Z(float theta) {
   register float costh=(float)cos(theta); register float sinth=(float)sin(theta);
   register float tmp_x=costh*X-sinth*Y; Y=sinth*X+costh*Y; X=tmp_x;
   return (*this); }
inline FVectorHR3&   FVectorHR3::Rotate_Around_X(float theta)
   { return (TMP_NEW FVectorHR3(*this))->Rotate_Self_Around_X(theta); }
inline FVectorHR3&   FVectorHR3::Rotate_Around_Y(float theta)
   { return (TMP_NEW FVectorHR3(*this))->Rotate_Self_Around_Y(theta); }
inline FVectorHR3&   FVectorHR3::Rotate_Around_Z(float theta)
   { return (TMP_NEW FVectorHR3(*this))->Rotate_Self_Around_Z(theta); }
inline FVectorHR3&   FVectorHR3::cross(const FVectorR3& v)
   { return *(TMP_NEW FVectorHR3((Y*v.Z)-(Z*v.Y),(Z*v.X)-(X*v.Z),(X*v.Y)-(Y*v.X))); }
inline FVectorHR3&   FVectorHR3::cross(const FVectorHR3& v)
   { return *(TMP_NEW FVectorHR3((Y*v.Z)-(Z*v.Y),(Z*v.X)-(X*v.Z),(X*v.Y)-(Y*v.X))); }
 
inline FCylindricalCoords& FVectorHR3::FCylindrical()
   { return (*(TMP_NEW FCylindricalCoords(Length(),(float)atan2(Y,X),Z))); }
inline FSphericalCoords& FVectorHR3::FSpherical()
   { return (*(TMP_NEW FSphericalCoords(Length(),(float)atan2(sqrt(X*X+Y*Y),Z),(float)atan2(Y,X)))); }

//================================================================================
//================================================================================
//
//
//
inline FPolarCoords::FPolarCoords():
   FVector(2),Radius(ELT(0)),Theta(ELT(1))
   { Radius=0.0; Theta=0.0; }
inline FPolarCoords::FPolarCoords(const FVector& v):
   FVector(2),Radius(ELT(0)),Theta(ELT(1))
   { Radius=v(0); Theta=v(1); }
inline FPolarCoords::FPolarCoords(const FPolarCoords& v):
   FVector(2),Radius(ELT(0)),Theta(ELT(1))
   { Radius=v.Radius; Theta=v.Theta; }
inline FPolarCoords::FPolarCoords(float _r,float _theta):
   FVector(2),Radius(ELT(0)),Theta(ELT(1))
   { Radius=_r; Theta=_theta; }

inline FPolarCoords::~FPolarCoords() {}

inline FPolarCoords& FPolarCoords::operator =(const FVector& v)
   { Radius=v(0); Theta=v(1); return (*this); }
inline FPolarCoords& FPolarCoords::operator =(const FPolarCoords& v)
   { Radius=v.Radius; Theta=v.Theta; return (*this); }

inline float FPolarCoords::Length()
   { return Radius; }
inline float FPolarCoords::dot(const FPolarCoords& v)
   { return Radius*v.Radius*(float)cos(fabs(Theta-v.Theta)); }

inline float FPolarCoords::Angle()
   { return Theta; }
inline FPolarCoords& FPolarCoords::Rotate_Self(float theta)
   { Theta+=theta; return (*this); }
inline FPolarCoords& FPolarCoords::Rotate(float theta)
   { return (TMP_NEW FPolarCoords(*this))->Rotate_Self(theta); }
 
inline FVectorR2& FPolarCoords::Cartesian()
   { return *(TMP_NEW FVectorR2(Radius*(float)cos(Theta),Radius*(float)sin(Theta))); }
inline FVectorHR2&   FPolarCoords::Homogeneous()
   { return *(TMP_NEW FVectorHR2(Radius*(float)cos(Theta),Radius*(float)sin(Theta))); }

//================================================================================
//================================================================================
//
//
//

inline FCylindricalCoords::FCylindricalCoords():
   FVector(3),Radius(ELT(0)),Theta(ELT(1)),Z(ELT(2))
   { Radius=0.0; Theta=0.0; Z=0.0; }
inline FCylindricalCoords::FCylindricalCoords(const FVector& v):
   FVector(3),Radius(ELT(0)),Theta(ELT(1)),Z(ELT(2))
   { Radius=v(0); Theta=v(1); Z=v(2); }
inline FCylindricalCoords::FCylindricalCoords(const FCylindricalCoords& v):
   FVector(3),Radius(ELT(0)),Theta(ELT(1)),Z(ELT(2))
   { Radius=v.Radius; Theta=v.Theta; Z=v.Z; }
inline FCylindricalCoords::FCylindricalCoords(float _radius,float _theta,float _z):
   FVector(3),Radius(ELT(0)),Theta(ELT(1)),Z(ELT(2))
   { Radius=_radius; Theta=_theta; Z=_z; }

inline FCylindricalCoords::~FCylindricalCoords() {}

inline FCylindricalCoords& FCylindricalCoords::operator =(const FVector& v)
   { Radius=v(0); Theta=v(1); Z=v(2); return (*this); }
inline FCylindricalCoords& FCylindricalCoords::operator =(const FCylindricalCoords& v)
   { Radius=v.Radius; Theta=v.Theta; Z=v.Z; return (*this); }

inline float FCylindricalCoords::Length()
   { return Radius; }

inline FVectorR3& FCylindricalCoords::Cartesian()
   { return *(TMP_NEW FVectorR3(Radius*(float)cos(Theta),Radius*(float)sin(Theta),Z)); }
inline FVectorHR3&   FCylindricalCoords::Homogeneous()
   { return *(TMP_NEW FVectorHR3(Radius*(float)cos(Theta),Radius*(float)sin(Theta),Z)); }

//================================================================================
//================================================================================
//
//
//
inline FSphericalCoords::FSphericalCoords():
   FVector(3),Radius(ELT(0)),Elevation(ELT(1)),Azimuth(ELT(2))
   { Radius=0.0; Elevation=0.0; Azimuth=0.0; }
inline FSphericalCoords::FSphericalCoords(const FVector& v):
   FVector(3),Radius(ELT(0)),Elevation(ELT(1)),Azimuth(ELT(2))
   { Radius=v(0); Elevation=v(1); Azimuth=v(2); }
inline FSphericalCoords::FSphericalCoords(const FSphericalCoords& v):
   FVector(3),Radius(ELT(0)),Elevation(ELT(1)),Azimuth(ELT(2))
   { Radius=v.Radius; Elevation=v.Elevation; Azimuth=v.Azimuth; }
inline FSphericalCoords::FSphericalCoords(float _r,float _e,float _a):
   FVector(3),Radius(ELT(0)),Elevation(ELT(1)),Azimuth(ELT(2))
   { Radius=_r; Elevation=_e; Azimuth=_a; }

inline FSphericalCoords::~FSphericalCoords() {}

inline FSphericalCoords& FSphericalCoords::operator =(const FVector& v)
   { Radius=v(0); Elevation=v(1); Azimuth=v(2); return (*this); }
inline FSphericalCoords& FSphericalCoords::operator =(const FSphericalCoords& v)
   { Radius=v.Radius; Elevation=v.Elevation; Azimuth=v.Azimuth; return (*this); }

inline float FSphericalCoords::Length()
   { return Radius; }

inline FVectorR3& FSphericalCoords::Cartesian()
   { return *(TMP_NEW FVectorR3(
                              Radius*(float)sin(Elevation)*(float)cos(Azimuth),
                              Radius*(float)sin(Elevation)*(float)sin(Azimuth),
                              Radius*(float)cos(Elevation))); }
inline FVectorHR3&   FSphericalCoords::Homogeneous()
   { return *(TMP_NEW FVectorHR3(
                              Radius*(float)sin(Elevation)*(float)cos(Azimuth),
                              Radius*(float)sin(Elevation)*(float)sin(Azimuth),
                              Radius*(float)cos(Elevation))); }

#undef TMP_NEW
#undef ELT
#undef NON
#undef VIRTUAL 

//======================================================================
//
// Default type establishment
//
#ifdef __Concat2
#undef __Concat2
#endif
#ifdef _Concat2
#undef _Concat2
#endif
#define __Concat2(a,b) a##b
#define _Concat2(a,b) __Concat2(a,b)
typedef _Concat2(DEF_PREC_VECTOR_PREFIX,VectorR2) VectorR2;
typedef _Concat2(DEF_PREC_VECTOR_PREFIX,VectorR3) VectorR3;
typedef _Concat2(DEF_PREC_VECTOR_PREFIX,VectorHR2) VectorHR2;
typedef _Concat2(DEF_PREC_VECTOR_PREFIX,VectorHR3) VectorHR3;
typedef _Concat2(DEF_PREC_VECTOR_PREFIX,PolarCoords) PolarCoords;
typedef _Concat2(DEF_PREC_VECTOR_PREFIX,SphericalCoords) SphericalCoords;
typedef _Concat2(DEF_PREC_VECTOR_PREFIX,CylindricalCoords) CylindricalCoords;

#undef __Concat2
#undef _Concat2

#endif

