/***********************************************************************

 DHMatrix.c
 Denavit-Hartenberg Matrix Class - DMatrix and FMatrix forms - Source Code
 Copyright (C) 1993
        Lyle Hall 1993/94
        Eric C. Welton 1993/94
        Beckman Institute - Urbana, Illinois
        For Neuronal Pattern Analysis Group and Hexapod Project

        These libraries are free and may be redistributed and/or modified,
        provided the above notice and this paragraph are included.
        These libraries may not be sold and are distributed without
        any warranty or implication of fitness for a purpose.


***********************************************************************/

#include <math.h>
#include <DHMatrix.h>

//======================================================================
DHDMatrix& DHDMatrix::SetValues() {
 // use the current settings to fill in the matrix
 if(inverted==0) {
	(*this)(0,0)=costh;             (*this)(0,1)=-1.0*cosal*sinth;
	(*this)(1,0)=sinth;             (*this)(1,1)=cosal*costh;
											  (*this)(2,1)=sinal;
	(*this)(0,2)=sinal*sinth;       (*this)(0,3)=a*costh;
	(*this)(1,2)=-1.0*sinal*costh;  (*this)(1,3)=a*sinth;
	(*this)(2,2)=cosal;             (*this)(2,3)=d;
                                   (*this)(3,3)=1.0;
	}
 else {
   (*this)(0,0)=costh;             (*this)(0,1)=sinth;
	(*this)(1,0)=-1.0*cosal*sinth;  (*this)(1,1)=cosal*costh;
	(*this)(2,0)=sinal*sinth;       (*this)(2,1)=-1.0*sinal*costh;
	(*this)(0,2)=0.0;               (*this)(0,3)=-1.0*a;
	(*this)(1,2)=sinal;             (*this)(1,3)=-1.0*d*sinal;
	(*this)(2,2)=cosal;             (*this)(2,3)=-1.0*d*cosal;
                                   (*this)(3,3)=1.0;
	}
 return (*this);
}

//======================================================================
DHDMatrix::DHDMatrix(   double _alpha,
                           double _a,
                           double _d,
                           double _theta):DMatrix(4,4) {

 inverted=0;
 theta=_theta; alpha=_alpha; a=_a; d=_d;
 sinth=sin(theta);
 costh=cos(theta);
 sinal=sin(alpha);
 cosal=cos(alpha);
 SetValues();
}

//======================================================================
DHDMatrix::DHDMatrix(const DHDMatrix& dh):DMatrix(dh)
{
 theta=dh.theta;
 alpha=dh.alpha;
 a=dh.a;
 d=dh.d;
 sinth=dh.sinth;
 costh=dh.costh;
 sinal=dh.sinal;
 cosal=dh.cosal;
 inverted=dh.inverted;
}


//======================================================================
DHDMatrix& DHDMatrix::set_theta(double _theta) {
 theta=_theta;
 sinth=sin(theta);
 costh=cos(theta);
 return SetValues();
}

//======================================================================
DHDMatrix& DHDMatrix::set_alpha(double _alpha) {
 alpha=_alpha;
 sinal=sin(alpha);
 cosal=cos(alpha);
 return SetValues();
}

//======================================================================
DHDMatrix& DHDMatrix::set_a(double _a) { a=_a; return SetValues(); }
DHDMatrix& DHDMatrix::set_d(double _d) { d=_d; return SetValues(); }

//======================================================================
DHDMatrix DHDMatrix::inv() {
 DHDMatrix	rettmp(*this);
 if(rettmp.inverted==0)
	rettmp.inverted=1;
 else
	rettmp.inverted=0;
 rettmp.SetValues();
 return DHDMatrix(rettmp);
}

//======================================================================
void DHDMatrix::print_values(FILE* f) {
 char* invflag="false";
 if(inverted==1) invflag="true";
   fprintf(f,"(th=%8.4g, al=%8.4g, a=%8.4g, d=%8.4g, inv=%s)",(double)theta,(double)alpha,(double)a,(double)d,invflag);
   }



//================================================================================
//================================================================================
//
// Implementation
//
//


//======================================================================
DHFMatrix& DHFMatrix::SetValues() {
 // use the current settings to fill in the matrix
 if(inverted==0) {
	(*this)(0,0)=costh;             (*this)(0,1)=-1.0*cosal*sinth;
	(*this)(1,0)=sinth;             (*this)(1,1)=cosal*costh;
											  (*this)(2,1)=sinal;
	(*this)(0,2)=sinal*sinth;       (*this)(0,3)=a*costh;
	(*this)(1,2)=-1.0*sinal*costh;  (*this)(1,3)=a*sinth;
	(*this)(2,2)=cosal;             (*this)(2,3)=d;
                                   (*this)(3,3)=1.0;
	}
 else {
   (*this)(0,0)=costh;             (*this)(0,1)=sinth;
	(*this)(1,0)=-1.0*cosal*sinth;  (*this)(1,1)=cosal*costh;
	(*this)(2,0)=sinal*sinth;       (*this)(2,1)=-1.0*sinal*costh;
	(*this)(0,2)=0.0;               (*this)(0,3)=-1.0*a;
	(*this)(1,2)=sinal;             (*this)(1,3)=-1.0*d*sinal;
	(*this)(2,2)=cosal;             (*this)(2,3)=-1.0*d*cosal;
                                   (*this)(3,3)=1.0;
	}
 return (*this);
}

DHFMatrix::DHFMatrix(   float _alpha,
                           float _a,
                           float _d,
                           float _theta):FMatrix(4,4) {
 inverted=0;
 theta=_theta; alpha=_alpha; a=_a; d=_d;
 sinth=sin(theta);
 costh=cos(theta);
 sinal=sin(alpha);
 cosal=cos(alpha);
 SetValues();
}

//======================================================================
DHFMatrix::DHFMatrix(const DHFMatrix& dh):FMatrix(dh)
{
 theta=dh.theta;
 alpha=dh.alpha;
 a=dh.a;
 d=dh.d;
 sinth=dh.sinth;
 costh=dh.costh;
 sinal=dh.sinal;
 cosal=dh.cosal;
 inverted=dh.inverted;
}

DHFMatrix& DHFMatrix::set_theta(float _theta) {
 theta=_theta;
 sinth=sin(theta);
 costh=cos(theta);
 return SetValues();
}

DHFMatrix& DHFMatrix::set_alpha(float _alpha) {
 alpha=_alpha;
 sinal=sin(alpha);
 cosal=cos(alpha);
 return SetValues();
}

DHFMatrix& DHFMatrix::set_a(float _a)  { a=_a; return SetValues(); }
DHFMatrix& DHFMatrix::set_d(float _d)  { d=_d; return SetValues(); }

DHFMatrix DHFMatrix::inv() {
 DHFMatrix	rettmp(*this);
 if(rettmp.inverted==0)
	rettmp.inverted=1;
 else
	rettmp.inverted=0;
 rettmp.SetValues();
 return DHFMatrix(rettmp);
}

void DHFMatrix::print_values(FILE* f) {
 char* invflag="false";
 if(inverted==1) invflag="true";
   fprintf(f,"(th=%8.4g, al=%8.4g, a=%8.4g, d=%8.4g, inv=%s)",(double)theta,(double)alpha,(double)a,(double)d,invflag);
   }

