/***********************************************************************

 DHMatrix.h
 Denavit-Hartenberg Matrix Class - DMatrix and FMatrix forms
 Copyright (C) 1993
        Lyle Hall 1993/94
        Eric C. Welton 1993/94
        Beckman Institute - Urbana, Illinois
        For Neuronal Pattern Analysis Group and Hexapod Project

        These libraries are free and may be redistributed and/or modified,
        provided the above notice and this paragraph are included.
        These libraries may not be sold and are distributed without
        any warranty or implication of fitness for a purpose.


***********************************************************************/

#ifndef DHMATRIX_HDR
#define DHMATRIX_HDR

#include <DMatrix.h>
#include <FMatrix.h>

class DHDMatrix:public DMatrix {

 int	  inverted;					// 0 is not inverted, 1=inverted

 double theta,alpha,a,d;      // the dh parameters
 double sinth,costh;          // only computed when angles change - for speed
 double sinal,cosal;          // purposes


 DHDMatrix& SetValues();            // use current parameter settings to fill in
 DHDMatrix& SetInverseValues();  // the rest of the matrix and/or it's inverse

public:

 //================================================== Constructors
 DHDMatrix(const DHDMatrix& dh);
 DHDMatrix(double _alpha=0,double _a=0,double _d=0,double _theta=0);

 //================================================== Parameter Settings
 DHDMatrix& set_theta(double _theta);
 DHDMatrix& set_alpha(double _alpha);
 DHDMatrix& set_a(double _a);
 DHDMatrix& set_d(double _d);
 DHDMatrix& set_theta(float _theta) { return set_theta((double)_theta); }
 DHDMatrix& set_alpha(float _alpha) { return set_alpha((double)_alpha); }
 DHDMatrix& set_a(float _a) { return set_a((double)_a); }
 DHDMatrix& set_d(float _d) { return set_d((double)_d); }


 //================================================== Parameter Offsets
 DHDMatrix& add_theta(double _theta);
 DHDMatrix& add_alpha(double _alpha);
 DHDMatrix& add_a(double _a);
 DHDMatrix& add_d(double _d);
 DHDMatrix& add_theta(float _theta) { return set_theta((double)_theta); }
 DHDMatrix& add_alpha(float _alpha) { return set_alpha((double)_alpha); }
 DHDMatrix& add_a(float _a) { return set_a((double)_a); }
 DHDMatrix& add_d(float _d) { return set_d((double)_d); }

 //================================================== Operations
 DHDMatrix inv();   // matrix inversion
 DHDMatrix& operator =(const DHDMatrix& dh)
   {	this->DMatrix::operator=((DMatrix&)dh);
		inverted=dh.inverted;
		theta=dh.theta; alpha=dh.alpha; a=dh.a; d=dh.d;
		sinth=dh.sinth; costh=dh.costh; sinal=dh.sinal; cosal=dh.cosal;
	return (*this); }
 DHDMatrix& operator =(const DMatrix& m)
   { this->DMatrix::operator=(m); return (*this); }

 //================================================== IO
 void print_values(FILE* f=stdout);

 };


class DHFMatrix:public FMatrix {

 int	  inverted;					// 0 is not inverted, 1=inverted

 float theta,alpha,a,d;    // the dh parameters
 float sinth,costh;           // only computed when angles change - for speed
 float sinal,cosal;           // purposes


 DHFMatrix& SetValues();            // use current parameter settings to fill in
 DHFMatrix& SetInverseValues();  // the rest of the matrix and/or it's inverse

public:

 //================================================== Constructors
 DHFMatrix(const DHFMatrix& dh);
 DHFMatrix(float _alpha=0,float _a=0,float _d=0,float _theta=0);

 //================================================== Parameter Settings
 DHFMatrix& set_theta(float _theta);
 DHFMatrix& set_alpha(float _alpha);
 DHFMatrix& set_a(float _a);
 DHFMatrix& set_d(float _d);
 DHFMatrix& set_theta(double _theta) { return set_theta((float)_theta); }
 DHFMatrix& set_alpha(double _alpha) { return set_alpha((float)_alpha); }
 DHFMatrix& set_a(double _a) { return set_a((float)_a); }
 DHFMatrix& set_d(double _d) { return set_d((float)_d); }


 //================================================== Parameter Offsets
 DHFMatrix& add_theta(float _theta);
 DHFMatrix& add_alpha(float _alpha);
 DHFMatrix& add_a(float _a);
 DHFMatrix& add_d(float _d);
 DHFMatrix& add_theta(double _theta) { return set_theta((float)_theta); }
 DHFMatrix& add_alpha(double _alpha) { return set_alpha((float)_alpha); }
 DHFMatrix& add_a(double _a) { return set_a((float)_a); }
 DHFMatrix& add_d(double _d) { return set_d((float)_d); }

 //================================================== Operations
 DHFMatrix inv();   // matrix inversion
 DHFMatrix& operator =(const DHFMatrix& dh)
   {	this->FMatrix::operator=((FMatrix&)dh);
		inverted=dh.inverted;
		theta=dh.theta; alpha=dh.alpha; a=dh.a; d=dh.d;
		sinth=dh.sinth; costh=dh.costh; sinal=dh.sinal; cosal=dh.cosal;
	return (*this); }
 DHFMatrix& operator =(const FMatrix& m)
   { this->FMatrix::operator=(m); return (*this); }

 //================================================== IO
 void print_values(FILE* f=stdout);

 };

#ifdef __Concat3
#undef __Concat3
#endif
#ifdef _Concat3
#undef _Concat3
#endif
#define __Concat3(a,b,c) a##b##c
#define _Concat3(a,b,c) __Concat3(a,b,c)

typedef _Concat3(DH,DEF_PREC_MATRIX_PREFIX,Matrix) DHMatrix;

#endif

