/***********************************************************************

 MatrixT.h
 MatrixT Class
 Copyright (C) 1993
        Lyle Hall 1993/94
        Eric C. Welton 1993/94
        Beckman Institute - Urbana, Illinois
        For Neuronal Pattern Analysis Group and Hexapod Project

        These libraries are free and may be redistributed and/or modified,
        provided the above notice and this paragraph are included.
        These libraries may not be sold and are distributed without
        any warranty or implication of fitness for a purpose.

 General Information

        The MatrixT library provides fundamental matrix operations.

***********************************************************************/

#ifndef MATRIX_HDR
#define MATRIX_HDR

#include <stdio.h>

//======================================================================
//
// Handle default debugging flags
//
#ifndef DEBUG_SINK
#define DEBUG_SINK stderr
#endif


//======================================================================
//
// Handle default subscription flag
//
#ifndef SUBSCRIPTION_OFFSET
#define SUBSCRIPTION_OFFSET 0
#endif

//======================================================================
//
// Handle default precision selection
//
//    1) if DEFAULT_PRECISION is not defined, set it to 'float'
//    2) if DEFAULT_PRECISION is defined check to see if it is
//       'float':    set DEF_PREC_MATRIX_NAME to 'FMatrix'
//                   set DEF_PREC_MATRIX_PREFIX to 'F'
//       'double':   set DEF_PREC_MATRIX_NAME to 'DMatrix'
//                   set DEF_PREC_MATRIX_PREFIX to 'D'
//    3) correctly redefine the type default type names
//    4) if the concatentation macros have not been defined,
//       then define them
//
#ifndef DEFAULT_PRECISION
#define DEFAULT_PRECISION float
#endif

#if DEFAULT_PRECISION == float
#define DEF_PREC_MATRIX_NAME FMatrix
#define DEF_PREC_MATRIX_PREFIX F
#elif DEFAULT_PRECISION == double
#define DEF_PREC_MATRIX_NAME DMatrix
#define DEF_PREC_MATRIX_PREFIX D
#else
#error DEFAULT_PRECISION is not defined correctly, must be double or float.
#endif

#ifdef __Concat2
#undef __Concat2
#endif
#ifdef _Concat2
#undef _Concat2
#endif
#define __Concat2(a,b) a##b
#define _Concat2(a,b) __Concat2(a,b)

#ifdef __Concat3
#undef __Concat3
#endif
#ifdef _Concat3
#undef _Concat3
#endif
#define __Concat3(a,b,c) a##b##c
#define _Concat3(a,b,c) __Concat3(a,b,c)


//======================================================================
//
// When the __DONT_DEFINE_MATRIX_TEMPLATE is undefined (e.g. the
// default situation) the following macros are defined.  These
// cause this file to implement the VectorT<EltType> template
//
//       TEMPLATE    template<class EltType>
//       TYPE        EltType
//       VCLASSNAME  VectorT
//       MCLASSNAME  MatrixT
//       VECT        VCLASSNAME<TYPE>
//       MAT         MCLASSNAME<TYPE>
//
// If these are defined differently it is possible to
// implement a non-template version of the Vector data type.
// This is done, for example, by the DVector header file, which
// defines the above as follows (and sets the __DONT... flag)
//
//       TEMPLATE
//       TYPE        double
//       VCLASSNAME  DVector
//       MCLASSNAME  DMatrix
//       VECT        DVector
//       MAT         DMatrix
//

#ifndef __DONT_DEFINE_MATRIX_TEMPLATE
#pragma interface
#define TEMPLATE template<class EltType>
#define TYPE EltType
#define VCLASSNAME VectorT
#define MCLASSNAME MatrixT
#define MAT MCLASSNAME<TYPE>
#define VECT VCLASSNAME<TYPE>
#else
#ifndef TEMPLATE
#define TEMPLATE
#endif
#ifndef MAT
#define MAT MCLASSNAME
#endif
#ifndef VECT
#define VECT VCLASSNAME
#endif
#endif

// If the header has not been included, prepare a forward reference
#ifndef VECTOR_HDR
TEMPLATE class VCLASSNAME;
#endif


// this is used where a reference could also be used - in the
// future the template might be expanded to facilitate non
// basic types, or types with expensive copy functions - in
// this case we want to use references to pass arguments.
#define TYPER TYPE


//======================================================================
//======================================================================
//======================================================================
//
// The actual class definition
//

TEMPLATE
class MCLASSNAME {
protected:
 int     width;
 int     height;
 TYPE*   data;
 friend class VECT;
 TYPE& error_return_value() const { static TYPE e=0.0; return e; }
 MCLASSNAME(TYPE* vals,int h,int w);
public:

 //================================================== Constructors
 MCLASSNAME();
 MCLASSNAME(int h,int w);
 MCLASSNAME(int h,int w,const TYPE* v);
 MCLASSNAME(const MAT& m);
 MCLASSNAME(const VECT& v);

 //================================================== Destructor
 virtual ~MCLASSNAME();

 //================================================== Information
 // ** Information (where p is precision, e.g.
 //                 assume zero if a(i,j) < p
 int columns() const            { return width; }
 int rows() const               { return height; }

 TYPE& Min(int* i=0,int* j=0) const;    // return Minimum elt
 TYPE& AbsMin(int* i=0,int* j=0) const; // return Minimum elt, absolute value
 TYPE& Max(int* i=0,int* j=0) const;    // return Maximum elt
 TYPE& AbsMax(int* i=0,int* j=0) const; // return Maximum elt, absolute value

 int NumZeros(const TYPER p=0) const;        // return number of elts <= p

 //================================================== Predicates
 // ** Predicates (where p is precision, e.g.
 //                assume zero if a(i,j) < p
 int iszero(const TYPER p=0) const;          // true if this is zero

 //================================================== Element Access
 // ** Element access
 TYPE&   operator()( int i, int j);
 TYPE&   operator()( int i, int j) const;
 VECT    row(int i) const;       // return row i
 VECT    col(int j) const;       // return col j
 VECT    diag(int i=0) const;    // return diag i
 MAT     sub_matrix( int ulx,    // return sub matrix
                     int uly,
                     int lrx,
                     int lry) const;

 //================================================== General Matrix Operations
 MAT& init(const TYPE* val);              // init from array
 MAT& fill(const TYPER val);              // set all elts = val
 MAT& abs();                           // take absolute value of each elt
 MAT& reshape(int i,int j);       // resize matrix to i x j
 MAT&  replace_column(int j,const VECT& v);// replace column j with vector v
 MAT&  replace_row(int i,const VECT& v);  // replace row i with vector v
 VECT  extract_column(int j);       // remove column and resize to m x n-1
 VECT  extract_row(int i);          // remove row and resize to m-1 x n

 //================================================== Comparison Operators
 int operator !=(const MAT& m) const;    // true if this(i,j) != a(i,j)
 int operator ==(const MAT& m) const;    // true if this(i,j) == a(i,j)
 int operator >=(const MAT& m) const;    // true if this(i,j) >= a(i,j)
 int operator <=(const MAT& m) const;    // true if this(i,j) <= a(i,j)
 int operator >(const MAT&  m) const;    // true if this(i,j)  > a(i,j)
 int operator <(const MAT&  m) const;    // true if this(i,j)  < a(i,j)

 //================================================== Assignment Operators
 MAT& operator =(const MAT& m);           // simple assignment
 operator double() const;           // assigns s to m[0] ( allows 1x1 mtx-scalar cnvrt)
 operator long() const;             // assigns s to m[0] ( allows 1x1 mtx-scalar cnvrt)
 MAT& operator +=(const MAT& m);       // this <- this + m
 MAT& operator -=(const MAT& m);       // this <- this - m
 MAT& operator *=(const TYPER s);         // this <- this * m, eltwise scalar
 MAT& operator /=(const TYPER s);         // this <- this / m, eltwise scalar

 //================================================== Matrix Operators
 MAT operator !() const;            // ret <- this.inverse
 MAT inv() const;                   // ret <- this.inverse
 MAT& inv_self();                   // this <- this.inverse
 MAT operator ~() const;            // ret <- this.transpose
 MAT transpose() const;             // ret <- this.transpose
 MAT& transpose_self();             // this <- this.transpose
 MAT operator +(const MAT& m)  const;     // ret <- this + m
 MAT operator -()  const;           // ret <- this * -1 
 MAT operator -(const MAT& m)  const;     // ret <- this - m
 MAT operator /(const TYPER s)  const;    // ret <- this / s, eltwise scalar
 MAT operator *(const TYPER s)  const;    // ret <- this * s, eltwise scalar
 MAT operator *(const VECT& v) const;     // ret <- this * v, matrix mult
 MAT operator *(const MAT& m)  const;     // ret <- this * m, matrix mult
friend MAT operator *(const TYPER s,const MAT& m); // ret <- m * s, eltwise scalar
friend MAT operator *(const VECT& v,const MAT& m); // ret <- v * m, matrix mult

 //================================================== Matrix IO
 void print(FILE* f=stdout) const;     // print an ascii version of the matrix
 int  read(FILE* f=stdin);             // read an ascii version of the matrix
 int  save(FILE* f,const char* fmt) const;  // save to f in format fmt
 int  save(const char* fname,const char* fmt) const;// create fname and save, use format fmt
 int  load(FILE* f,const char* fmt);        // load from f, expect format fmt
 int  load(const char* fname,const char* fmt);   // open fname and load, expect format fmt
 };

#ifndef __DONT_DEFINE_VECTOR_TEMPLATE
#undef TYPE
#undef TYPER
#undef VCLASSNAME
#undef MCLASSNAME
#endif
#undef TEMPLATE
#undef MAT
#undef VECT

//======================================================================
//
// When the user is defining a template type we must include the
// implementation file.  If this header is not defining a template,
// then we don't want to do that - we want the implementation
// to be placed in a source file and turned into an object file,
// so we ignore the implemenation here.
//
#ifndef __DONT_DEFINE_MATRIX_TEMPLATE
#include "Matrix_Imp.h"
#endif

#endif
