/***********************************************************************

 VectorT.h
 VectorT Class Header File
 Copyright (C) 1993
        Lyle Hall 1993/94
        Eric C. Welton 1993/94
        Beckman Institute - Urbana, Illinois
        For Neuronal Pattern Analysis Group and Hexapod Project

        These libraries are free and may be redistributed and/or modified,
        provided the above notice and this paragraph are included.
        These libraries may not be sold and are distributed without
        any warranty or implication of fitness for a purpose.

 General Information

        The VectorT library provides fundamental vector operations.

***********************************************************************/

#ifndef VECTOR_HDR 
#define VECTOR_HDR 

#include <stdio.h>


//======================================================================
//
// Handle default debugging flags
//
#ifndef DEBUG_SINK
#define DEBUG_SINK stderr
#endif

//======================================================================
//
// Handle default subscription flag
//
#ifndef SUBSCRIPTION_OFFSET
#define SUBSCRIPTION_OFFSET 0
#endif

//======================================================================
//
// Handle default precision selection
//
//    1) if DEFAULT_PRECISION is not defined, set it to 'float'
//    2) if DEFAULT_PRECISION is defined check to see if it is
//       'float':    set DEF_PREC_VECTOR_NAME to 'FVector'
//                   set DEF_PREC_VECTOR_PREFIX to 'F'
//       'double':   set DEF_PREC_VECTOR_NAME to 'DVector'
//                   set DEF_PREC_VECTOR_PREFIX to 'D'
//    3) correctly redefine the type default type names
//    4) if the concatentation macros have not been defined,
//       then define them
//
#ifndef DEFAULT_PRECISION
#define DEFAULT_PRECISION float
#endif

#if DEFAULT_PRECISION == float
#define DEF_PREC_VECTOR_NAME FVector
#define DEF_PREC_VECTOR_PREFIX F
#elif DEFAULT_PRECISION == double
#define DEF_PREC_VECTOR_NAME DVector
#define DEF_PREC_VECTOR_PREFIX D
#else
#error DEFAULT_PRECISION is not defined correctly, must be double or float.
#endif

#ifdef __Concat2
#undef __Concat2
#endif
#ifdef _Concat2
#undef _Concat2
#endif
#define __Concat2(a,b) a##b
#define _Concat2(a,b) __Concat2(a,b)

#ifdef __Concat3
#undef __Concat3
#endif
#ifdef _Concat3
#undef _Concat3
#endif
#define __Concat3(a,b,c) a##b##c
#define _Concat3(a,b,c) __Concat3(a,b,c)

//======================================================================
//
// When the __DONT_DEFINE_VECTOR_TEMPLATE is undefined (e.g. the
// default situation) the following macros are defined.  These
// cause this file to implement the VectorT<EltType> template
//
//       TEMPLATE    template<class EltType>
//       TYPE        EltType
//       VCLASSNAME  VectorT
//       MCLASSNAME  MatrixT
//       VECT        VCLASSNAME<TYPE>
//       MAT         MCLASSNAME<TYPE>
//
// If these are defined differently it is possible to
// implement a non-template version of the Vector data type.
// This is done, for example, by the DVector header file, which
// defines the above as follows (and sets the __DONT... flag)
//
//       TEMPLATE    
//       TYPE        double
//       VCLASSNAME  DVector
//       MCLASSNAME  DMatrix
//       VECT        DVector
//       MAT         DMatrix
//

#ifndef __DONT_DEFINE_VECTOR_TEMPLATE
#pragma interface
#define TEMPLATE template<class EltType>
#define TYPE EltType
#define VCLASSNAME VectorT
#define MCLASSNAME MatrixT
#define MAT MCLASSNAME<TYPE>
#define VECT VCLASSNAME<TYPE>
#else
#ifndef TEMPLATE
#define TEMPLATE
#endif
#ifndef MAT
#define MAT MCLASSNAME
#endif
#ifndef VECT
#define VECT VCLASSNAME
#endif
#endif

// If the header has not been included, prepare a forward reference
#ifndef MATRIX_HDR 
TEMPLATE class MCLASSNAME;
#endif


// this is used where a reference could also be used - in the
// future the template might be expanded to facilitate non
// basic types, or types with expensive copy functions - in
// this case we want to use references to pass arguments.
#define TYPER TYPE

//============================================================
//
// The VectorType enum must be defined only once, and since
// this header may be included multiple times we need a
// separate #define protection flag to ensure thess lines only
// get compiled once.
//
#ifndef VECTOR_TYPE_ENUM
#define VECTOR_TYPE_ENUM
enum VectorType { RowV, ColV };
#define Vector  DEF_PREC_VECTOR_NAME
#endif

//======================================================================
//======================================================================
//======================================================================
//
// The actual class definition
//

TEMPLATE
class VCLASSNAME {
protected:
 VectorType _type;
 int        numelts;
 TYPE*         data;
 friend class MAT;
 TYPE& error_return_value() const { static TYPE e=0.0; return e; }
 VCLASSNAME(TYPE* newvals,int sz,VectorType t);
public:

 //================================================== Constructors
 VCLASSNAME();
 VCLASSNAME(int sz,VectorType v=RowV);
 VCLASSNAME(int sz,const TYPE* vals,VectorType v=RowV);
 VCLASSNAME(const MAT& v);
 VCLASSNAME(const VECT& v);

 //================================================== Destructor
 virtual ~VCLASSNAME();

 //================================================== Information
 // ** Information (where p is precision, e.g.
 //                 assume zero if a(i,j) < p
 int size() const          { return numelts; };

// 7/1/98 TEST
 VectorType type() { return _type; }
//VectorType type() const   { return _type; }

 TYPE& Min(int* i=0) const;          // ret Min elt
 TYPE& AbsMin(int* i=0) const;       // ret Min elt, absolute value
 TYPE& Max(int* i=0) const;          // ret Max elt
 TYPE& AbsMax(int* i=0) const;       // ret Max elt, absolute value

 double PNorm(double p) const;       // return the p-norm
 double Length() const;              // return 2-norm
 int NumZeros(const TYPER p=0) const;// return number of elts <= p

 //================================================== Predicates
 // ** Predicates (where p is precision, e.g.
 //                assume zero if a(i,j) < p
 int iszero(const TYPER p=0) const;  // true if this is zero

 //================================================== Element Access
 // ** Element access
 TYPE& operator()(int i);
 TYPE&   operator()(int i) const;
 VECT sub_vector( int start, int stop) const;   // return sub vector

 //================================================== General Vector Operations
 VECT& init(const TYPE* val);       // init from array
 VECT& fill(const TYPER val);       // set all elts = val
 VECT& abs();                       // take absolute value of each elt
 VECT& extract_elt(int j);          // remove column and resize to n-1

 //================================================== Comparison Operators
 int operator !=(const VECT& v) const;    // true if this(i,j) != a(i,j)
 int operator ==(const VECT& v) const;    // true if this(i,j) == a(i,j)
 int operator >=(const VECT& v) const;    // true if this(i,j) >= a(i,j)
 int operator <=(const VECT& v) const;    // true if this(i,j) <= a(i,j)
 int operator >(const VECT&  v) const;    // true if this(i,j)  > a(i,j)
 int operator <(const VECT&  v) const;    // true if this(i,j)  < a(i,j)

 //================================================== Assignment Operators
 VECT& operator =(const VECT& v);    // simple assignment
 VECT& operator +=(const VECT& v);   // this <- this + m
 VECT& operator -=(const VECT& v);   // this <- this - m
 VECT& operator *=(const TYPER s);   // this <- this * m, eltwise scalar
 VECT& operator /=(const TYPER s);   // this <- this / m, eltwise scalar
 operator double() const;            // assigns s to m[0] ( allows 1x1 mtx-scalar cnvrt)
 operator long() const;              // assigns s to m[0] ( allows 1x1 mtx-scalar cnvrt)

 //================================================== Vector Operators
 VECT operator ~() const;            // ret <- this.transpose
 VECT transpose() const;             // ret <- this.transpose
 VECT& transpose_self();             // this <- this.transpose
 VECT operator +(const VECT& m)  const;   // ret <- this + m
 VECT operator -()  const;                // ret <- this * -1
 VECT operator -(const VECT& m)  const;   // ret <- this - m
 VECT operator /(const TYPER s)  const;   // ret <- this / s, eltwise scalar
 VECT operator *(const TYPER s)  const;   // ret <- this * s, eltwise scalar
 MAT  operator *(const VECT& v) const;    // ret <- this * v, matrix mult
 MAT  operator *(const MAT& m)  const;    // ret <- this * m, matrix mult
friend VECT operator *(const TYPER s,const VECT& m);  // ret <- m * s, eltwise scalar
friend MAT operator *(const MAT& v,const VECT& m);  // ret <- v * m, matrix mult
 double  dot(const VECT& v) const;
 double  inner_product(const VECT& v) const;
 MAT     outer_product(const VECT& v) const;
 VECT    normalize() const;
 VECT&   normalize_self();

 //================================================== Vector IO
 void print(FILE* f=stdout) const;     // print an ascii version of the matrix
 int  read(FILE* f=stdin);             // read an ascii version of the matrix
 int  save(FILE* f,const char* fmt=0) const;  // save to f in format fmt
 int  save(const char* fname,const char* fmt=0) const;// create fname and save, use format fmt
 int  load(FILE* f,const char* fmt=0);        // load from f, expect format fmt
 int  load(const char* fname,const char* fmt=0);   // open fname and load, expect format fmt

 };

#ifndef __DONT_DEFINE_VECTOR_TEMPLATE
#undef TYPE
#undef TYPER
#undef VCLASSNAME
#undef MCLASSNAME
#endif
#undef TEMPLATE
#undef MAT
#undef VECT

//======================================================================
//
// When the user is defining a template type we must include the
// implementation file.  If this header is not defining a template,
// then we don't want to do that - we want the implementation
// to be placed in a source file and turned into an object file,
// so we ignore the implemenation here.
//
#ifndef __DONT_DEFINE_VECTOR_TEMPLATE
#include "Vector_Imp.h"
#endif

#endif
