/***********************************************************************

 VectorT_Imp.h
 VectorT Class Member Function Definitions Implementation
 Copyright (C) 1993
        Lyle Hall 1993/94
        Eric C. Welton 1993/94
        Beckman Institute - Urbana, Illinois
        For Neuronal Pattern Analysis Group and Hexapod Project

        These libraries are free and may be redistributed and/or modified,
        provided the above notice and this paragraph are included.
        These libraries may not be sold and are distributed without
        any warranty or implication of fitness for a purpose.

 General Information

        The VectorT library provides fundamental matrix operations.

***********************************************************************/

#ifndef VECTOR_IMPLEMENTATION_HDR
#define VECTOR_IMPLEMENTATION_HDR

#include <VectorT.h>
#include <math.h>

//======================================================================
//
// #define flag status at this point should be:
//
// DEBUG_SINK              correct from Vector.h
// SUBSCRIPTION_OFFSET     correct from Vector.h
// DEFAULT_PRECISION       correct from Vector.h ('float' or 'double')
// DEF_PREC_VECTOR_NAME    correct from Vector.h ('DVector' or 'FVector')
// DEF_PREC_VECTOR_PREFIX  correct from Vector.h ('D' or 'F')
// __Concat2               correct from Vector.h
// _Concat2                correct from Vector.h
// __Concat3               correct from Vector.h
// _Concat3                correct from Vector.h
//
// __DONT_DEFINE_VECTOR_TEMPLATE should be defined by any user
//                               or subclassing files prior to
//                               the inclusion of this file.
// 
// TEMPLATE                undef, will def here is DONT... flag undef
// ELT                     undef, will def here is DONT... flag undef
// VCLASSNAME              undef, will def here is DONT... flag undef
// MCLASSNAME              undef, will def here is DONT... flag undef
// VECT                    undef, will def here is DONT... flag undef
// MAT                     undef, will def here is DONT... flag undef
// TYPER                   undef, need to define here
//
// VECTOR_TYPE_ENUM        correct from Vector.h
//


// determine whether this is building a header or object source
#ifndef INLINE
#define INLINE inline
#endif

//======================================================================
//
// When the __DONT_DEFINE_VECTOR_TEMPLATE is undefined (e.g. the
// default situation) the following macros are defined.  These
// cause this file to implement the VectorT<TYPE> template
//
//       TEMPLATE    template<class EltType>
//       TYPE        EltType
//       VCLASSNAME  VectorT
//       MCLASSNAME  MatrixT
//       VECT        VCLASSNAME<TYPE>
//       MAT         MCLASSNAME<TYPE>
//
// If these are defined differently it is possible to
// implement a non-template version of the Vector data type.
// This is done, for example, by the DVector header file, which
// defines the above as follows (and sets the __DONT... flag)
//
//       TEMPLATE
//       TYPE        double
//       VCLASSNAME  DVector
//       MCLASSNAME  DMatrix
//       VECT        DVector
//       MAT         DMatrix
//

#ifndef __DONT_DEFINE_VECTOR_TEMPLATE
#pragma interface
#define TEMPLATE template<class Type>
#define TYPE Type
#define VCLASSNAME VectorT
#define MCLASSNAME MatrixT
#define MAT MCLASSNAME<TYPE>
#define VECT VCLASSNAME<TYPE>
#else
#ifndef TEMPLATE
#define TEMPLATE
#endif
#ifndef MAT
#define MAT MCLASSNAME
#endif
#ifndef VECT
#define VECT VCLASSNAME
#endif
#endif

// this is used where a reference could also be used - in the
// future the template might be expanded to facilitate non
// basic types, or types with expensive copy functions - in
// this case we want to use references to pass arguments.
#define TYPER TYPE



//======================================================================
//
// Header build control macros
//

//========================= Error handling

#define ERROR(val,ret) { fprintf(stderr,"%s<> library error, line %d, file %s\n:::: error=%s\n",__CLASSNAMESTRING,__LINE__,__FILE__,#val); abort(); return ret; }

//========================= Element Access

// the internal cycling variable name
#define INDEX __ii

// if the user has requested optimized element access, then, wherever
// possible we access the data directly, instead of through the subscription
// operator - the subscription operator should be used when debugging (i.e.
// turn off the optimization) since that provides the user with
// additional break points, error checking, and status information
#ifdef OPTIMIZE_ELEMENT_ACCESS
// yes, optimize
#define FOR_EACH_ELEMENT   for(register int INDEX=0; INDEX<numelts; INDEX++)
#define ELTKN(k)           data[k]
#define ELTK(k)            data[k-SUBSCRIPTION_OFFSET]
#define ELT                data[INDEX]
#define REFELTKN(v,k)      (v.data[k])
#define REFELTK(v,k)       (v.data[k-SUBSCRIPTION_OFFSET])
#define REFELT(v)          (v.data[INDEX])

#else
// no, don't optimize
#define FOR_EACH_ELEMENT   for(register int INDEX=0; INDEX<numelts; INDEX++)
#define ELTKN(k)           ((*this)(k+SUBSCRIPTION_OFFSET))
#define ELTK(k)            ((*this)(k))
#define ELT                ((*this)(INDEX))
#define REFELTKN(v,k)      (v(k+SUBSCRIPTION_OFFSET))
#define REFELTK(v,k)       (v(k))
#define REFELT(v)          (v(INDEX))

#endif

//========================= Absolute Value (we might optimize this later)
#define ABS(a) (a<0?-a:a)
#define ABSELT (ELT<0?-ELT:ELT)

//========================= Free-Store Management
#define GET_NEW_ELTS(name,num) name=new TYPE[num]
#define FREE_ELT_SPACE(name)   delete name

//========================= Vector Access Protection
#ifdef PROTECT_VECTOR_ACCESS
#define CHECK_FOR_VALID_VECTOR(ret)    \
             if(numelts<0) \
               ERROR(VectorT_Invalid_Access,ret);
#define CHECK_V_FOR_VALID_VECTOR(v,ret)               \
            if((&v)==0)                               \
               { ERROR(VectorT_Null_Operand,ret); }   \
            else                                      \
               { if(v.numelts<0)                      \
                  ERROR(VectorT_Invalid_Access,ret); }
#else
#define CHECK_FOR_VALID_VECTOR(ret)
#define CHECK_V_FOR_VALID_VECTOR(v,ret)
#endif

//========================= Matrix Access Protection
#ifdef PROTECT_MATRIX_ACCESS
#define CHECK_FOR_VALID_MATRIX(ret)    \
             if((width<0)||(height<0)) \
               ERROR(MatrixT_Invalid_Access,ret);
#define CHECK_M_FOR_VALID_MATRIX(m,ret)               \
            if((&m)==0)                               \
               { ERROR(MatrixT_Null_Operand,ret); }   \
            else                                      \
               { if((m.width<0)||(m.height<0))        \
                  ERROR(MatrixT_Invalid_Access,ret); }
#else
#define CHECK_FOR_VALID_MATRIX(ret)
#define CHECK_M_FOR_VALID_MATRIX(m,ret)
#endif

//========================= Ctor and Dtor Debugging Auxiliaries
#define ___CREATE_STRING(a) #a
#define __CREATE_STRING(a) ___CREATE_STRING(a)
#define __CLASSNAMESTRING __CREATE_STRING(VCLASSNAME)


//======================================================================
//======================================================================
//======================================================================
//
// Actual class function definition
//

// ================================================== Constructors
TEMPLATE INLINE VECT::VCLASSNAME(TYPE* new_val_ptr,int sz,VectorType vtype) {
#ifdef DEBUG_CTORS_AND_DTORS
 fprintf(DEBUG_SINK,"%s(%08lx,%d,%d) this=%08lx\n",__CLASSNAMESTRING,
      (unsigned long)new_val_ptr,sz,(int)_type,(unsigned long)this);
#endif
 numelts=sz;
 _type=vtype;
 data=new_val_ptr;
 }

TEMPLATE INLINE VECT::VCLASSNAME() {
#ifdef DEBUG_CTORS_AND_DTORS
 fprintf(DEBUG_SINK,"%s() this=%08lx\n",__CLASSNAMESTRING,(unsigned long)this);
#endif
 numelts=0;
 data=0;
 }

TEMPLATE INLINE VECT::VCLASSNAME(int sz,VectorType t) {
 numelts=sz;
 _type=t;
 GET_NEW_ELTS(data,sz);
#ifdef DEBUG_CTORS_AND_DTORS
 fprintf(DEBUG_SINK,"%s(%d,%d) data=%08lx this=%08lx\n",__CLASSNAMESTRING,
      sz,(int)_type,(unsigned long)data,(unsigned long)this);
#endif
 }

TEMPLATE INLINE VECT::VCLASSNAME(int sz,const TYPE* vals,VectorType t) {
 numelts=sz;
 _type=t;
 GET_NEW_ELTS(data,sz);
#ifdef DEBUG_CTORS_AND_DTORS
 fprintf(DEBUG_SINK,"%s(%d,%08lx,%d) data=%08lx this=%08lx\n",__CLASSNAMESTRING,
   sz,(unsigned long)vals,(int)_type,(unsigned long)data,(unsigned long)this);
#endif
 init(vals);
 }

TEMPLATE INLINE VECT::VCLASSNAME(const MAT& m) {
 data=0;
#ifdef DEBUG_CTORS_AND_DTORS
 fprintf(DEBUG_SINK,"%s(Mat(%d,%d,%08lx,%08lx) this=%08lx\n",__CLASSNAMESTRING,
            m.height,m.width,(unsigned long)m.data,(unsigned long)&m,
               (unsigned long)this);
#endif
 if(m.width==1) {
   numelts=m.height;
   _type=ColV;
   GET_NEW_ELTS(data,numelts);
   FOR_EACH_ELEMENT ELT=m(INDEX,0);
   }
 else
   if(m.height==1) {
      numelts=m.width;
      _type=RowV;
      GET_NEW_ELTS(data,numelts);
      FOR_EACH_ELEMENT ELT=m(0,INDEX);
      }
#ifndef DONT_PROTECT_INVALID_OPERATIONS
   else {
      ERROR(VectorT_Vector_Ctor_From_Matrix_Not_Given_Valid_Matrix,/**/);
      }
#endif
 }

TEMPLATE INLINE VECT::VCLASSNAME(const VECT& v) {
 data=0;
 _type=v._type;
 numelts=v.numelts;
 GET_NEW_ELTS(data,numelts);
#ifdef DEBUG_CTORS_AND_DTORS
 fprintf(DEBUG_SINK,"%s(Vect(%d,%d,%08lx,%08lx) data=%08lx this=%08lx\n",
      __CLASSNAMESTRING,v.numelts,(int)v._type,(unsigned long)v.data,(unsigned long)&v,
         (unsigned long)data,(unsigned long)this);
#endif
 init(v.data);
 }

TEMPLATE INLINE VECT::~VCLASSNAME() {
#ifdef DEBUG_CTORS_AND_DTORS
 fprintf(DEBUG_SINK,"~%s() numelts=%d type=%d data=%08lx this=%08lx\n",
      __CLASSNAMESTRING,numelts,(int)_type,(unsigned long)data,(unsigned long)this);
#endif
 numelts=-1;
 FREE_ELT_SPACE(data);
 data=0;
 }

 //================================================== Information
 // ** Information (where p is precision, e.g.
 //                 assume zero if a(i,j) < p
TEMPLATE INLINE TYPE& VECT::Min(int* ip) const
{
 CHECK_FOR_VALID_VECTOR(error_return_value())
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 // if this is a zero size matrix, return error_zero, set i,j=0 if applicable
 if(data==0) {
   if(ip!=0) *ip=-1;
   return (error_return_value());
   }
#endif
 register int ir=0;
 TYPE val=data[0];
 register TYPE k;
 FOR_EACH_ELEMENT {
   k=ELT;
   if(k<val) { ir=INDEX; val=k; }
   }

 // save ir in user space if requested, then return the elt ref
 if(ip!=0)  *ip=ir;
 return ELTK(ir);
}

// return Minimum absolute value elt
TEMPLATE INLINE TYPE& VECT::AbsMin(int* ip) const
{
 CHECK_FOR_VALID_VECTOR(error_return_value())
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 // if this is a zero size matrix, return error_zero, set i,j=0 if applicable
 if(data==0) {
   if(ip!=0)   *ip=-1;
   return (error_return_value());
   }
#endif
 // ir,jr = temp for holding location of min
 register int ir=0;
 TYPE val=ABS(data[0]);
 register TYPE k;

 FOR_EACH_ELEMENT {
   k=ABSELT;
   if(k<val) { ir=INDEX;; val=k; }
   }

 // save ir in user space if requested, then return the elt ref
 if(ip!=0)  *ip=ir;
 return ELTK(ir);
}

//================================================== return Maximum elt
TEMPLATE INLINE TYPE& VECT::Max(int* ip) const
{
 CHECK_FOR_VALID_VECTOR(error_return_value())
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 // if this is a zero size matrix, return error_zero, set i,j=0 if applicable
 if(data==0) {
   if(ip!=0)   *ip=-1;
   return (error_return_value());
   }
#endif

 // ir,jr = temp for holding location of min
 register int ir=0;
 TYPE val=data[0];
 FOR_EACH_ELEMENT
   if(ELT>val) { ir=INDEX;; val=ELT; }

 // save ir in user space if requested, then return the elt ref
 if(ip!=0)  *ip=ir;
 return ELTK(ir);
}

//================================================== return Maximum elt, absolute value
TEMPLATE INLINE TYPE& VECT::AbsMax(int* ip) const
{
 CHECK_FOR_VALID_VECTOR(error_return_value())
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 // if this is a zero size matrix, return error_zero, set i,j=0 if applicable
 if(data==0) {
   if(ip!=0)   *ip=-1;
   return (error_return_value());
   }
#endif

 // ir = temp for holding location of min
 register int ir=0;
 TYPE val=data[0];
 if(val<0) val= -val;
 register TYPE k;
 FOR_EACH_ELEMENT {
   k=ABSELT;
   if(k>val) { ir=INDEX; val=k; }
   }

 // save ir in user space if requested, then return the elt ref
 if(ip!=0)  *ip=ir;
 return ELTK(ir);
}


//================================================== return number of elts <= p
TEMPLATE INLINE double VECT::PNorm(double p) const
{
 CHECK_FOR_VALID_VECTOR(1.0)
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 // since the 0-norm doesn't make sense
 if(p=0.0)
   return 1.0;
#endif

 register double sum=0.0;
 FOR_EACH_ELEMENT { register TYPE k=ABSELT; sum+=pow(k,p); }
 return pow(sum,1.0/p);
}

TEMPLATE INLINE double VECT::Length() const {
 CHECK_FOR_VALID_VECTOR(1.0)
 register double sum=0.0;
 FOR_EACH_ELEMENT { register TYPE k=ELT; sum+=(k*k); }
 return sqrt(sum);
}

TEMPLATE INLINE int VECT::NumZeros(const TYPER pin) const
{
 CHECK_FOR_VALID_VECTOR(-1)
 const TYPE p=ABS(pin);
 register int ct=0;
 FOR_EACH_ELEMENT {
   register TYPE k=ABSELT;
   if(k<=p) ct++;
   }
 return ct;
}

 //================================================== Predicates
 // ** Predicates (where p is precision, e.g.
 //                assume zero if a(i,j) < p

//================================================== true if this is zero
TEMPLATE INLINE int VECT::iszero(const TYPER p) const
   { return (NumZeros(p)==(numelts)); }


 //================================================== Element Access
 // ** Element access
TEMPLATE INLINE TYPE& VECT::operator()(int i) {
#ifdef PROTECT_VECTOR_ACCESS
 CHECK_FOR_VALID_VECTOR(error_return_value())
#if SUBSCRIPTION_OFFSET
 if(i<0) {
#else
 if(i<SUBSCRIPTION_OFFSET) {
#endif
   ERROR(VectorT_Bound_Too_Low,error_return_value());
   }
#if SUBSCRIPTION_OFFSET
 if(i>=(numelts-SUBSCRIPTION_OFFSET)) {
#else
 if(i>=numelts) {
#endif
   ERROR(VectorT_Bound_Too_High,error_return_value());
   }
 if(numelts<0) {
   ERROR(VectorT_Access_after_object_was_destroyed,error_return_value());
   }
#endif
#ifdef PROTECT_POINTER_ACCESS
 if(data==0) {
   ERROR(VectorT_Data_Pointer_Null,error_return_value());
   }
#endif
#if SUBSCRIPTION_OFFSET
 return data[i-SUBSCRIPTION_OFFSET];
#else
 return data[i];
#endif
 } /// }}

TEMPLATE INLINE TYPE& VECT::operator()(int i) const {
#ifdef PROTECT_VECTOR_ACCESS
 CHECK_FOR_VALID_VECTOR(error_return_value())
#if SUBSCRIPTION_OFFSET
 if(i<0) {
#else
 if(i<SUBSCRIPTION_OFFSET) {
#endif
   ERROR(VectorT_Bound_Too_Low,error_return_value());
   }
#if SUBSCRIPTION_OFFSET
 if(i>=(numelts-SUBSCRIPTION_OFFSET)) {
#else
 if(i>=numelts) {
#endif
   ERROR(VectorT_Bound_Too_High,error_return_value());
   }
 if(numelts<0) {
   ERROR(VectorT_Access_after_object_was_destroyed,error_return_value());
   }
#endif
#ifdef PROTECT_POINTER_ACCESS
 if(data==0) {
   ERROR(VectorT_Data_Pointer_Null,error_return_value());
   }
#endif
#if SUBSCRIPTION_OFFSET
 return data[i-SUBSCRIPTION_OFFSET];
#else
 return data[i];
#endif
 } //}}

TEMPLATE INLINE VECT VECT::sub_vector( int start, int stop) const {
 CHECK_FOR_VALID_VECTOR(VECT())
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 if(stop<start) { register int tmp=stop; stop=start; start=tmp; }
 if(start<SUBSCRIPTION_OFFSET)
   ERROR(VectorT_Start_Value_Too_Low_For_Sub_Vector,VECT());
 if(start>=(numelts+SUBSCRIPTION_OFFSET))
   ERROR(VectorT_Start_Value_Too_High_For_Sub_Vector,VECT());
 if(stop<SUBSCRIPTION_OFFSET)
   ERROR(VectorT_Stop_Value_Too_Low_For_Sub_Vector,VECT());
 if(stop>=(numelts+SUBSCRIPTION_OFFSET))
   ERROR(VectorT_Stop_Value_Too_High_For_Sub_Vector,VECT());
#endif
 return VECT(&ELTK(start),stop-start,_type);
}

 //================================================== General Vector Operations
// init from array
TEMPLATE INLINE VECT&  VECT::init(const TYPE* val) {
 CHECK_FOR_VALID_VECTOR((*this))
#ifdef PROTECT_POINTER_ACCESS
 if(val==0)
   ERROR(VectorT_Init_Passed_NULL_Value_Pointer,(*this));
#endif
 FOR_EACH_ELEMENT
   ELT = val[INDEX];
 return *this;
 }

// set all elts = val
TEMPLATE INLINE VECT&  VECT::fill(const TYPER val) {
 CHECK_FOR_VALID_VECTOR((*this))
 FOR_EACH_ELEMENT
   ELT=val;
 return (*this);
 }

// take absolute value of each elt
TEMPLATE INLINE VECT&  VECT::abs() {
 CHECK_FOR_VALID_VECTOR((*this))
 FOR_EACH_ELEMENT
   ELT=ABSELT;
 return (*this);
 }

TEMPLATE INLINE VECT& VECT::extract_elt(int j) {
 CHECK_FOR_VALID_VECTOR((*this))
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 if(j<SUBSCRIPTION_OFFSET)
   ERROR(VectorT_Extract_TYPE_Element_Num_Too_Low(),(*this));
 if(j>=(numelts+SUBSCRIPTION_OFFSET))
   ERROR(VectorT_Extract_TYPE_Element_Num_Too_Hight(),(*this));
#endif
#if SUBSCRIPTION_OFFSET
 j-=SUBSCRIPTION_OFFSET;
#endif
 register int k=j;
 for(register int i=j+1;i<numelts;i++,k++)
   ELTKN(k)=ELTKN(i);

 return (*this);
 }


 //================================================== Comparison Operators
// true if this(i,j) = a(i,j)
TEMPLATE INLINE int VECT::operator ==(const VECT&  v) const
{
 CHECK_FOR_VALID_VECTOR(-1)
 CHECK_V_FOR_VALID_VECTOR(v,-1)
 if(v.numelts!=numelts) return 0;
#ifdef DISTINGUISH_BETWEEIN_VECTOR_TYPES
 if(v._type!=_type) return 0;
#endif
 FOR_EACH_ELEMENT
   if(ELT != REFELT(v)) return 0;
 return 1;
}

TEMPLATE INLINE int VECT::operator !=(const VECT&  v) const
   { return (!((*this)==v)); }

// true if this(i,j) >= a(i,j)
TEMPLATE INLINE int VECT::operator >=(const VECT&  v) const
{
 CHECK_FOR_VALID_VECTOR(-1)
 CHECK_V_FOR_VALID_VECTOR(v,-1)
 if(numelts>v.numelts) return 1;
 if(numelts!=v.numelts) return 0;
#ifdef DIFFERENT_VECTOR_TYPES_CANT_BE_EQUAL
 if(v._type!=_type) return 0;
#endif
 FOR_EACH_ELEMENT
   if(ELT < REFELT(v)) return 0;
 return 1;
}

// true if this(i,j) <= a(i,j)
TEMPLATE INLINE int VECT::operator <=(const VECT&  v) const {
 CHECK_FOR_VALID_VECTOR(-1)
 CHECK_V_FOR_VALID_VECTOR(v,-1)
 if(numelts<v.numelts) return 1;
 if(numelts!=v.numelts) return 0;
#ifdef DIFFERENT_VECTOR_TYPES_CANT_BE_EQUAL
 if(v._type!=_type) return 0;
#endif
 FOR_EACH_ELEMENT
   if(ELT > REFELT(v)) return 0;
 return 1;
}

// true if this(i,j) > a(i,j)
TEMPLATE INLINE int VECT::operator >(const VECT&  v) const {
 CHECK_FOR_VALID_VECTOR(-1)
 CHECK_V_FOR_VALID_VECTOR(v,-1)
 if(numelts>v.numelts) return 1;
 if(numelts!=v.numelts) return 0;
#ifdef DIFFERENT_VECTOR_TYPES_CANT_BE_EQUAL
 if(v._type!=_type) return 0;
#endif
 FOR_EACH_ELEMENT
   if(ELT <= REFELT(v)) return 0;
 return 1;
}

// true if this(i,j) < a(i,j)
TEMPLATE INLINE int VECT::operator <(const VECT&  v) const {
 CHECK_FOR_VALID_VECTOR(-1)
 CHECK_V_FOR_VALID_VECTOR(v,-1)
 if(numelts<v.numelts) return 1;
 if(numelts!=v.numelts) return 0;
#ifdef DIFFERENT_VECTOR_TYPES_CANT_BE_EQUAL
 if(v._type!=_type) return 0;
#endif
 FOR_EACH_ELEMENT
   if(ELT >= REFELT(v)) return 0;
 return 1;
}

//================================================== Assignment Operators
// this <- v
TEMPLATE INLINE VECT& VECT::operator =(const VECT&  v) {
 CHECK_FOR_VALID_VECTOR((*this))
 CHECK_V_FOR_VALID_VECTOR(v,(*this))
 if(data!=0) { // then this is not a null matrix so we check to make sure the
               // sizes agree, then we use the same space as the other one.
   if(v.numelts<numelts) {
      numelts=v.numelts;
      _type=v._type;
      }
   else {
      FREE_ELT_SPACE(data);
      numelts=v.numelts;
      _type=v._type;
      GET_NEW_ELTS(data,numelts);
      }
   }
 else {
   numelts=v.numelts;
   _type=v._type;
   GET_NEW_ELTS(data,numelts);
   }

 init(v.data);
 return (*this);
}

TEMPLATE INLINE VECT::operator double() const {
 CHECK_FOR_VALID_VECTOR(0.0)
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 if(numelts<=1)
   ERROR(VectorT_Cast_To_Double_With_Bad_Value,0.0);
#endif
 return (double)ELTK(SUBSCRIPTION_OFFSET);
 }

TEMPLATE INLINE VECT::operator long() const {
 CHECK_FOR_VALID_VECTOR(0)
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 if(numelts<=1)
   ERROR(VectorT_Cast_To_Double_With_Bad_Value,0);
#endif
 return (long)ELTK(SUBSCRIPTION_OFFSET);
 }

TEMPLATE INLINE VECT& VECT::operator +=(const VECT& v) {
 CHECK_FOR_VALID_VECTOR((*this))
 CHECK_V_FOR_VALID_VECTOR(v,(*this))
#ifndef DONT_PROTECT_INVALID_OPERATIONS
#ifdef OPERATIONS_ON_VECTORS_WITH_SAME_LENGTH_BUT_DIFFERENT_TYPES_ARE_INVALID
 if(v._type!=_type)
   ERROR(VectorT_Plus_Equal_With_Different_Types,(*this));
#endif
 if(v.numelts!=numelts)
   ERROR(VectorT_Plus_Equal_With_Different_Lengths,(*this));
#endif
 FOR_EACH_ELEMENT
   ELT+=REFELT(v);
 return (*this);
}

TEMPLATE INLINE VECT& VECT::operator -=(const VECT& v) {
 CHECK_FOR_VALID_VECTOR((*this))
 CHECK_V_FOR_VALID_VECTOR(v,(*this))
#ifndef DONT_PROTECT_INVALID_OPERATIONS
#ifdef OPERATIONS_ON_VECTORS_WITH_SAME_LENGTH_BUT_DIFFERENT_TYPES_ARE_INVALID
 if(v._type!=_type)
   ERROR(VectorT_Minus_Equal_With_Different_Types,(*this));
#endif
 if(v.numelts!=numelts)
   ERROR(VectorT_Minus_Equal_With_Different_Lengths,(*this));
#endif
 FOR_EACH_ELEMENT
   ELT-=REFELT(v);
 return (*this);
}

TEMPLATE INLINE VECT& VECT::operator /=(const TYPER s) {
 CHECK_FOR_VALID_VECTOR((*this))
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 if(s==0.0)
   ERROR(VectorT_Divide_Equal_Given_Zero_Argument,(*this));
#endif
 FOR_EACH_ELEMENT
   ELT/=s;
 return (*this);
}

TEMPLATE INLINE VECT& VECT::operator *=(const TYPER s) {
 CHECK_FOR_VALID_VECTOR((*this))
 FOR_EACH_ELEMENT
   ELT*=s;
 return (*this);
 }

 //================================================== Vector Operators

// matrix transpose
TEMPLATE INLINE VECT VECT::operator ~() const {
 CHECK_FOR_VALID_VECTOR(VECT())
 if(_type==RowV)
   return VECT(numelts,data,ColV);
 else
   return VECT(numelts,data,RowV);
}

TEMPLATE INLINE VECT VECT::transpose() const {
 CHECK_FOR_VALID_VECTOR(VECT())
 if(_type==RowV)
   return VECT(numelts,data,ColV);
 else
   return VECT(numelts,data,RowV);
}

TEMPLATE INLINE VECT& VECT::transpose_self() {
 CHECK_FOR_VALID_VECTOR((*this))
 _type=(_type==RowV?ColV:RowV);
 return (*this);
}



TEMPLATE INLINE VECT VECT::operator +(const VECT& v) const {
 CHECK_FOR_VALID_VECTOR(VECT())
 CHECK_V_FOR_VALID_VECTOR(v,VECT())
#ifdef DISTINGUISH_BETWEEN_VECTOR_TYPES
 if(v._type!=_type)
   ERROR(VectorT_Attempt_To_Add_Different_Vector_Types(),VECT());
#endif
 TYPE* vals; GET_NEW_ELTS(vals,numelts);
 FOR_EACH_ELEMENT vals[INDEX]=ELT+REFELT(v);
 return VECT(vals,numelts,_type);
 }

TEMPLATE INLINE VECT VECT::operator -() const {
 CHECK_FOR_VALID_VECTOR(VECT())
 TYPE* vals; GET_NEW_ELTS(vals,numelts);
 FOR_EACH_ELEMENT vals[INDEX]=-ELT;
 return VECT(vals,numelts,_type);
 }

TEMPLATE INLINE VECT VECT::operator -(const VECT&  v) const {
 CHECK_FOR_VALID_VECTOR(VECT())
 CHECK_V_FOR_VALID_VECTOR(v,VECT())
#ifdef DISTINGUISH_BETWEEN_VECTOR_TYPES
 if(v._type!=_type)
   ERROR(VectorT_Attempt_To_Subtract_Different_Vector_Types(),VECT());
#endif
 TYPE* vals; GET_NEW_ELTS(vals,numelts);
 FOR_EACH_ELEMENT vals[INDEX]=ELT-REFELT(v);
 return VECT(vals,numelts,_type);
 }

TEMPLATE INLINE VECT VECT::operator /(const TYPER  s) const {
 CHECK_FOR_VALID_VECTOR(VECT())
 TYPE* vals; GET_NEW_ELTS(vals,numelts);
 FOR_EACH_ELEMENT vals[INDEX]=ELT/s;
 return VECT(vals,numelts,_type);
 }

TEMPLATE INLINE VECT VECT::operator *(const TYPER  s) const {
 CHECK_FOR_VALID_VECTOR(VECT())
 TYPE* vals; GET_NEW_ELTS(vals,numelts);
 FOR_EACH_ELEMENT vals[INDEX]=ELT*s;
 return VECT(vals,numelts,_type);
 }

TEMPLATE INLINE MAT VECT::operator *(const VECT&  v) const {
 CHECK_FOR_VALID_VECTOR(MAT())
 CHECK_V_FOR_VALID_VECTOR(v,MAT())
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 if(v._type==_type)
   ERROR(VectorT_Attempted_Vect_Vect_Multiply_With_Same_Vect_Types,MAT());
 if((v.numelts!=numelts)&&(_type!=RowV))
   ERROR(VectorT_Attempted_Row_Col_Multiply_With_Different_Lengths,MAT());
#endif
 if(_type==RowV) {
   // take inner product
   TYPE product=0;
   FOR_EACH_ELEMENT product+=(ELT*REFELT(v));
   return MAT(1,1,&product);
   }
 else {
   // take outer product
   TYPE* vals; GET_NEW_ELTS(vals,numelts*v.numelts);
   FOR_EACH_ELEMENT {
      for(register int i=0;i<v.numelts;i++)
         vals[(INDEX*v.numelts)+i]=ELT*REFELTKN(v,i);
      }
   return MAT(vals,numelts,v.numelts);
   }
 }

TEMPLATE INLINE MAT VECT::operator *(const MAT&  m) const {
 CHECK_FOR_VALID_VECTOR(MAT())
 CHECK_M_FOR_VALID_MATRIX(m,MAT())
 printf("STILL WORKING\n");
 return MAT();
 }

TEMPLATE INLINE VECT operator *(const TYPER s,const VECT& v) { return v*s; }
TEMPLATE INLINE MAT operator *(const MAT& m,const VECT& v) {
 MAT tmat(v);
 return MAT(m*tmat);
 }

TEMPLATE INLINE double VECT::dot(const VECT& v) const {
 CHECK_FOR_VALID_VECTOR(0.0)
 CHECK_V_FOR_VALID_VECTOR(v,0.0)
#ifndef DONT_PROTECT_INVALID_OPERATIONS
#ifdef DISTINGUISH_BETWEEN_VECTOR_TYPES
 if((_type!=RowV)||(v._type==ColV))
   ERROR(VectorT_Attempted_Dot_Product_With_Incorrect_Vector_Types,0.0);
#endif
 if(numelts!=v.numelts)
   ERROR(VectorT_Attempted_Dot_Product_With_Incorrect_Vector_Lengths,0.0);
#endif
 double prod=0.0;
 FOR_EACH_ELEMENT prod+=(REFELT(v)*ELT);
 return prod;
 }

TEMPLATE INLINE double VECT::inner_product(const VECT& v) const { return dot(v); }

TEMPLATE INLINE MAT VECT::outer_product(const VECT& v) const {
 CHECK_FOR_VALID_VECTOR(MAT())
 CHECK_V_FOR_VALID_VECTOR(v,MAT())
#ifndef DONT_PROTECT_INVALID_OPERATIONS
#ifdef DISTINGUISH_BETWEEN_VECTOR_TYPES
 if((_type!=ColV)||(v._type==RowV))
   ERROR(VectorT_Attempted_Outer_Product_With_Incorrect_Vector_Types,MAT());
#endif
#endif
 printf("STILL WORKING\n");
 return MAT();
 }

TEMPLATE INLINE VECT VECT::normalize() const {
 CHECK_FOR_VALID_VECTOR(VECT())
 register double len=Length();
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 if(len==0.0) return VECT();
#endif
 TYPE* vals; GET_NEW_ELTS(vals,numelts);
 register int k=0;
 FOR_EACH_ELEMENT vals[k++]=(ELT/len);
 return VECT(vals,numelts,_type);
 }

TEMPLATE INLINE VECT& VECT::normalize_self() {
 CHECK_FOR_VALID_VECTOR((*this))
 register double len=Length();
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 if(len==0.0) return (*this);
#endif
 return (*this)/=len;
 }

 //================================================== Vector IO
// print an ascii version of the matrix
TEMPLATE INLINE void VECT::print(FILE* file) const {
 CHECK_FOR_VALID_VECTOR(;)
 if(_type==RowV) {
   fprintf(file,"1 %d [",numelts);
   FOR_EACH_ELEMENT fprintf(file,"%g ",(double)ELT);
   fprintf(file,"]\n");
   }
 else {
   fprintf(file,"1 %d [",numelts);
   FOR_EACH_ELEMENT fprintf(file,"%g\n",(double)ELT);
   fprintf(file,"]\n");
   }
}

// read an ascii version of the matrix
TEMPLATE INLINE int VECT::read(FILE* file) {
 CHECK_FOR_VALID_VECTOR(-1)
 int height, width;
 fscanf(file, "%d %d [", &height, &width);

 if(height==1)
   { _type=RowV; numelts=width; }
 else
   if(width==1)
      { _type=ColV; numelts=height; }
   else
      { ERROR(VectorT_Attempted_To_Read_A_Non_Vector,-1); }

 FREE_ELT_SPACE(data);
 GET_NEW_ELTS(data,numelts);

 double val;
 if(_type==RowV) {
   FOR_EACH_ELEMENT {
      fscanf(file,"%lg ",&val);
      ELT=(TYPE)val;
      }
   fscanf(file,"]\n");
   }
 else {
   fscanf(file,"\n");
   FOR_EACH_ELEMENT {
      fscanf(file,"%lg\n",&val);
      ELT=(TYPE)val;
      }
   fscanf(file,"]\n");
   }
 return 0;
}


// save to f in format fmt
TEMPLATE INLINE int VECT::save(FILE* f,const char* fmt) const
{
  printf("STILL WORKING\n");
  return 0;
}

// create fname and save, use format fmt
TEMPLATE INLINE int VECT::save(const char* fname,const char* fmt) const
{
 CHECK_FOR_VALID_VECTOR(-1)
 FILE* f;
 if((f=fopen(fname,"w"))==0)
   return -1;
 int ret;
 ret=save(f,fmt);
 fclose(f);
 return ret;
}

// load from f, expect format fmt
TEMPLATE INLINE int VECT::load(FILE* f,const char* fmt)
{
 printf("STILL WORKING\n");
 return 0;
}

// open fname and load, expect format fmt
TEMPLATE INLINE int VECT::load(const char* fname,const char* fmt) {
 CHECK_FOR_VALID_VECTOR(-1)
 FILE* f;
 if((f=fopen(fname,"r"))==NULL) return -1;
 int ret=load(f,fmt);
 fclose(f);
 return ret;
}


#undef INLINE
#undef ROW_INDEX
#undef COL_INDEX 
#undef FOR_EACH_ELEMENT
#undef FOR_EACH_ROW_ELEMENT
#undef FOR_EACH_COL_ELEMENT
#undef REFELT
#undef REFELTK
#undef REFELTKN
#undef ELT
#undef ELTK
#undef ELTKN
#undef ABS
#undef ABSELT
#undef GET_NEW_ELTS
#undef FREE_ELT_SPACE
#undef CHECK_FOR_VALID_VECTOR
#undef CHECK_V_FOR_VALID_VECTOR
#undef CHECK_FOR_VALID_MATRIX
#undef CHECK_M_FOR_VALID_MATRIX
#undef ___CREATE_STRING
#undef __CREATE_STRING
#undef __CLASSNAMESTRING

#ifndef __DONT_DEFINE_VECTOR_TEMPLATE
#undef TYPE
#undef TYPER
#undef VCLASSNAME
#undef MCLASSNAME
#endif
#undef TEMPLATE
#undef MAT
#undef VECT

#endif
