// grob.C
// The general methods for storing, manipulating, creating, and displaying our GRaphic OBjects
// note: these classes require glut

#include "grob.h"
#include "JRutils.h"
#include "cfpext.h"

// temp addition for saving matrices to file
#include "../../cfplibs/biobot/misc.h"
#include "../../cfplibs/biobot/misc.def"

#include <iostream>
#include <cmath>

//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace Grob{

#include "grobinline.h"
#include "grobcamerainline.h"




// ----------------------------------------------------------------------------------
// -- Grob::Static flags for which modes we are currently supporting
// ----------------------------------------------------------------------------------

#ifdef GrobGL
bool Grob::using_gl=true;
#else
bool Grob::using_gl=false;
#endif

#ifdef GrobPOV
bool Grob::using_pov=true;
#else
bool Grob::using_pov=false;
#endif

#ifdef GrobVRML
bool Grob::using_vrml=true;
#else
bool Grob::using_vrml=false;
#endif


// ----------------------------------------------------------------------------------
// -- Grob::supports_gl()
// more static functions
// ----------------------------------------------------------------------------------
bool Grob::supports_gl()
{
  // return true if this grob can perform GL stuff
  #ifdef GrobGL
  return true;
  #endif
  return false;
}

// ----------------------------------------------------------------------------------
// -- Grob::supports_pov()
// ----------------------------------------------------------------------------------
bool Grob::supports_pov()
{
  // return true if this grob can perform Pov stuff
  #ifdef GrobGL
  return true;
  #endif
  return false;
}

// ----------------------------------------------------------------------------------
// -- Grob::supports_vrml()
// ----------------------------------------------------------------------------------
bool Grob::supports_vrml()
{
  // return true if this grob can perform vrml stuff
  #ifdef GrobGL
  return true;
  #endif
  return false;
}





// ----------------------------------------------------------------------------------
// -- Grob::use_gl
// ----------------------------------------------------------------------------------
bool Grob::use_gl()
{
  if (supports_gl())
    return (using_gl=true);
  else
    return false;
}

// ----------------------------------------------------------------------------------
// -- Grob::use_pov
// ----------------------------------------------------------------------------------
bool Grob::use_pov()
{
  if (supports_pov())
    return (using_pov=true);
  else
    return false;
}

// ----------------------------------------------------------------------------------
// -- Grob::use_vrml
// ----------------------------------------------------------------------------------
bool Grob::use_vrml()
{
  if (supports_vrml())
    return (using_vrml=true);
  else
    return false;
}










// ----------------------------------------------------------------------------------
// -- Grob::initialize()
// ----------------------------------------------------------------------------------

void Grob::initialize()
{
  // initialize it so that it is undefined
  location=perVector(3);
  orientation=perMatrix(3,3);

  // initialize it
  location.fill(0.0);
  orientation.fill(0.0);
  orientation(0,0)=1;
  orientation(1,1)=1;
  orientation(2,2)=1;
  oscale[0]=oscale[1]=oscale[2]=1.0;
  
  // zero out the matrix
  olmat[3]=olmat[7]=olmat[11]=0.0;

  makematrix();
  
  // initialize components
  gl_initialize();
}




// ----------------------------------------------------------------------------------
// -- Grob::~Grob()
// ----------------------------------------------------------------------------------

Grob::~Grob()
{
  // free any allocated displaylists
  gl_uninitialize();
}






// ----------------------------------------------------------------------------------
// -- Grob::setscale()
// ----------------------------------------------------------------------------------

void Grob::setscale(perVectorR3 inscale)
{
  // set location
  oscale[0]=inscale(0);
  oscale[1]=inscale(1);
  oscale[2]=inscale(2);

  // set update flags
  gl_needsupdate=true;
}




void Grob::setscale(float x,float y,float z)
{
  // set location
  oscale[0]=x;
  oscale[1]=y;
  oscale[2]=z;

  // set update flags
  gl_needsupdate=true;
}

void Grob::setscale(float inscale[3])
{
  // set location
  oscale[0]=inscale[0];
  oscale[1]=inscale[1];
  oscale[2]=inscale[2];

  // set update flags
  gl_needsupdate=true;
}

void Grob::setscale(int ind,float scale)
{
  // set location
  oscale[ind]=scale;

  // set update flags
  gl_needsupdate=true;
}




















// ---------------------------------------------------------------------------------------------
// -- Helper functions which make 3 part display lists of shapes
// -- Right now they only know how to make GL components, but we would like to add more later
// ---------------------------------------------------------------------------------------------


void Grob::setupcontext()
{
  // prepare context
  #ifdef GrobGL
  GrobGlHelper::setuplights();
  #endif
}


void Grob::makePointList()
{
  #ifdef GrobGL
  int lownum,mednum,highnum,finenum,paintnum;
  lownum=GrobGlHelper::makePoint(1,lowdetail);
  mednum=GrobGlHelper::makePoint(2,meddetail);
  highnum=GrobGlHelper::makePoint(3,highdetail);
  finenum=GrobGlHelper::makePoint(3,finedetail);
  paintnum=GrobGlHelper::makePoint(3,paintdetail);
  set_gl(lownum,mednum,highnum,finenum,paintnum);
  #endif
}


void Grob::makeSphereList(double radius)
{
  #ifdef GrobGL
  int lownum,mednum,highnum,finenum,paintnum;
  lownum=GrobGlHelper::makeSphere(radius,5,5,0,lowdetail);
  mednum=GrobGlHelper::makeSphere(radius,5,5,1,meddetail);
  highnum=GrobGlHelper::makeSphere(radius,9,9,1,highdetail);
  finenum=GrobGlHelper::makeSphere(radius,16,16,1,finedetail);
  paintnum=GrobGlHelper::makeSphere(radius,12,12,1,paintdetail);
  set_gl(lownum,mednum,highnum,finenum,paintnum);
  #endif
}


void Grob::makeCubeList(double xlength,double ylength,double zlength)
{
  #ifdef GrobGL
  int lownum,mednum,highnum,finenum,paintnum;
  lownum=GrobGlHelper::makeCube(xlength,ylength,zlength,0,lowdetail);
  mednum=GrobGlHelper::makeCube(xlength,ylength,zlength,1,meddetail);
  highnum=mednum;
  finenum=mednum;
  paintnum=GrobGlHelper::makeCube(xlength,ylength,zlength,1,paintdetail);
  set_gl(lownum,mednum,highnum,finenum,paintnum);
  #endif
}

void Grob::makeRoundCubeList(double xlength,double ylength,double zlength,double shift)
{
  #ifdef GrobGL
  int lownum,mednum,highnum,finenum,paintnum;
  lownum=GrobGlHelper::makeRoundCube(xlength,ylength,zlength,shift,0,lowdetail);
  mednum=GrobGlHelper::makeRoundCube(xlength,ylength,zlength,shift,1,meddetail);
  highnum=GrobGlHelper::makeRoundCube(xlength,ylength,zlength,shift,1,highdetail);
  finenum=GrobGlHelper::makeRoundCube(xlength,ylength,zlength,shift,1,finedetail);
  paintnum=GrobGlHelper::makeRoundCube(xlength,ylength,zlength,shift,1,paintdetail);
  set_gl(lownum,mednum,highnum,finenum,paintnum);
  #endif
}

void Grob::makeLegconeList(double pradius,double dradius, double length,int color,int sleeve,double offst,int endeffector)
{
  #ifdef GrobGL
  int lownum,mednum,highnum,finenum,paintnum;
  lownum=GrobGlHelper::makeLegcone(pradius,dradius,length,5,2,color/5,color/8,0,offst,endeffector,0,lowdetail);
  mednum=GrobGlHelper::makeLegcone(pradius,dradius,length,5,3,color/5,color/8,0,offst,endeffector,1,meddetail);
  highnum=GrobGlHelper::makeLegcone(pradius,dradius,length,8,4,color/5,color/8,sleeve,offst,endeffector,1,highdetail);
  finenum=GrobGlHelper::makeLegcone(pradius,dradius,length,14,8,color/5,color/8,sleeve,offst,endeffector,1,finedetail);
  paintnum=GrobGlHelper::makeLegcone(pradius,dradius,length,10,6,color/5,color/8,sleeve,offst,endeffector,1,paintdetail);
  set_gl(lownum,mednum,highnum,finenum,paintnum);
  #endif
}

void Grob::makeConeList(double radius, double length)
{
  #ifdef GrobGL
  int lownum,mednum,highnum,finenum,paintnum;
  lownum=GrobGlHelper::makeCone(radius,length,5,2,0,lowdetail);
  mednum=GrobGlHelper::makeCone(radius,length,5,3,1,meddetail);
  highnum=GrobGlHelper::makeCone(radius,length,8,4,1,highdetail);
  finenum=GrobGlHelper::makeCone(radius,length,14,8,1,finedetail);
  paintnum=GrobGlHelper::makeCone(radius,length,10,6,1,paintdetail);
  set_gl(lownum,mednum,highnum,finenum,paintnum);
  #endif
}



void Grob::makeTriangletileList(double p10,double p11,double p12,double p20,double p21,double p22,double p30,double p31,double p32,double thickness)
{
  #ifdef GrobGL
  int lownum,mednum,highnum,finenum,paintnum;
  lownum=GrobGlHelper::makeTritile(p10,p11,p12,p20,p21,p22,p30,p31,p32,thickness,0,lowdetail);
  mednum=GrobGlHelper::makeTritile(p10,p11,p12,p20,p21,p22,p30,p31,p32,thickness,1,meddetail);
  highnum=mednum;
  finenum=mednum;
  //  finenum=GrobGlHelper::makeTexturetile(p10,p11,p12,p20,p21,p22,p30,p31,p32,thickness,1,finedetail);
  set_gl(lownum,mednum,highnum,finenum,finenum);
  #endif
}



void Grob::makeArrowList(double length,int color)
{
  #ifdef GrobGL
  int lownum,mednum,highnum,finenum,paintnum;
  lownum=GrobGlHelper::makeArrow(length, 4,4, color/5,0,lowdetail);
  mednum=GrobGlHelper::makeArrow(length, 6,6, color/5,1,meddetail);
  highnum=GrobGlHelper::makeArrow(length, 9,9, color/5,1,highdetail);
  finenum=GrobGlHelper::makeArrow(length, 16,16, color/5,1,finedetail);
  paintnum=GrobGlHelper::makeArrow(length, 12,12, color/5,1,paintdetail);
  set_gl(lownum,mednum,highnum,finenum,paintnum);
  #endif
}



void Grob::makeAxesList(double viewradius,int linewidth)
{
  #ifdef GrobGL
  int lownum,mednum,highnum,finenum;
  lownum=GrobGlHelper::makeAxes(viewradius,lowdetail,linewidth);
  mednum=GrobGlHelper::makeAxes(viewradius,meddetail,linewidth);
  highnum=GrobGlHelper::makeAxes(viewradius,highdetail,linewidth);
  finenum=GrobGlHelper::makeAxes(viewradius,finedetail,linewidth);
  set_gl(lownum,mednum,highnum,finenum,finenum);
  #endif
}



void Grob::makeBarList(double xlength,double ylength,double zlength)
{
  #ifdef GrobGL
  int lownum,mednum,highnum,finenum,paintnum;
  lownum=GrobGlHelper::makeBar(xlength,ylength,zlength,lowdetail);
  mednum=GrobGlHelper::makeBar(xlength,ylength,zlength,meddetail);
  highnum=GrobGlHelper::makeBar(xlength,ylength,zlength,highdetail);
  finenum=GrobGlHelper::makeBar(xlength,ylength,zlength,finedetail);
  paintnum=GrobGlHelper::makeBar(xlength,ylength,zlength,paintdetail);
  set_gl(lownum,mednum,highnum,finenum,paintnum);
  #endif
}

















void Grob::replacelist(GrobDetail indetail,int newlistnum)
{
  // replace a specific detail list for the grob
  unusedetail(indetail);
  gl_listnum[(int)indetail]=newlistnum;
}

void Grob::unusedetail(GrobDetail indetail)
{
  // the detail list is no longer needed
  // so, IF it is not used by another detail level, it is cfpe to gldelete it
  int dnum=(int)indetail;
  int count;

  if (gl_listnum[dnum]!=-1)
    {
      // now null out any equiv. details
      for (count=0;count<GROBDETAILS;++count)
	{
	  if (count==dnum)
	    continue;
	  if (gl_listnum[count]==gl_listnum[dnum])
	    break;
	}
      if (count>=(int)GROBDETAILS)
	{
	  // no one else uses it, so it's cfpe to delete
	  GrobGlHelper::freelistnum(gl_listnum[dnum]);
	}
    }
  gl_listnum[dnum]=-1;
}





void Grob::replaceLow()
{
  // replace the wirefram lownum displaylist with a point
  // useful for grobs used in plots for which points are preferable to wireframes
  unusedetail(lowdetail);
  gl_listnum[(int)lowdetail]=GrobGlHelper::makePoint(3,lowdetail);
}

void Grob::replaceHigh()
{
  // replace highest res model with next lower down, used for objects we don't need so accurate
  unusedetail(finedetail);
  gl_listnum[(int)finedetail]=gl_listnum[(int)highdetail];
}

void Grob::makeAxesList(VisibleAxis *xaxisp,VisibleAxis *yaxisp,VisibleAxis *zaxisp,float cxscale,float cyscale,float czscale,GrobSpacerType usespacer,int linewidth,bool clearbox)
{
  #ifdef GrobGL
  int lownum,mednum,highnum,finenum;
  lownum=GrobGlHelper::makeAxes(xaxisp,yaxisp,zaxisp,lowdetail,cxscale,cyscale,czscale,linewidth,clearbox,usespacer);
  mednum=GrobGlHelper::makeAxes(xaxisp,yaxisp,zaxisp,meddetail,cxscale,cyscale,czscale,linewidth,clearbox,usespacer);
  highnum=GrobGlHelper::makeAxes(xaxisp,yaxisp,zaxisp,highdetail,cxscale,cyscale,czscale,linewidth,clearbox,usespacer);
  finenum=GrobGlHelper::makeAxes(xaxisp,yaxisp,zaxisp,finedetail,cxscale,cyscale,czscale,linewidth,clearbox,usespacer);
  set_gl(lownum,mednum,highnum,finenum,finenum);
  #endif
}




void Grob::drawline(double sx,double sy,double sz,double ex,double ey,double ez)
{
  #ifdef GrobGL
  // immediately draw a line
  GrobGlHelper::drawLine(sx,sy,sz,ex,ey,ez);
  #endif
}








// ----------------------------------------------------------------------------------
// -- Grob GL member functions
// ----------------------------------------------------------------------------------


// ----------------------------------------------------------------------------------
// -- Grob::gl_initialize()
// ----------------------------------------------------------------------------------

void Grob::gl_initialize()
{
  // initialize gl preparations
  detail=lowdetail;

  // displaylists: clear them all
  for (int count=0;count<GROBDETAILS;++count)
    gl_listnum[count]=-1;
  gl_currentnum=-1;

  // clear update flag
  gl_needsupdate=true;
}



// ----------------------------------------------------------------------------------
// -- Grob::gl_uninitialize()
// ----------------------------------------------------------------------------------

void Grob::gl_uninitialize()
{
  // free the displaylists
  int count,count2;

  for (count=0;count<GROBDETAILS;++count)
    unusedetail((GrobDetail)count);
}




void Grob::set_gl(GrobDetail indetail,int displaylistnum)
{
  // set the displaylistnum of the specified detail level (or all, or all undefined)

  if (indetail==lowdetail || gl_listnum[(int)lowdetail]==-1)
    gl_listnum[(int)lowdetail]=displaylistnum;
  if (indetail==meddetail || gl_listnum[(int)meddetail]==-1)
    gl_listnum[(int)meddetail]=displaylistnum;
  if (indetail==highdetail || gl_listnum[(int)highdetail]==-1)
    gl_listnum[(int)highdetail]=displaylistnum;
  if (indetail==finedetail || gl_listnum[(int)finedetail]==-1)
    gl_listnum[(int)finedetail]=displaylistnum;
  if (indetail==paintdetail || gl_listnum[(int)paintdetail]==-1)
    gl_listnum[(int)paintdetail]=displaylistnum;

  if (detail==indetail || detail==-1 || gl_currentnum==-1)
    {
      detail=indetail;
      gl_currentnum=displaylistnum;
    }
}


void Grob::set_gl(int inlow,int inmed,int inhigh, int infine,int inpaint)
{
  // set the displaylistnums of all detail levels
  gl_listnum[(int)lowdetail]=inlow;
  gl_listnum[(int)meddetail]=inmed;
  gl_listnum[(int)highdetail]=inhigh;
  gl_listnum[(int)finedetail]=infine;
  gl_listnum[(int)paintdetail]=inpaint;

  if (gl_currentnum==-1)
    detail=lowdetail;

  gl_currentnum=gl_listnum[(int)detail];
}






// ----------------------------------------------------------------------------------
// -- Grob::ready_gl()
// ----------------------------------------------------------------------------------
bool Grob::ready_gl()
{
  if (!using_gl)
    return false;

  if (gl_currentnum==-1)
    return false;
  return true;
}









































// ----------------------------------------------------------------------------------
// -- Class GrobGlHelper performs GL related displaylist building
// ----------------------------------------------------------------------------------



#ifdef GrobGL




// ---------------------------------------------------------------------------------------------
// -- Static helper functions which i think are useful in general
// ---------------------------------------------------------------------------------------------


int GrobGlHelper::getlistnum()
{
  // allocate a new displaylist num
  int retv;
  static int glb_listnum=100;

  // GL, at least my version Mesa, is fucked up in that it won't give us any listnums
  // or let us create displayslists unless we have a GL window up.
  // furthermore, to make matters worse, X sometimes takes time to put up windows
  //  and so just making a window prior to trying to allocate listnums is not sufficient.
  // SO, we rely on a two-part trick.. first we insist that our gui manager make a GL window,
  //  that's handled elsewhere.  Here we wait for it to work for a couple of seconds
  for (int count=1;count<=10;++count)
    {
      retv=(int)glGenLists(1);
      if (retv)
	break;
      sleep(1);
    }

  if (!retv)
    {
      cout << "ERROR: gl says there are no more displaylists to allocate.\n     This can happen if a GL window is not created prior to creating shapes.\n";
      //      retv=++glb_listnum;
    }
  return retv;
}


void GrobGlHelper::freelistnum(int listnum)
{
  // free the listnum
  glDeleteLists(listnum,1);
}





void GrobGlHelper::snifferrors(char *announce)
{
  // check for any GL errors
  GLenum error;

  printf("$sniffing for GL errors during '%s'\n",announce);
  while ((error=glGetError())!= GL_NO_ERROR)
    {
      printf("$GL ERROR: %s\n",gluErrorString(error));
    }
}


void GrobGlHelper::SetMaterial(GLfloat spec[], GLfloat amb[], GLfloat diff[], GLfloat shin[])
{
  glMaterialfv(GL_FRONT, GL_SPECULAR, spec);
  glMaterialfv(GL_FRONT, GL_SHININESS, shin);
  glMaterialfv(GL_FRONT, GL_AMBIENT, amb);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, diff);
  glColor4fv(amb);
}



// ---------------------------------------------------------------------------------------------
// -- Static helper functions which i think should be fixed
// ---------------------------------------------------------------------------------------------



void GrobGlHelper::setuplights()
{
  GLfloat lightOneColor[] = {0.4, 1.0, 0.4, 1.0}; /* green-tinted */
  GLfloat lightZeroColor[] = {0.2, 0.2, 0.9, 0.5};  /* red-tinted */
  GLfloat noColor[] = {0, 0, 0, 1};  /* ? */
  GLfloat tinyColor[] = {.1, .1, .1, .1};  /* ? */
  GLfloat smallColor[] = {.4, .4, .4, .4};  /* ? */
  GLfloat someColor[] = {.6, .6, .6, .6};  /* ? */

  glEnable(GL_DEPTH_TEST); /* Enable depth buffering. */

  // deluxe light - just some random colors
  //  GLfloat ambient[] = {0.2, 0.5, 0.7, .2};
  //  GLfloat diffuse[] = {.6, .4, .5, .2};

  GLfloat specular1[] = {1, 1, 1, 1};
  GLfloat specular2[] = {.1, .1, .1, 1};
  GLfloat specular3[] = {0, 0, 0, 1};
  glLightfv(GL_LIGHT0,GL_AMBIENT,noColor);
  glLightfv(GL_LIGHT0,GL_DIFFUSE,smallColor);
  glLightfv(GL_LIGHT0,GL_SPECULAR,specular1);
  glLightfv(GL_LIGHT1,GL_AMBIENT,noColor);
  glLightfv(GL_LIGHT1,GL_DIFFUSE,someColor);
  glLightfv(GL_LIGHT1,GL_SPECULAR,specular1);

  glLightfv(GL_LIGHT2,GL_AMBIENT,noColor);
  glLightfv(GL_LIGHT2,GL_DIFFUSE,smallColor);
  glLightfv(GL_LIGHT2,GL_SPECULAR,specular1);
  glLightfv(GL_LIGHT3,GL_AMBIENT,noColor);
  glLightfv(GL_LIGHT3,GL_DIFFUSE,smallColor);
  glLightfv(GL_LIGHT3,GL_SPECULAR,specular1);
  glLightfv(GL_LIGHT4,GL_AMBIENT,noColor);
  glLightfv(GL_LIGHT4,GL_DIFFUSE,smallColor);
  glLightfv(GL_LIGHT4,GL_SPECULAR,specular1);

  /* Set up projection transform. */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  // Dummy push so we can pop on model recalc.
  glPushMatrix();
}








// ---------------------------------------------------------------------------------------------
// -- More static helper functions which could be useful for making shapes
// -- They use GLUT and also need to be made more general w.r.t. colors and materials
// -- They return ints refering to the new listnum they created
// ---------------------------------------------------------------------------------------------





int GrobGlHelper::startlist()
{
  // allocate and begin a new gl displaylist
  int dlnum=getlistnum();
  pushlighting();
  glPushMatrix();
  glLoadIdentity();
  glNewList(dlnum,GL_COMPILE);
  return dlnum;
}

void GrobGlHelper::endlist()
{
  // allocate and begin a new gl displaylist
  //  glPopAttrib();
  glEndList();
  glPopMatrix();
  poplighting();
}



int GrobGlHelper::makePoint(int size,GrobDetail detail)
{
  int displaylistnum=startlist();
  drawPoint(size,detail);
  endlist(); return displaylistnum;
}


int GrobGlHelper::makeSphere(double radius,int slices,int stacks, char solid,GrobDetail detail)
{
  int displaylistnum=startlist();
  lightson();
  drawSphere(radius,slices,stacks,solid,detail);
  endlist(); return displaylistnum;
}



int GrobGlHelper::makeCube(double xlength,double ylength,double zlength,char solid,GrobDetail detail)
{
  int displaylistnum=startlist();
  lightson();
  drawCube(xlength,ylength,zlength,solid,detail);
  endlist(); return displaylistnum;
}

int GrobGlHelper::makeRoundCube(double xlength,double ylength,double zlength,double shift,char solid,GrobDetail detail)
{
  int displaylistnum=startlist();
  lightson();
  drawRoundCube(xlength,ylength,zlength,shift,solid,detail);
  endlist(); return displaylistnum;
}



int GrobGlHelper::makeLegcone(double proximal_radius,double distal_radius, double length, int slices,int stacks,double color, double color2, int sleeve,double offst, int endeffector, char solid,GrobDetail detail)
{
  int displaylistnum=startlist();
  lightson();
  drawLegcone(proximal_radius,distal_radius,length,slices,stacks,color,color2,sleeve,offst,endeffector,solid,detail);
  endlist(); return displaylistnum;
}




int GrobGlHelper::makeTritile(double x1, double y1,double z1,double x2,double y2,double z2,double x3,double y3, double z3,double thickness, char solid,GrobDetail detail)
{
  int displaylistnum=startlist();
  lightson();
  drawTritile(x1,y1,z1,x2,y2,z2,x3,y3,z3,thickness,solid,detail);
  endlist(); return displaylistnum;
}



int GrobGlHelper::makeTexturetile(double x1, double y1,double z1,double x2,double y2,double z2,double x3,double y3, double z3,double thickness, char solid,GrobDetail detail)
{
  // based on texture code form the Redbook, a not all-together satisfying textured 3d tile
  int displaylistnum=startlist();
  lightson();
  drawTexturetile(x1,y1,z1,x2,y2,z2,x3,y3,z3,thickness,solid,detail);
  endlist(); return displaylistnum;
}



int GrobGlHelper::makeCone(double radius,double length,int slices,int stacks,char solid,GrobDetail detail)
{
  double color1=1;
  double color2=1;
  int displaylistnum=startlist();
  lightson();
  drawCone(radius,length,slices,stacks,color1,color2,solid,detail);
  endlist(); return displaylistnum;
}



int GrobGlHelper::makeArrow(double length, int slices,int stacks, int color, char solid,GrobDetail detail)
{
  int displaylistnum=startlist();
  lightson();
  drawArrow(length,slices,stacks,color,solid,detail);
  endlist(); return displaylistnum;
}


int GrobGlHelper::makeLine(double length)
{
  int displaylistnum=startlist();
  lightsoff();
  glBegin(GL_LINES);
  glLineWidth(1);
  glVertex3f(0, 0, 0); glVertex3f(length, 0, 0);
  glEnd();
  endlist(); return displaylistnum;
}



int GrobGlHelper::makeBar(double xlength,double ylength,double zlength,GrobDetail detail)
{
  int displaylistnum=startlist();
  lightson();
  drawBar(xlength,ylength,zlength,detail);
  endlist(); return displaylistnum;
}


























// ---------------------------------------------------------------------------------------------
// -- Static helper functions which could be useful for making shapes
// -- They use GLUT and also need to be made more general w.r.t. colors and materials
// -- They draw directly and are used by other funcs which make displaylists
// ---------------------------------------------------------------------------------------------


void GrobGlHelper::drawPoint(int size,GrobDetail detail)
{
  pushlighting();
  lightsoff();
  glPointSize(size);
  glBegin(GL_POINTS);
  if (detail!=paintdetail)
    glColor3f(.9,.9,1.0);
  glVertex3f(0, 0, 0);
  glEnd();
  glPointSize(1);
  poplighting();
}

void GrobGlHelper::drawPoint(float x, float y, float z,GrobDetail detail)
{
  glBegin(GL_POINTS);
  if (detail!=paintdetail)
    glColor3f(1,1,1);
  glVertex3f(x, y, z);
  glEnd();
}



void GrobGlHelper::drawSphere(double radius,int slices,int stacks, char solid,GrobDetail detail)
{
  //  Color
  //  GLfloat triColor2[] = {0.2, 0.4, 0.7, .7};
  //  GLfloat specular[] = {1, 1, 1, 1};
  //  GLfloat shininess[] = {128};
  //  GLfloat triColor2[] = {0.3, 0.4, 0.8, .9};
  //  GLfloat triColor2[] = {0.5, 0.4, 0.9, 1};

  GLfloat specular[] = {.7, .7, .7, .8};
  GLfloat shininess[] = {75};
  GLfloat triColor2[] = {0.8, 0.6, .1, .2};

  if (detail!=paintdetail)
    {
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor2);
      glColor4fv(triColor2);
    }
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);
  
  if (solid==1)
    glutSolidSphere(radius,slices,stacks);
  else
    glutWireSphere(radius,slices,stacks);
}




void GrobGlHelper::drawCube(double xlength,double ylength,double zlength,char solid,GrobDetail detail)
{
  // Color
  //  GLfloat triColor2[] = {0.5, 1, 0.5, 1}; // green
  GLfloat triColor2[] = {.3, .5, .4, 1};
  GLfloat specular[] = {.3, .3, .3, .3};
  GLfloat shininess[] = {75};

  if (detail!=paintdetail)
    {
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor2);
      glColor4fv(triColor2);
    }
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

  drawMechBox(xlength,ylength,zlength,solid);
}




void GrobGlHelper::drawLegcone(double proximal_radius,double distal_radius, double length, int slices,int stacks,double color, double color2, int sleeve,double offst, int endeffector, char solid,GrobDetail detail)
{
  // Color
  GLfloat triColor[] = {0.1, 0.6, 1, 1};
  GLfloat specular[] = {1, 1, 1, 1};
  GLfloat shininess[] = {100};

  // end effectors not drawn at lowdetail
  if (detail==lowdetail)
    endeffector=0;

  if (detail!=paintdetail)
    {
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor);
      glColor4fv(triColor);
    }
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);
  
  // draw the cone
  glRotatef(90.0,0.0,1.0,0.0);
  drawMechCylinder(proximal_radius,distal_radius,length,slices,stacks,solid,color,endeffector);
  
  if (detail!=paintdetail)
    {
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor);
      glColor4fv(triColor);
    }
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

  if (offst!=0.0)
    {
      glRotatef(90.0,0.0,1.0,0.0);
      // offset link
      drawMechCylinder(proximal_radius,proximal_radius,offst,slices,stacks,solid,1,0);

      if (sleeve==0 && solid)
      	{
      	  glTranslatef(0,0,offst);
      	  drawSphere(proximal_radius,slices,stacks,solid,detail);
      	  glTranslatef(0,0,-offst);
      	}
    }

  GLfloat triColor2[] = {0.5, 0.4, 0.9, 1};
  GLfloat specular2[] = {.4, .4, .4, .4};
  GLfloat shininess2[] = {20};
  if (detail!=paintdetail)
    {
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor2);
      glColor4fv(triColor2);
    }
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular2);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess2);

  if (sleeve!=0)
    {
      // proximal sleeve
      glTranslatef(0,0,-proximal_radius*1.1);
      drawMechCylinder(proximal_radius*1.2,proximal_radius*1.2,proximal_radius*2.2,slices,stacks,solid,color2,3);
      glTranslatef(0,0,proximal_radius*1.1);

      // only one of the following two should be used, both here for experimentation
      if (offst!=0)
	{
	  // proximal offset sleeve
	  // this puts a sleeve at the base of first link
	  glTranslatef(0,0,offst);
	  glTranslatef(0,0,-proximal_radius*1.1);
	  drawMechCylinder(proximal_radius*1.2,proximal_radius*1.2,proximal_radius*2.2,slices,stacks,solid,color2,3);
	  glTranslatef(0,0,proximal_radius);
	  glTranslatef(0,0,-offst);
	}
      if (offst && !endeffector && false)
	{
	  // distal offset sleeve
	  // this fails to put a sleeve at first base, but looks more mechanically
	  //  accurate at highdetail, beacuse joints revolve better.
	  glTranslatef(-length,0,0);
	  glTranslatef(0,0,-distal_radius*1.1);
	  drawMechCylinder(distal_radius*1.2,distal_radius*1.2,distal_radius*2.2,slices,stacks,solid,color2,3);
	  glTranslatef(0,0,distal_radius*1.1);
	  glTranslatef(length,0,0);
	}
    }
  else
    {
      if (solid)
	{
	  //	  drawSphere(proximal_radius,slices,stacks,solid,detail);
	  drawSphere(proximal_radius*1.5,slices,stacks,solid,detail);
	}
    }
}






void GrobGlHelper::drawTriangle(double x1, double y1,double z1,double x2,double y2,double z2,double x3,double y3, double z3,double thickness, char solid,GrobDetail detail)
{
  // make a 3d tile
  char i, j = 0;
  float depthplus=.0001;

  //  GLfloat topColor[] = {0.55, 0.55, 0.6, 1};
  GLfloat wireColor[] = {1, 1, 1, 1};
  GLfloat topColor[] = {0.85, 0.25, 0.4, .5};
  GLfloat triColor[] = {0.8, 0.8, 0.8, 1};
  //  GLfloat specular[] = {0.1, 0.1, 0.1, 1};
  //  GLfloat shininess[] = {1.0};
  GLfloat specular[] = {1, 1, 1, 1};
  GLfloat shininess[] = {100};

  if (detail!=paintdetail)
    glColor4fv(wireColor);

  // The wire border around the triangle
  glBegin(GL_LINE_LOOP);
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, wireColor);
  glNormal3f(-1.0, 0.0, 0.0);
  glVertex3f(x1, y1, z1);
  glVertex3f(x2, y2, z2);
  glVertex3f(x3, y3, z3);
  glEnd();

  if (!solid)
    return;

  // now the solid tile
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

  glBegin(GL_TRIANGLES);
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, topColor);
  glNormal3f(-1.0, 0.0, 0.0);
  glVertex3f(x1, y1, z1+depthplus);
  glVertex3f(x2, y2, z2+depthplus);
  glVertex3f(x3, y3, z3+depthplus);
  glEnd();
}




void GrobGlHelper::drawTritile(double x1, double y1,double z1,double x2,double y2,double z2,double x3,double y3, double z3,double thickness, char solid,GrobDetail detail)
{
  // make a 3d tile
  char i, j = 0;

  //  GLfloat topColor[] = {0.2, 0.2, 0.2, 1};
  //  GLfloat sideColor1[] = {0.4, 0.2, 0.2, 1};
  //  GLfloat sideColor2[] = {0.1, 0.6, 0.2, 1};
  //  GLfloat triColor[] = {0.4, 0.2, 0.2, 1};
  //  GLfloat specular[] = {0.3, 0.3, 0.3, 1};
  //  GLfloat shininess[] = {1.0};

  GLfloat topColor[] = {0.55, 0.55, 0.6, 1};
  GLfloat sideColor1[] = {0.40, 0.45, 0.45, 1};
  GLfloat sideColor2[] = {0.4, 0.4, 0.55, 1};
  GLfloat triColor[] = {0.8, 0.8, 0.8, 1};
  GLfloat specular[] = {0.1, 0.1, 0.1, 1};
  GLfloat shininess[] = {1.0};

  // new - lower floor
  glTranslatef(0,0,thickness);
  
  if (detail!=paintdetail)
    {
      //      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor);
      glColor4fv(triColor);
    }
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

  //  glRotatef(90.0, 0.0, 0.0, 1.0);

  // TOP
  if (!solid)
    glBegin(GL_LINE_LOOP);
  else
    glBegin(GL_TRIANGLES);
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, topColor);
  glNormal3f(-1.0, 0.0, 0.0);
  glVertex3f(x1, y1, z1);
  glVertex3f(x2, y2, z2);
  glVertex3f(x3, y3, z3);
  glEnd();

  // BOTTOM
  if (!solid)
    glBegin(GL_LINE_LOOP);
  else
    glBegin(GL_TRIANGLES);
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, topColor);
  glNormal3f(1.0, 0.0, 0.0);
  glVertex3f(x1, y1, z1+thickness);
  glVertex3f(x2, y2, z2+thickness);
  glVertex3f(x3, y3, z3+thickness);
  glEnd();

  // SIDE1
  if (!solid)
    glBegin(GL_LINE_LOOP);
  else
    glBegin(GL_QUADS);
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, sideColor1);
  glNormal3f(0, 0, -1);
  glVertex3f(x1, y1, z1);
  glVertex3f(x2, y2, z2);
  glVertex3f(x2, y2, z2+thickness);
  glVertex3f(x1, y1, z1+thickness);
  glEnd();

  // SIDE2
  if (!solid)
    glBegin(GL_LINE_LOOP);
  else
    glBegin(GL_QUADS);
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, sideColor2);
  glNormal3f(0.0, 0.0, -1);
  glVertex3f(x2, y2, z2);
  glVertex3f(x3, y3, z3);
  glVertex3f(x3, y3, z3+thickness);
  glVertex3f(x2, y2, z2+thickness);
  glEnd();

  // SIDE3
  if (!solid)
    glBegin(GL_LINE_LOOP);
  else
    glBegin(GL_QUADS);
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, sideColor2);
  glNormal3f(0.0, 0.0, -1);
  glVertex3f(x3, y3, z3);
  glVertex3f(x1, y1, z1);
  glVertex3f(x1, y1, z1+thickness);
  glVertex3f(x3, y3, z3+thickness);
  glEnd();

  glTranslatef(0,0,-thickness);
}




void GrobGlHelper::drawTexturetile(double x1, double y1,double z1,double x2,double y2,double z2,double x3,double y3, double z3,double thickness, char solid,GrobDetail detail)
{
  // based on texture code form the Redbook, a not all-together satisfying textured 3d tile
  // just a test!
  // very INEFFICIENT and written when i didn't know opengl at all - should be REPLACED!
  GLfloat topColor[] = {0.2, 0.2, 0.2, 0.2};
  GLfloat sideColor1[] = {0.4, 0.2, 0.2, 0.5};
  GLfloat sideColor2[] = {0.1, 0.6, 0.2, 0.5};
  GLfloat triColor[] = {0.4, 0.2, 0.2, 0.5};
  GLfloat specular[] = {0.3, 0.3, 0.3, 0.3};
  GLfloat shininess[] = {5.0};
  int checkImageWidth=64;
  int checkImageHeight=64;
  GLubyte checkImage[checkImageHeight][checkImageWidth][4];
  GLubyte otherImage[checkImageHeight][checkImageWidth][4];
  GLuint texName[2];
  
  int i, j, c;
  for (i = 0; i < checkImageHeight; i++) {
    for (j = 0; j < checkImageWidth; j++) {
      c = ((((i&0x8)==0)^((j&0x8))==0))*180;
      c+=74;
      checkImage[i][j][0] = (GLubyte) c;
      checkImage[i][j][1] = (GLubyte) c;
      checkImage[i][j][2] = (GLubyte) c;
      checkImage[i][j][3] = (GLubyte) 180;
      c = ((((i&0x10)==0)^((j&0x10))==0))*180;
      c+=74;
      otherImage[i][j][0] = (GLubyte) c;
      otherImage[i][j][1] = (GLubyte) 74;
      otherImage[i][j][2] = (GLubyte) 74;
      otherImage[i][j][3] = (GLubyte) 180;
    }
  }

  // Now the init part
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  
  glGenTextures(2, texName);
  glBindTexture(GL_TEXTURE_2D, texName[0]);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, checkImageWidth,checkImageHeight,0,GL_RGBA,GL_UNSIGNED_BYTE,checkImage);
  
  glBindTexture(GL_TEXTURE_2D, texName[1]);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  
  // i don't know (can't tell) the dif. between linear and nearest
  //   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  //   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, checkImageWidth,checkImageHeight,0,GL_RGBA,GL_UNSIGNED_BYTE,otherImage);
  
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glEnable(GL_TEXTURE_2D);
  
  // New Colors
  if (detail!=paintdetail)
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);
  
  //  glRotatef(90.0, 0.0, 0.0, 1.0);
  
  // TOP
  glBindTexture(GL_TEXTURE_2D, texName[0]);
  if (!solid)
    glBegin(GL_LINE_LOOP);
  else
    glBegin(GL_TRIANGLES);

  glNormal3f(-1.0, 0.0, 0.0);
  glTexCoord2f(0.0, 0.0);
  glVertex3f(x1, y1, z1);
  glTexCoord2f(2, 5);
  glVertex3f(x3, y3, z3);
  glTexCoord2f(4, 0.0);
  glVertex3f(x2, y2, z2);
  glEnd();

  glFlush();
  glDisable(GL_TEXTURE_2D);
  
  // BOTTOM
  if (!solid)
    glBegin(GL_LINE_LOOP);
  else
    glBegin(GL_TRIANGLES);
  if (detail!=paintdetail)
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, topColor);
  glNormal3f(1.0, 0.0, 0.0);
  glVertex3f(x1, y1, z1+thickness);
  glVertex3f(x2, y2, z2+thickness);
  glVertex3f(x3, y3, z3+thickness);
  glEnd();

  // SIDE1
  if (!solid)
    glBegin(GL_LINE_LOOP);
  else
    glBegin(GL_QUADS);
  if (detail!=paintdetail)
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, sideColor1);
  glNormal3f(0, 0, -1);
  glVertex3f(x1, y1, z1);
  glVertex3f(x2, y2, z2);
  glVertex3f(x2, y2, z2+thickness);
  glVertex3f(x1, y1, z1+thickness);
  glEnd();

  // SIDE2
  if (!solid)
    glBegin(GL_LINE_LOOP);
  else
    glBegin(GL_QUADS);
  if (detail!=paintdetail)
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, sideColor2);
  glNormal3f(0.0, 0.0, -1);
  glVertex3f(x2, y2, z2);
  glVertex3f(x3, y3, z3);
  glVertex3f(x3, y3, z3+thickness);
  glVertex3f(x2, y2, z2+thickness);
  glEnd();

  // SIDE3
  if (!solid)
    glBegin(GL_LINE_LOOP);
  else
    glBegin(GL_QUADS);
  if (detail!=paintdetail)
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, sideColor2);
  glNormal3f(0.0, 0.0, -1);
  glVertex3f(x3, y3, z3);
  glVertex3f(x1, y1, z1);
  glVertex3f(x1, y1, z1+thickness);
  glVertex3f(x3, y3, z3+thickness);
  glEnd();
}







void GrobGlHelper::drawCone(double radius,double length,int slices,int stacks,double color1,double color2,char solid,GrobDetail detail)
{
  GLUquadricObj *upper = gluNewQuadric();
  GLfloat triColor[] = {0.8, 0.6, .1, 1};
  GLfloat specular[] = {.2, .2, .2, 1};
  GLfloat shininess[] = {256};
  GLfloat brightColor[] = {.9, .9, .9, 1};
  
  // Color
  if (detail!=paintdetail)
    {
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor);
      glColor4fv(triColor);
    }
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);
  
  // draw the cone
  glTranslatef(0,0,-length/2);
  
  if (solid==0)
    gluQuadricDrawStyle(upper, (GLenum) GLU_LINE);
  
  if (solid==1)
    glutSolidCone(radius,length,slices,stacks);
  else
    glutWireCone(radius,length,slices,stacks);

  glTranslatef(0,0,length/2);
}




void GrobGlHelper::drawArrow(double length, int slices,int stacks, int color, char solid,GrobDetail detail)
{
  // NOTE: this particular function is does not work right yet
  double radius=length/25;
  GLfloat triColor[] = {0.1, 0.6, 1, 1};
  GLfloat specular[] = {1, 1, 1, 1};
  GLfloat shininess[] = {100};
  
  if (detail!=paintdetail)
    {
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor);
      glColor4fv(triColor);
    }

  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);
  
  // draw the cylinder with cone endcap
  glRotatef(90.0,0.0,1.0,0.0);
  drawMechCylinder(radius,radius,length,slices,stacks,solid,color,4);
}




void GrobGlHelper::drawBar(double xlength,double ylength,double zlength,GrobDetail detail)
{
  // draw a bar at different levels of detail
  // high/fine detail is just a bar of material
  // but med detail is a wire frame with hidden surface removal (slower then highdetail, naughty)
  // and low detail is a compromise, it's a simple line, but capped with a hiddden surfaceremoved square
  //  these higher-than-usual costs for lowdetail are nesc. because otherwise bars are just too obscured to see.
  GLfloat barColor[] = {0.8, 0.6, .1, 1};
  GLfloat specular[] = {.2, .2, .2, 1};
  GLfloat shininess[] = {256};
  GLfloat noshininess[] = {0};
  //  GLfloat capColor[] = {.8, .3, .3, 1};
  GLfloat capColor[] = {0.8, 0.6, .1, 1};
  //  GLfloat brightColor[] = {1, .9, .8, 1};
  GLfloat brightColor[] = {.9, .9, .9, 1};
  GLfloat darkColor[] = {.3, .2, .1, 1};

  if (detail==lowdetail)
    {
      glPushMatrix();

      // just draw the top slice of the bar with a line underneath
      // draw the line underneath
      drawLine(0,0,zlength/2,0,0,-zlength/2);
      glTranslatef(xlength/-2,ylength/-2,zlength/2);
      depthoverwrite();

      // draw the solid square cap
      if (detail!=paintdetail)
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, capColor);
      glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
      glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, noshininess);
      drawSquare(xlength,ylength,1);

      // draw a border
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, brightColor);
      glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
      glColor4fv(brightColor);
      drawSquare(xlength,ylength,0);

      // put state back where we found it
      depthnormal();
      glPopMatrix();
    }
  else if (detail==meddetail)
    {
      // we want to make a hidden surface removed bar with border
      // first draw interior with cap color
      if (detail!=paintdetail)
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, capColor);
      glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
      glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, noshininess);
      drawMechBox(xlength,ylength,zlength,1);
      // draw a border, going into overwrite mode
      depthoverwrite();
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, brightColor);
      glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
      glColor4fv(brightColor);
      drawMechBox(xlength,ylength,zlength,0);
      depthnormal();
    }
  else
    {
      if (detail!=paintdetail)
	{
	  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, barColor);
	  glColor4fv(barColor);
	}
      glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
      glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);
      drawMechBox(xlength,ylength,zlength,1);
      // do we want a border
      // draw a border, going into overwrite mode
      depthoverwrite();
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, brightColor);
      glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
      glColor4fv(brightColor);
      drawMechBox(xlength,ylength,zlength,0);
      depthnormal();
    }
}




























void GrobGlHelper::drawMechBox(double width, double height, double depth, char solid)
{
  // based on glutmech.c, author	: Simon Parkinson-Bates.
  char i, j = 0;
  float x = width / 2.0, y = height / 2.0, z = depth / 2.0;

  for (i = 0; i < 4; i++) {
    glRotatef(90.0, 0.0, 0.0, 1.0);
    if (j) {
      if (!solid)
        glBegin(GL_LINE_LOOP);
      else
        glBegin(GL_QUADS);
      glNormal3f(-1.0, 0.0, 0.0);
      glVertex3f(-x, y, z);
      glVertex3f(-x, -y, z);
      glVertex3f(-x, -y, -z);
      glVertex3f(-x, y, -z);
      glEnd();
      if (solid) {
        glBegin(GL_TRIANGLES);
        glNormal3f(0.0, 0.0, 1.0);
        glVertex3f(0.0, 0.0, z);
        glVertex3f(-x, y, z);
        glVertex3f(-x, -y, z);
        glNormal3f(0.0, 0.0, -1.0);
        glVertex3f(0.0, 0.0, -z);
        glVertex3f(-x, -y, -z);
        glVertex3f(-x, y, -z);
        glEnd();
      }
      j = 0;
    } else {
      if (!solid)
        glBegin(GL_LINE_LOOP);
      else
        glBegin(GL_QUADS);
      glNormal3f(-1.0, 0.0, 0.0);
      glVertex3f(-y, x, z);
      glVertex3f(-y, -x, z);
      glVertex3f(-y, -x, -z);
      glVertex3f(-y, x, -z);
      glEnd();
      if (solid) {
        glBegin(GL_TRIANGLES);
        glNormal3f(0.0, 0.0, 1.0);
        glVertex3f(0.0, 0.0, z);
        glVertex3f(-y, x, z);
        glVertex3f(-y, -x, z);
        glNormal3f(0.0, 0.0, -1.0);
        glVertex3f(0.0, 0.0, -z);
        glVertex3f(-y, -x, -z);
        glVertex3f(-y, x, -z);
        glEnd();
      }
      j = 1;
    }
  }
}






void GrobGlHelper::drawMechBox2(double width, double height, double depth, char solid)
{
  // based on glutmech.c, author	: Simon Parkinson-Bates.
  char i, j = 0;
  float x = width, y= height, z = depth;

  for (i = 0; i < 4; i++) {
    glRotatef(90.0, 0.0, 0.0, 1.0);
    if (j) {
      if (!solid)
        glBegin(GL_LINE_LOOP);
      else
        glBegin(GL_QUADS);
      glNormal3f(-1.0, 0.0, 0.0);
      glVertex3f(0, y, z);
      glVertex3f(0, 0, z);
      glVertex3f(0, 0, 0);
      glVertex3f(0, y, 0);
      glEnd();
      if (solid) {
        glBegin(GL_TRIANGLES);
        glNormal3f(0.0, 0.0, 1.0);
        glVertex3f(0.0, 0.0, z);
        glVertex3f(0, y, z);
        glVertex3f(0, 0, z);
        glNormal3f(0.0, 0.0, -1.0);
        glVertex3f(0.0, 0.0, 0);
        glVertex3f(0, 0, 0);
        glVertex3f(0, y, 0);
        glEnd();
      }
      j = 0;
    } else {
      if (!solid)
        glBegin(GL_LINE_LOOP);
      else
        glBegin(GL_QUADS);
      glNormal3f(-1.0, 0.0, 0.0);
      glVertex3f(0, x, z);
      glVertex3f(0, 0, z);
      glVertex3f(0, 0, 0);
      glVertex3f(0, x, 0);
      glEnd();
      if (solid) {
        glBegin(GL_TRIANGLES);
        glNormal3f(0.0, 0.0, 1.0);
        glVertex3f(0.0, 0.0, z);
        glVertex3f(0, x, z);
        glVertex3f(0, 0, z);
        glNormal3f(0.0, 0.0, -1.0);
        glVertex3f(0.0, 0.0, 0);
        glVertex3f(0, 0, 0);
        glVertex3f(0, x, 0);
        glEnd();
      }
      j = 1;
    }
  }
}






void GrobGlHelper::drawMechCylinder(double baseradius,double topradius,double height,int slices,int stacks,char solid,float color, int endeffector)
{
  // based on glutmech.c, author	: Simon Parkinson-Bates.
  GLUquadricObj *upper = gluNewQuadric();
  GLfloat triColor2[] = {0.2, 0.8, .6, 1};
  int i;

  if (solid==0)
    gluQuadricDrawStyle(upper, (GLenum) GLU_LINE);

  if (endeffector==1)
    {
      // cut height so total height with tip is the same)
      height-=topradius*2;
    }

  gluCylinder(upper, baseradius, topradius, height, slices, stacks);

  // cap the ends of the cylinder
  if (solid)
    drawDisk(upper, 0, baseradius, slices,stacks,solid);

  if (endeffector)
    {
      // end-effector

      glTranslatef(0,0,height);
      if (endeffector==1)
	{
	  // a cone cap
	  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor2);
	  glColor4fv(triColor2);

	  if (solid==1)
	    glutSolidCone(topradius,topradius*2,slices,stacks);
	  else
	    glutWireCone(topradius,topradius*2,slices,stacks);
	}
      else if (endeffector==2)
	{
	  // a large radius disk cap
	  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor2);
	  glColor4fv(triColor2);

	  glRotatef(180,0,1,0);
	  drawDisk(upper, 0, topradius*2, slices,stacks,solid);
	  glRotatef(-180,0,1,0);
	}
      else if (endeffector==3)
	{
	  // just simple cap the disk otherwise it's hollow
	  drawDisk(upper, 0, topradius, slices,stacks,solid);
	}
      else if (endeffector==4)
	{
	  // a large cone cap?
	  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor2);
	  glColor4fv(triColor2);

	  if (solid==1)
	    glutSolidCone(topradius*2,topradius*.5,slices,stacks);
	  else
	glutWireCone(topradius*2,topradius*.5,slices,stacks);
	}
      else if (endeffector==5)
	{
	  // a large radius disk cap
	  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor2);
	  glColor4fv(triColor2);

	  glRotatef(180,0,1,0);
	  glTranslatef(topradius,0,0);
	  drawDisk(upper, 0, topradius*2, slices,stacks,solid);
	  glTranslatef(-topradius,0,0);
	  glRotatef(-180,0,1,0);
	}
      else if (endeffector==6)
	{
	  // kludge for dif. orientations of legs 3/4/99 - jesse unify 5&6
	  // a large radius disk cap
	  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor2);
	  glColor4fv(triColor2);

	  glRotatef(180,0,1,0);
	  glTranslatef(-topradius,0,0);
	  drawDisk(upper, 0, topradius*2, slices,stacks,solid);
	  glTranslatef(topradius,0,0);
	  glRotatef(-180,0,1,0);
	}
      glTranslatef(0,0,-1*height);
    }
}





void GrobGlHelper::drawDisk(GLUquadricObj *upper,double inradius,double baseradius,int slices,int stacks, char solid)
{
  gluDisk(upper, 0, baseradius, slices,stacks);
}


void GrobGlHelper::drawLinearrow(double x1,double y1,double z1,double x2,double y2,double z2, int hatdetail)
{
  // Draw a GL arrow
  GLfloat forceColor[] = {1.0, 0.0, 0.0, 1};
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, forceColor);
  glColor3f(1.0,0.0,0.0);

  // first the line
  glLineWidth(2);
  glBegin(GL_LINES);
  glVertex3f(x1,y1,z1);
  glVertex3f(x2,y2,z2);
  glEnd();

  if (hatdetail!=-1)
    {
      // now add the hat (a cone), scaled with pscale
    }
}







void GrobGlHelper::drawMessage(bool bitfont,double x, double y, double z, double xscale, double yscale, double zscale, const char *message,int axis,int startwidth,int endwidth)
{
  double zf=zscale/200;

  if (bitfont)
    drawBitmapText(GLUT_BITMAP_HELVETICA_18, x,y,z+zf, xscale, message);
  else
    {
      if (startwidth<=0)
	drawStrokeText(GLUT_STROKE_ROMAN, x,y,z+zf, xscale,yscale,zscale, message,axis);
      else
	{
	  glLineWidth(startwidth);
	  drawStrokeText(GLUT_STROKE_ROMAN, x,y,z+zf, xscale,yscale,zscale, message,axis);
	  glLineWidth(endwidth);
	}
    }
}

















int GrobGlHelper::glutbitmapwidth(void *font,const char *message)
{
  int width=0;
  int cwidth=8;

  // fltk or glut, one of them doesn't like glutbitmapwidth
  //      while (*message)
  //        {
  //	  width+glutBitmapWidth((int)GLUT_BITMAP_HELVETICA_18, (int)*message);
  //          width+glutBitmapWidth(1, (int)*message);
  //          message++;
  //        }

  if (font==GLUT_BITMAP_HELVETICA_10)
    cwidth=5;
  else if (font==GLUT_BITMAP_HELVETICA_12)
    cwidth=6;
  else if (font==GLUT_BITMAP_HELVETICA_18)
    cwidth=8;

  width=strlen(message)*cwidth;
  return width;
}




void Grob::makeTextBoxList(string titlestr,string substring,int width,int height)
{
  #ifdef GrobGL
  int dummynum,lownum,mednum,highnum,finenum;

  // for some reason i cannot figure out, the first displaylistnum of this title that gets created doesn't
  //  work, so we create it and then throw it away.
  dummynum=GrobGlHelper::makeTextBox(titlestr,substring,width,height,lowdetail);
  GrobGlHelper::freelistnum(dummynum);

  lownum=GrobGlHelper::makeTextBox(titlestr,substring,width,height,lowdetail);
  mednum=GrobGlHelper::makeTextBox(titlestr,substring,width,height,meddetail);
  highnum=mednum;
  finenum=mednum;
  set_gl(lownum,mednum,highnum,finenum,finenum);
  #endif
}



void GrobGlHelper::drawBitmapText(void *font,double x,double y,double z,double scale, const char *message)
{
  glPushMatrix();
  pushlighting();
  lightsoff();

  glRasterPos3f(x,y,z);
  glColor3f(1.0,1.0,1.0);

  while (*message)
    {
      // for some reason if i don't do this it flickers in the prescence of mult. windows
      glutBitmapCharacter(font, *message);
      message++;
    }

  poplighting();
  glPopMatrix();
}






void GrobGlHelper::drawStrokeText(void *font,double x,double y,double z,double xscale,double yscale,double zscale, const char *message,int axis)
{
  glPushMatrix();
  pushlighting();
  lightsoff();

  //  glColor3f(1.0,1.0,1.0);
  glTranslatef(x,y,z);
  glScalef(xscale, yscale, zscale);
  rotatetoaxis(axis);

  drawStrokeText(font,message);

  poplighting();
  glPopMatrix();
}




void GrobGlHelper::drawStrokeText(void *font,const char *message)
{
  // simple drawstroke, no properties preserved
  while (*message)
    {
      glutStrokeCharacter(font, *message);
      message++;
    }
}








void GrobGlHelper::drawBackbox(double x,double y,double z,double width,double height)
{
  GLfloat triColor[] = {0.1, 0.6, .1, 1};
  GLfloat specular[] = {1, 1, 1, 1};
  GLfloat shininess[] = {100};

  glPushMatrix();
  glBegin(GL_QUADS);

  //  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor);
  //  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  //  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

  glColor3f(.1,0.6,.1);

  //  glNormal3f(0.0, 0.0, 1.0);
  glVertex3f(x, y, z);
  glVertex3f(x+width, y, z);
  glVertex3f(x+width, y-height, z);
  glVertex3f(x, y-height, z);
  glColor3f(1,1,1);
  glEnd();
  glPopMatrix();
}



void GrobGlHelper::drawBox(float sx,float sy,float sz,float ex, float ey, float ez)
{
  glColor3f(1.0,1.0,1.0);
  glBegin(GL_LINE_LOOP);
  glVertex3f(sx, sy, sz);
  glVertex3f(ex, sy, sz);
  glVertex3f(ex, ey, sz);
  glVertex3f(sx, ey, sz);
  glVertex3f(sx, sy, sz);
  glVertex3f(sx, sy, ez);
  glVertex3f(ex, sy, ez);
  glVertex3f(ex, ey, ez);
  glVertex3f(sx, ey, ez);
  glVertex3f(sx, sy, ez);
  glEnd();
  glBegin(GL_LINES);
  glVertex3f(sx, ey, sz);
  glVertex3f(sx, ey, ez);
  glVertex3f(ex, sy, sz);
  glVertex3f(ex, sy, ez);
  glVertex3f(ex, ey, sz);
  glVertex3f(ex, ey, ez);
  glEnd();
}


void GrobGlHelper::drawWall(float sx,float sy,float sz,float ex, float ey, float ez,float thickness,WallType wtype,GrobDetail indetail)
{
  float dx=ex-sx;
  float dy=ey-sy;
  float dz=ez-sz;

  // nice yellow
  //  GLfloat triColor[] = {0.8, 0.6, .1, .2};

  //  GLfloat triColor[] = {0.1, 0.2, .6, .8};
  //  GLfloat specular[] = {.8, .9, 1, .1};
  //  GLfloat shininess[] = {.1};

  GLfloat noColor[] = {0, 0, 0, 1};
  GLfloat triColor1[] = {0.1, 0.5, .7, .2};
  GLfloat triColor2[] = {0.1, 0.4, .7, .5};
  GLfloat triColor3[] = {0.1, 0.3, .7, .8};
  GLfloat specular[] = {0, 0, 0, 1};
  GLfloat shininess[] = {0};

  pushlighting();

  // check for at least 2d
  if ((dx==0 && dy==0) || (dx==0 && dz==0) || (dy==0 && dz==0))
    indetail=lowdetail;
  
  glColor3f(.7,.7,.7);

  if (wtype==solid_wall && indetail!=lowdetail)
    {
      if (dx==0)
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, triColor1);
      else if (dy==0)
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, triColor2);
      else
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, triColor3);
      glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, noColor);
      glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
      glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);
      glBegin(GL_QUADS);
      lightson();
    }
  else
    {
      lightsoff();
      glBegin(GL_LINE_LOOP);
    }

  if (dx==0)
    {
      glNormal3f(1, .1, .1);
      glVertex3f(sx, sy, sz);
      glVertex3f(sx, ey, sz);
      glVertex3f(sx, ey, ez);
      glVertex3f(sx, sy, ez);
    }
  else if (dy==0)
    {
      glNormal3f(.1, 1, .1);
      glVertex3f(sx, sy, sz);
      glVertex3f(ex, sy, sz);
      glVertex3f(ex, sy, ez);
      glVertex3f(sx, sy, ez);
    }
  else if (dz==0)
    {
      glNormal3f(.1, .1, -1);
      glVertex3f(sx, sy, sz);
      glVertex3f(ex, sy, sz);
      glVertex3f(ex, ey, sz);
      glVertex3f(sx, ey, sz);
    }

  glEnd();
  glColor3f(1,1,1);

  poplighting();

  if (wtype==solid_wall && indetail!=lowdetail)
    {
      // to solid walls we also draw wire on top
      drawWall(sx,sy,sz,ex,ey,ez,thickness,wire_wall,lowdetail);
    }
}




int GrobGlHelper::makeTextBox(string titlestr, string substring, int width, int height,GrobDetail detail)
{
  int len1,len2,sy,y,y2,y3;
  int titx,tity,titw,tith;
  double z=-100;
  double zf=.5;
  
  int displaylistnum=startlist();
  lightsoff();
  
  // set up orthographic coord<->pixel projection
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0,width,0,height,-1*(fabs(z)+1),fabs(z)+1);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glViewport(0,0,width,height);
  
  // compute title sizes and locations
  sy=height-10;
  y=y2=y3=height-24;
  if (titlestr!="")
    {
      len1=GrobGlHelper::glutbitmapwidth(GLUT_BITMAP_HELVETICA_18,titlestr.c_str());
      y2-=15;
      y3-=15;
    }
  else
    len1=0;
  
  if (substring!="")
    {
      len2=GrobGlHelper::glutbitmapwidth(GLUT_BITMAP_HELVETICA_12,substring.c_str());
      y3-=15;
    }
  else
    len2=0;
  
  // compute size of title border
  if (len1>len2)
    titw=len1;
  else
    titw=len2;
  tith=y-y3;
  titx=(width-titw)/2;
  tity=sy;
  
  // draw a title border box?
  titw+=40;titx-=20;
  tith+=10;tity+=5;
  titx=5-1;titw=width-10;
  GrobGlHelper::drawBackbox(titx,tity,z,titw,tith);
  
  // draw titles
  if (true)
    {
      // bitmap version
      if (titlestr!="")
	GrobGlHelper::drawBitmapText(GLUT_BITMAP_HELVETICA_18,(width-len1)/2,y,-z+zf,1,titlestr.c_str());
      if (substring!="")
	GrobGlHelper::drawBitmapText(GLUT_BITMAP_HELVETICA_12,(width-len2)/2,y2,-z+zf,1,substring.c_str());
    }
  else if (false)
    {
      // stroke version
      if (titlestr!="")
	GrobGlHelper::drawStrokeText(GLUT_STROKE_ROMAN,(width-len1)/2,y,-z+zf,.12,titlestr.c_str());
      if (substring!="")
	GrobGlHelper::drawStrokeText(GLUT_STROKE_ROMAN,(width-len2)/2,y2,-z+zf,.08,substring.c_str());
    }
  endlist();  return displaylistnum;
}











int GrobGlHelper::makeAxes(double inradius, GrobDetail detail,int linewidth)
{
  // draw some nice label axes for the user to get their bearings
  int i,j;
  int boxcount=12;
  double viewradius=inradius/3;
  double boxsize=viewradius/30;
  double textsize=boxsize/20;
  double textwidth=textsize*70;
  GLfloat axes_ambuse2[] =   { 0.5, 0.0, 0.0, 1.0 };
  GLfloat axes_ambuse[] =   { 1.0, 0.0, 0.0, 1.0 };
  bool bitfont=false;
  int st,et;

  int displaylistnum=startlist();
  lightsoff();

  glMatrixMode(GL_MODELVIEW);
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, axes_ambuse);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, axes_ambuse);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, axes_ambuse);
  glBegin(GL_LINES);
  glLineWidth(1);
  glColor3f(1.0,1.0,1.0);
  glVertex3f(viewradius, 0, 0); glVertex3f(-viewradius, 0, 0);
  glVertex3f(0, viewradius, 0); glVertex3f(0, -viewradius, 0);
  glVertex3f(0, 0, viewradius); glVertex3f(0, 0, -viewradius);
  glEnd();

  if (detail!=(int)meddetail)
    {
      lightson();
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, axes_ambuse2);
      glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, axes_ambuse2);
      glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, axes_ambuse2);
      double boxstep=(viewradius+viewradius)/(boxcount+1);
      for (i = 0; i < 3; i++)
	{
	  glPushMatrix();
	  glTranslatef(((-viewradius)+boxstep)*(i==0), ((-viewradius)+boxstep)*(i==1), ((-viewradius)+boxstep)*(i==2));
	  for (j = 0; j < boxcount; j++)
	    {
	      glutSolidCube(boxsize);
	      glTranslatef(boxstep*(i==0), boxstep*(i==1), boxstep*(i==2));
	    }
	  glPopMatrix();
	}
    }

  // label axes
  if (bitfont)
    {
      drawMessage(true,viewradius,0,0,textsize,"X",3);
      drawMessage(true,0,viewradius,0,textsize,"Y",3);
      drawMessage(true,0,0,viewradius,textsize,"Z",3);
    }
  else
    {
      if (detail==(int)lowdetail || detail==(int)meddetail)
	st=et=-1;
      else
	{
	  st=linewidth;
	  et=st+1;
	}
      drawMessage(false,viewradius+textwidth*2,0,(textwidth/2),textsize,"X",3,st,et);
      drawMessage(false,-1*textwidth/2,viewradius,-1*textwidth,textsize,"Y",3,st,et);
      drawMessage(false,-1*textwidth/2,0,viewradius+textwidth*2,textsize,"Z",3,st,et);
    }
      
  endlist();  return displaylistnum;
}






















void GrobGlHelper::rotatetoaxis(int axis)
{
  // perform any rotations to come into alignment with axis (x=0,y=1,z=2)
  switch(axis)
    {
    case 0:
      break;
    case 1:
      //      glRotatef(-90,0,0,1);
      glRotatef(90,0,0,1);
      break;
    case 2:
      // silly, the below must be compressable
      //      glRotatef(180,1,0,0);
      //      glRotatef(-90,0,0,1);
      glRotatef(90,0,0,1);
      glRotatef(-90,0,1,0);
      break;
    case 3:
      // for orienting world axes, upright for person viewing normal world (z down, y into screen)
      glRotatef(180,0,0,1);
      glRotatef(-90,1,0,0);
      break;
    }
}










void GrobGlHelper::drawSquare(double x, double y, char solid)
{
  
  if (solid)
    glBegin(GL_QUADS);
  else
    glBegin(GL_LINE_LOOP);
  
  glNormal3f(0, 0, -1.0);
  glVertex3f(0, 0, 0);
  glVertex3f(x, 0, 0);
  glVertex3f(x, y, 0);
  glVertex3f(0, y, 0);
  glEnd();
}


















void GrobGlHelper::drawRoundCube(double xlength,double ylength,double zlength,double shift,char solid,GrobDetail detail)
{
  // Color
  //  GLfloat triColor2[] = {0.5, 1, 0.5, 1}; // green
  GLfloat triColor2[] = {.3, .5, .4, 1};
  GLfloat specular[] = {.3, .3, .3, .3};
  GLfloat shininess[] = {75};
  int slices,stacks;
  float radius;
  float length;

  glTranslatef(0,shift,0);

  if (detail!=paintdetail)
    {
      glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor2);
      glColor4fv(triColor2);
    }
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

  if (detail==lowdetail)
    {
      // use a box to approx. in lowdetail mode
      drawMechBox(xlength/1.3,ylength/1.1,zlength/2,solid);
    }
  else
    {
      switch(detail)
	{
	case lowdetail:
	  slices=7;
	  stacks=4;
	  break;
	case meddetail:
	  slices=7;
	  stacks=4;
	  break;
	case highdetail:
	  slices=9;
	  stacks=6;
	  break;
	case finedetail:
	  slices=9;
	  stacks=4;
	  break;
	case paintdetail:
	  slices=9;
	  stacks=4;
	  break;
	}
      xlength/=2.0;
      ylength/=2.0;
      zlength/=2.0;
      radius=xlength;
      glScalef(1.0,ylength/radius,zlength/radius);
      if (solid==1)
	glutSolidSphere(radius,slices,stacks);
      else
	glutWireSphere(radius,slices,stacks);
      glScalef(1.0,radius/ylength,-radius/zlength);


      // now draw anntennae
      GLfloat triColor2[] = {0.5, 0.4, 0.9, 1};
      //      GLfloat specular2[] = {.4, .4, .4, .4};
      //      GLfloat shininess2[] = {20};
      GLfloat specular2[] = {.7, .7, .7, .7};
      GLfloat shininess2[] = {70};

      if (detail!=paintdetail)
	{
	  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, triColor2);
	  glColor4fv(triColor2);
	}
      glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular2);
      glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess2);
      
      GLUquadricObj *upper = gluNewQuadric();
      if (solid==0)
	gluQuadricDrawStyle(upper, (GLenum) GLU_LINE);

      radius/=10;
      slices/=2;
      stacks/=2;
      slices=4;
      stacks=4;
      length=xlength/1.5;

      glPushMatrix();
      glTranslatef(-xlength/2.8,ylength/1.4,zlength/2.2);
      glRotatef(-8,0,1,0);
      //      glRotatef(55,1,0,0);
      glRotatef(-55,1,0,0);
      glScalef(1.5,1,6);
      if (solid==1)
	glutSolidCone(radius,length,slices,stacks);
      else
	glutWireCone(radius,length,slices,stacks);
      glPopMatrix();
      glTranslatef(xlength/2.8,ylength/1.4,zlength/2.2);
      glRotatef(8,0,1,0);
      //      glRotatef(55,1,0,0);
      glRotatef(-55,1,0,0);
      glScalef(1.5,1,6);
      if (solid==1)
      	glutSolidCone(radius,length,slices,stacks);
      else
      	glutWireCone(radius,length,slices,stacks);
    }
}
















void Grob::makeMeshList(int gridsx,int gridsy,float xdist,float ydist,float *heights)
{
  #ifdef GrobGL
  int lownum,mednum,highnum,finenum,paintnum;
  float thickness=.3;

  lownum=GrobGlHelper::makeMesh(gridsx,gridsy,xdist,ydist,heights,thickness,0,lowdetail);
  mednum=GrobGlHelper::makeMesh(gridsx,gridsy,xdist,ydist,heights,thickness,1,meddetail);;
  highnum=GrobGlHelper::makeMesh(gridsx,gridsy,xdist,ydist,heights,thickness,1,highdetail);
  finenum=highnum;
  set_gl(lownum,mednum,highnum,finenum,finenum);
  #endif
}








int GrobGlHelper::makeMesh(int gridsx,int gridsy,float xdist,float ydist,float *heights,float thickness, char solid,GrobDetail detail)
{
  int ix,iy;
  float x[3],y[3],z[3];
  int displaylistnum=startlist();
  lightson();

  // draw the surface grid, one cell at a time
  for (ix=0;ix<gridsx-1;++ix)
    {
      for (iy=0;iy<gridsy-1;++iy)
	{
	  //printf("Makemesh working on location %d,%d\n",ix,iy);
	  // each grid location is made up of two triangles
	  // first the upper right

	  x[0]=ix*xdist;
	  y[0]=iy*ydist;
	  z[0]=heights[iy*gridsx+ix];
	  x[1]=(ix+1)*xdist;
	  y[1]=iy*ydist;
	  z[1]=heights[iy*gridsx+(ix+1)];
	  x[2]=(ix+1)*xdist;
	  y[2]=(iy+1)*ydist;
	  z[2]=heights[(iy+1)*gridsx+(ix+1)];
	  drawTriangle(x[0],y[0],z[0],x[1],y[1],z[1],x[2],y[2],z[2],thickness,solid,detail);

	  // then the lower left
	  x[0]=ix*xdist;
	  y[0]=iy*ydist;
	  z[0]=heights[iy*gridsx+ix];
	  x[1]=ix*xdist;
	  y[1]=(iy+1)*ydist;
	  z[1]=heights[(iy+1)*gridsx+ix];
	  x[2]=(ix+1)*xdist;
	  y[2]=(iy+1)*ydist;
	  z[2]=heights[(iy+1)*gridsx+(ix+1)];
	  drawTriangle(x[0],y[0],z[0],x[1],y[1],z[1],x[2],y[2],z[2],thickness,solid,detail);
	}
    }

  endlist(); return displaylistnum;
}

























#endif


// }
// Namespace close-brace
