// grob.h
// Grob class for storing and drawing graphics in a somewhat abstract, flexible way

// forward declarations
class Grob;
class GrobGlHelper;



#ifndef INC_grobh
#define INC_grobh

#include "matrixinc.h"
#include "grobsymbols.h"
#include "grobcamera.h"
#include "cfpaxis.h"
#include "JRinterpolater.h"

#include "../../cfplibs/biobot/misc.def"


#include <FL/gl.h>
#include <FL/glut.H>

#include <string>

//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace Grob{






// ----------------------------------------------------------------------------------
// Grob - the main class for holding shapes and drawing them flexibly in dif. styles
//        and to different graphics libs
// ----------------------------------------------------------------------------------

class Grob
{
  friend class GrobVideoCamera;
private:
  static bool using_gl,using_pov,using_vrml;
public:
  static bool use_gl();
  static bool use_pov();
  static bool use_vrml();
  static bool supports_gl();
  static bool supports_pov();
  static bool supports_vrml();
private:
  float olmat[16];										// for drawing
  float oscale[3];									// for drawing
  GrobDetail detail;
  perVectorR3 location;
  perMatrix orientation;
public:
  Grob() {initialize();};
  ~Grob();
private:
  void initialize();
public:
  static void setupcontext();									// prepare context
public:
  void readlocation(FILE *fp);
  void writelocation(FILE *fp);
  void readorientation(FILE *fp);
  void writeorientation(FILE *fp);
  void setlocation(perVectorR3 inlocation);							// set location
  void setlocation(float inloc[3]);								// set location
  void setorientation(perMatrix inorientation);							// set orientation
  void setscale(perVectorR3 inscale);								// set scale
  void setscale(float inscale[3]);								// set scale
  void setscale(float x,float y,float z);							// set scale
  void setscale(int ind,float scale);
  void setdetail(GrobDetail indetail) {detail=indetail;gl_currentnum=gl_listnum[(int)indetail];};
  perVector getlocation() {return location;};							// accessor
  perMatrix getorientation() {return orientation;};						// accessor
  void setlocation(int index,float val) {location(index)=val;gl_needsupdate=true;};
  void setlocation(float ix,float iy, float iz) {location(0)=ix;location(1)=iy;location(2)=iz;gl_needsupdate=true;};
  float getlocation(int i){return location(i);};
  void makematrix();						// make gl matrix from location&orientation
public:
  void draw(GrobVideoCamera *cam);
  void draw(GrobVideoCamera *cam,GrobDetail indetail) {setdetail(indetail);draw(cam);};

  // GL Members
private:
  bool gl_needsupdate;										// flag for update
  int gl_listnum[GROBDETAILS];									// displaylists
  int gl_currentnum;									        // cur diplaylistnm
private:
  void gl_initialize();										// initialize
  void gl_uninitialize();									// uninitialize
  void gl_update();										// set up for draw
public:
  Grob(int inlow,int inmed,int inhigh,int infine,int inpaint){initialize();set_gl(inlow,inmed,inhigh,infine,inpaint);};
  Grob(int listnum) {initialize();set_gl(listnum,listnum,listnum,listnum,listnum);};		// quick gl const.
  void set_gl(GrobDetail indetail,int displaylistnum);						// set one detail
  void set_gl(int inlow,int inmed,int inhigh,int infine,int inpaint);				// set all detail
  void set_gl(int listnum) {set_gl(listnum,listnum,listnum,listnum,listnum);};			// set all detail
  void draw_gl();										// draw it
  bool ready_gl();										// ready for gldraw
public:
  void makePointList();
  void makeSphereList(double radius);
  void makeCubeList(double xlength,double ylength,double zlength);
  void makeRoundCubeList(double xlength,double ylength,double zlength,double shift);
  void makeConeList(double radius, double length);
  void makeLegconeList(double pradius,double dradius, double length,int color,int sleeve,
		       double offst,int endeffector);
  void makeTriangletileList(double p10,double p11,double p12,double p20,double p21,double p22,
			    double p30,double p31,double p32,double thickness);
  void makeArrowList(double length,int color);
  void makeTextBoxList(string titlestr,string substring,int width,int height);
  void makeAxesList(double viewradius,int linewidth=-1);
  void makeAxesList(VisibleAxis *xaxisp,VisibleAxis *yaxisp,VisibleAxis *zaxisp,float cxscale,float cyscale,float czscale,GrobSpacerType usespacer,int linewidth=-1,bool clearbox=false);
  void makeBarList(double xlength,double ylength,double zlength);
  void makeMeshList(int gridsx,int gridsy,float xdist,float ydist,float *heights);
public:
  void unusedetail(GrobDetail indetail);
  void replacelist(GrobDetail indetail,int newlistnum);
  void replaceLow();
  void replaceHigh();
public:
  static void drawline(double sx,double sy,double sz,double ex,double ey,double ez);
};






















#ifdef GrobGL


// ----------------------------------------------------------------------------------
// -- Grob GL Helper Class GrobGlHelper
// ----------------------------------------------------------------------------------
//: Implements some static functions which create convenient shapes

class GrobGlHelper
{
public:
  static int getlistnum();							// get a free gl display listnum
  static void freelistnum(int listnum);						// free a free gl display listnum
public:
  static void snifferrors(char *announce);
  static void SetMaterial(GLfloat spec[], GLfloat amb[], GLfloat diff[], GLfloat shin[]);
  static void setuplights();
private:
  static int startlist();
  static void endlist();
public:
  static int makePoint(int size,GrobDetail detail);
  static int makeSphere(double radius,int slices,int stacks, char solid,GrobDetail detail);
  static int makeCube(double xlength,double ylength,double zlength,char solid,GrobDetail detail);
  static int makeRoundCube(double xlength,double ylength,double zlength,double shift,char solid,GrobDetail detail);
  static int makeLegcone(double proximal_radius,double distal_radius, double length, int slices,int stacks,
			 double color, double color2, int sleeve,double offst, int endeffector, char solid,GrobDetail detail);
  static int makeTritile(double x1, double y1,double z1,double x2,double y2,double z2,double x3,double y3,
			 double z3,double thickness, char solid,GrobDetail detail);
  static int makeTexturetile(double x1, double y1,double z1,double x2,double y2,double z2,double x3,double y3,
			     double z3,double thickness, char solid,GrobDetail detail);
  static int makeCone(double radius,double length,int slices,int stacks,char solid,GrobDetail detail);
  static int makeArrow(double length, int slices,int stacks, int color, char solid,GrobDetail detail);
  static int makeAxes(double inradius, GrobDetail detail,int linewidth=-1);
  static int makeLine(double length);
  static int makeTextBox(string titlestr, string substring, int width, int height,GrobDetail detail);
  static int makeAxes(VisibleAxis *xaxisp,VisibleAxis *yaxisp,VisibleAxis *zaxisp,GrobDetail detail,float cxscale,float cyscale,float czscale,int linewidth=-1,bool clearbox=false,GrobSpacerType usespacer=box_spacer);
  static int makeBar(double xlength,double ylength,double zlength,GrobDetail detail);
  static int makeMesh(int gridsx,int gridsy,float xdist,float ydist,float *heights,float thickness, char solid,GrobDetail detail);
public:
  static void drawPoint(int size,GrobDetail detail);
  static void drawPoint(float x, float y, float z,GrobDetail detail);
  static void drawSphere(double radius,int slices,int stacks, char solid,GrobDetail detail);
  static void drawCube(double xlength,double ylength,double zlength,char solid,GrobDetail detail);
  static void drawRoundCube(double xlength,double ylength,double zlength,double shift,char solid,GrobDetail detail);
  static void drawLegcone(double proximal_radius,double distal_radius, double length, int slices,int stacks,
			 double color, double color2, int sleeve,double offst, int endeffector, char solid,GrobDetail detail);
  static void drawTriangle(double x1, double y1,double z1,double x2,double y2,double z2,double x3,double y3,
			 double z3,double thickness, char solid,GrobDetail detail);
  static void drawTritile(double x1, double y1,double z1,double x2,double y2,double z2,double x3,double y3,
			 double z3,double thickness, char solid,GrobDetail detail);
  static void drawTexturetile(double x1, double y1,double z1,double x2,double y2,double z2,double x3,double y3,
			     double z3,double thickness, char solid,GrobDetail detail);
  static void drawCone(double radius,double length,int slices,int stacks,double color1,double color2,char solid,GrobDetail detail);
  static void drawArrow(double length, int slices,int stacks, int color, char solid,GrobDetail detail);
  static void drawBar(double xlength,double ylength,double zlength,GrobDetail detail);
public:
  static void drawAxis(float *widthp,float *heightp,float *depthp,VisibleAxis *axisp,float sx,float sy,float sz,float ex,float ey,float ez,bool bitfont,GrobDetail indetail,float cxscale,float cyscale,float czscale,float space,int linewidth=-1);
  static void drawLine(double sx,double sy,double sz,double ex,double ey,double ez);
public:
  static void drawMechBox(double width, double height, double depth, char solid);
  static void drawMechBox2(double width, double height, double depth, char solid);
  static void drawMechCylinder(double baseradius,double topradius,double height,int slices,int stacks,char solid,
			       float color,int endeffector);
  static void drawDisk(GLUquadricObj *upper, double inradius, double baseradius, int slices, int stacks, char solid);
  static void drawLinearrow(double x1,double y1,double z1,double x2,double y2,double z2, int hatdetail);
  static void drawMessage(bool bitfont,double x, double y, double z, double xscale, double yscale, double zscale, const char *message,int axis=0,int startwidth=0,int endwidth=0);
  static void drawMessage(bool bitfont,double x, double y, double z, double scale, const char *message,int axis=0,int startwidth=0,int endwidth=0) {drawMessage(bitfont,x,y,z,scale,scale,scale,message,axis,startwidth,endwidth);};
  static void drawBitmapText(void *font,double x,double y,double z,double scale, const char *message);
  static void drawStrokeText(void *font,double x,double y,double z,double xscale,double yscale,double zscale, const char *message,int axis=0);
  static void drawStrokeText(void *font,double x,double y,double z,double scale, const char *message)
    {drawStrokeText(font,x,y,z,scale,scale,scale,message);};
  static void drawStrokeText(void *font,const char *message);
  static void drawBackbox(double x,double y,double z,double width,double height);
  static void drawBox(float sx,float sy,float sz,float ex, float ey, float ez);
  static void drawWall(float sx,float sy,float sz,float ex, float ey, float ez,float thickness,WallType wtype,GrobDetail indetail);
  static void drawFloor(VisibleAxis *xaxisp,VisibleAxis *yaxisp,VisibleAxis *zaxisp,float heightx,float heighty,float heightz,float depthx,float depthy,float depthz,float bsx,float bex,float bsy,float bey,float bsz2,float bez);
  static void drawAxisLabel(VisibleAxis *axisp,float *rwidth,float *rheight,float *rdepth,float sx,float sy,float sz,float length,float xscale, float yscale, float zscale, float space,GrobDetail detail,int axis=0,int linewidth=-1);
  static void drawSquare(double width, double height, char solid);
public:
  static void rotatetoaxis(int axis);
  static int glutbitmapwidth(void *font,const char *message);
  static float AxisWriter_Ticks(VisibleAxis *axisp,float xpos,float wscale,float hscale,float dscale,float width,float length,float space,int linewidth,float vertdir);
  static float AxisWriter_Label(VisibleAxis *axisp,string label,float wscale,float hscale,float dscale,float width,float space,int linewidth,float virtdir);
  static float AxisWriter_Enums(VisibleAxis *axisp,float wscale,float hscale,float dscale,float width,float space,int linewidth,float virdir);
  static float draw_labeltext(const char *message,float textscale,float wscale,float hscale,float dscale,float xpos,float width,float minscreensize,float virtdir,bool outsideskip=false,bool spacetest=false);
  static void axisnumbertext(char *charnum,float val,float min,float max);
public:
  static void lightson(){glEnable(GL_LIGHTING);};
  static void lightsoff(){glDisable(GL_LIGHTING);};
  static void pushlighting(){glPushAttrib(GL_LIGHTING_BIT);};
  static void poplighting(){glPopAttrib();};
  static float getlinewidth(){GLfloat width;glGetFloatv(GL_LINE_WIDTH,&width); return (float)width;};
  static void setlinewidth(float linewidth){glLineWidth(linewidth);};
  static void depthoverwrite() {glDepthFunc(GL_LEQUAL);};
  static void depthnormal() {glDepthFunc(GL_LESS);};
public:
  static JRinterpolater* rgbramp_hot(int levels=100);
  static JRinterpolater* rgbramp_cool(int levels=100);
  static JRinterpolater* rgbramp_jet(int levels=100);
  static JRinterpolater* rgbramp_hsv(int levels=100);
  static JRinterpolater* rgbramp_pink(int levels=100);
  static JRinterpolater* rgbramp_bone(int levels=100);
  static JRinterpolater* rgbramp_copper(int levels=100);
  static JRinterpolater* rgbramp_gray(int levels=100);
};



#endif



//#include "grobinline.h"







// }
// Namespace close-brace
#endif


