// grobcamera.h
// Camera classes for drawing Grobs

// forward delcarations
class GrobVideoCamera;
class GrobVideoCameraDrawer;



#ifndef INC_grobcamerah
#define INC_grobcamerah


#include "matrixinc.h"
#include "cfpaxis.h"

#include "grobsymbols.h"
#include "grob.h"

#include "../../cfplibs/biobot/misc.def"

#include <FL/gl.h>
#include <FL/glut.H>

#include <iostream>
#include <string>
#include <list>

//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace Grob{




// ----------------------------------------------------------------------------------
// CameraView
// ----------------------------------------------------------------------------------
//:Helper class which can be used to save and restore common viewing parameters (euler angles, zoom values, etc.)

class CameraView
  {
    friend class GrobVideoCamera;
  private:
    float e1,e2,e3;
    float zoomfactor;
    float ydolly;
    GrobProjectionMode pmode;
    bool use_eulers,use_dolly,use_zoomfactor,use_projectionmode;
  public:
    CameraView(float ine1,float ine2,float ine3);
    CameraView(float ine1,float ine2,float ine3,float inzoom);
    CameraView(float ine1,float ine2,float ine3,float inzoom,float inydolly);
    CameraView(float ine1,float ine2,float ine3,float inzoom,GrobProjectionMode inpmode);
  public:
    void seteulers(float ine1,float ine2,float ine3) {e1=ine1;e2=ine2;e3=ine3;use_eulers=true;};
    void setzoom(float inzoom) {zoomfactor=inzoom;use_zoomfactor=true;};
    void setydolly(float inydolly) {ydolly=inydolly;use_dolly=true;};
    void setpmode(GrobProjectionMode inpmode) {pmode=inpmode;use_projectionmode=true;};
    void restore(float *oe1,float *oe2,float *oe3,float *ozoom,float *oydolly,GrobProjectionMode *opmode);
  private:
    void initialize();
  };







// ----------------------------------------------------------------------------------
// GrobVideoCamera
// ----------------------------------------------------------------------------------

class GrobVideoCamera
{
  friend class Grob;
  friend class Configpp;
  friend class SWfltk_widget;
private:
  // specialize camera drawing parameters;
  string title, subtitle;
  bool gvc_prepared;
  GrobDetail detail;
  Grob *focus_object;
  GrobProjectionMode projectionmode;
  float distance;
  bool distance_lock;
  bool focus_lock;
  float aperture,aspect,znear,zfar;
  perVectorR3 focus_location;
  float zoomfactor;
  float viewradius;
  bool autocenter;
  bool autoscale;
  bool using_gl;
  int refreshrate;									// how often to refdraw
  int refreshcounter;									// how many request so far
  int redundantcounter;
  float cxscale,cyscale,czscale;
  float xskew,yskew,zskew;
  bool needsrecalc;
  bool needsredraw;
  int viewsx,viewsy,vieww,viewh;
  float clearcolor[4];
  int linewidth;
  bool clearbox;
  GrobSpacerType usespacer;
  int curview;
protected:
  VisibleAxis *xaxisp,*yaxisp,*zaxisp;
protected:
  vector<CameraView *> views;
protected:
  int win_width,win_height;
  int win_xpos,win_ypos;
protected:
  perVectorR3 location;
  perMatrix orientation;
  GrobVideoCameraDrawer *gdrawerp;
  bool redrawrequested;
  bool drawing_engaged;
protected:
  bool showaxes;
  bool showtitle;
  bool showlegend;
  bool showwalls;
  Grob *gaxes;
  Grob *gtitle;
  Grob *glegend;
  Grob *gwalls;
private:
  perVectorR3 basisa,basisb,basisc;
  perVectorR3 upvector;
  float tanthetah,tanthetav;
  float euler1,euler2,euler3;
  float q[4];
public:
  GrobVideoCamera(){construct();};
  ~GrobVideoCamera() {destruct();};
public:
  void setdetail(GrobDetail indetail) {detail=indetail;detailchange();};					// accessor
  void setidetail(int indetail) {setdetail((GrobDetail)indetail);};				// accessor
  GrobDetail getdetail(){return detail;};							// accessor
  void setaxes(VisibleAxis *inxaxisp,VisibleAxis *inyaxisp=NULL,VisibleAxis *inzaxisp=NULL);	// accessor
private:
  void construct();
  void gvc_prepare();
  void gvc_gl_prepare();
  void destruct();
  void gl_destruct();
  void detailchange();
  void add_defaultviews();
private:
  void makeaxes();
  void maketitle();
  void makelegend();
  void makewalls();
public:
  void begin_draw();							// must be called before starting draws
  void end_draw();							// must be called after ending draw
  void begin_incdraw();                  				// must be called before starting draws
  void end_incdraw();				                        // must be called before starting draws
  void end_drawinc();							// must be called after ending draw
  void grobdraw(Grob *grob);						// draw the grob
  void grobdraw(Grob *grob,GrobDetail expdetail);			// draw the grob w/ specific detail
  bool canincdraw() {if (needsredraw || !gvc_prepared) return false; else return true;};   // can inc draw
private:
  void recalc();
  void update_internals();
  void seteye();
  void show_lights_scene();
  void show_lights_headmounted();
  void show_title();
  void doautocenter(GrobVideoCameraDrawer *drawerp);
  void doautoscale(GrobVideoCameraDrawer *drawerp);
public:
  void setzoom(float inzoom){zoomfactor=inzoom;change();};				// accessor
  float getzoom(){return zoomfactor;};							// accessor
  void setdist(float indist){distance=indist;change();};				// accessor
  float getdist(){return distance;};							// accessor
  void setapt(float inapt){aperture=inapt;change();};					// accessor
  float getapt(){return aperture;};							// accessor
public:
  void settitle(string intit,string insub);						// accessor
  void setwinsize(int w,int h,int x=-1,int y=-1);					// accessor
  void setshowaxes(bool doshow=true){showaxes=doshow;needsredraw=true;};		// accessor
  void seteuler1(float val){euler1=val;needsrecalc=needsredraw=true;};				// accessor
  void seteuler2(float val){euler2=val;needsrecalc=needsredraw=true;};				// accessor
  void seteuler3(float val){euler3=val;needsrecalc=needsredraw=true;};				// accessor
  float geteuler1(){return euler1;};				// accessor
  float geteuler2(){return euler2;};				// accessor
  float geteuler3(){return euler3;};				// accessor
  void setquat(float inq[4]) {q[0]=inq[0];q[1]=inq[1];q[2]=inq[2];q[3]=inq[3];needsrecalc=needsredraw=true;};
  void setfocus(float x,float y, float z);
  void setautoscale(bool inas=true){autoscale=inas;};
  void setclearbox(bool inclearbox=true){clearbox=inclearbox;makeaxes();};
  void setspacer(GrobSpacerType inspacer){usespacer=inspacer;makeaxes();};
  void seteulers(float e1,float e2, float e3) {euler1=e1;euler2=e2;euler3=e3;needsrecalc=needsredraw=true;};
public:
  void setxdolly(float x){setfocus(x,focus_location(1),focus_location(2));};
  void setydolly(float y){setfocus(focus_location(0),y,focus_location(2));};
  void setzdolly(float z){setfocus(focus_location(0),focus_location(1),z);};
public:
  void setview(int viewnum);
  int addview(CameraView *cviewp);
  int saveview();
  void nextview() {if (++curview>=views.size()) curview=0;setview(curview);};
public:
  void setpos(perVectorR3 inlocation,perMatrix inorientation);				// set location
  void setlocation(perVectorR3 inlocation);						// set location
  void setorientation(perMatrix inorientation);						// set orientation
  void setlinewidth(int inlinewidth=-1);
  void biglinewidth(int inc=1);
private:
  void update_eulers();
  void update_quats();
  void location_change();
  void change();
private:
  void world2eye(float ix,float iy,float iz,float &ox,float &oy,float &oz);
  void eye2norm(float ix,float iy,float iz,float &ox,float &oy,float &oz);
  void norm2screen(float ix,float iy,float iz,float &ox,float &oy);
public:
  setdrawer(GrobVideoCameraDrawer *ingdrawerp);				// set the drawer of this window
  void update_notify(GrobVideoCameraDrawer *drawerp, bool need=false);	// the drawer has new stuff
  void axes_notify(GrobVideoCameraDrawer *drawerp, bool need=false);	// the drawer changed axes
  virtual void request_redraw(GrobVideoCameraDrawer *drawerp);		// tell the drawer to go ahead
};





//:The GrobVideoCameraDrawer class shold be inherited by any class (like a simulation world)
//  which wants to be registered as the official updater of the videocamera.
// The idea is the something like the GUI will automatically notify a GrobVideoCameraDrawer
//  when it needs the window/camera repainted.
class GrobVideoCameraDrawer
{
protected:
  int listeners;
  list<GrobVideoCamera *>::iterator gi;
  list<GrobVideoCamera *>::iterator gend;
  list<GrobVideoCamera *>::iterator gbegin;
  list<GrobVideoCamera *> gvc_listeners;
protected:
  VisibleAxis *xaxisp,*yaxisp,*zaxisp;
public:
  GrobVideoCameraDrawer();
  ~GrobVideoCameraDrawer() {;};
public:
  void add_vidcam_listener(GrobVideoCamera *newcam);				// add a videocam listener
  virtual void grobvidcam_draw(GrobVideoCamera *cam)=0;				// draw to camera (pure virtual)
  virtual bool grobvidcam_supportsincdraw() {return false;};                    // can we incremental draw?
  virtual bool grobvidcam_canincdraw() {return false;};                         // can we incremental draw?
  virtual void grobvidcam_incdraw(GrobVideoCamera *cam) {grobvidcam_draw(cam);};// incremental draw if pos.
  virtual GrobSpacerType wantspacer(){return box_spacer;};
  void getpaxes(VisibleAxis* &cxaxisp,VisibleAxis* &cyaxisp,VisibleAxis* &czaxisp);
protected:
  void cameranotify();						// tell any listeners that we have newstuff to draw
  void cameranotify_inc() {cameranotify();};			// tell any listeners that we have new inc to draw
  void cameranotify_axes();					// tell any listeners that we have new axes
public:
  virtual string drawertype(){return "unknown";};                       // rtti
};



//#include "grobcamerainline.h"




// }
// Namespace close-brace
#endif


